/*	Pin Configuration
 *
 *	PB0: M4-FLT
 *	PB1: M1-FLT
 *	PB2: CS
 *	PB3: MOSI
 *	PB4: MISO
 *	PB5: SCK
 *
 *	PC0: M4-AOUT
 *	PC1: M1-AOUT
 *	PC2: M2-AOUT
 *	PC3: M5-AOUT
 *	PC4: M3-AOUT
 *
 *	PD0: M5-FLT
 *	PD1: M2-FLT
 *	PD2: M3-FLT
 *	PD5: LED red
 *	PD6: LED green
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "motor_data.h"
#include "spi.h"

#define M1F PB1
#define M2F PD1
#define M3F PD2
#define M4F PB0
#define M5F PD0

#define M1A PC1
#define M2A PC2
#define M3A PC4
#define M4A PC0
#define M5A PC3

#define LEDR PD5
#define LEDG PD6

MotorData data;

struct _MotorInternal
{
	uint16_t offset[5];
} internal;

void adcCalibration();

int main()
{
	uint8_t i;

	// init ports
	DDRD = (1 << LEDR) | (1 << LEDG);

	PORTD = (1 << LEDR) | (1 << LEDG);

	PORTD &= ~(1 << LEDR);

	// init SPI
	SPIInit();

	// init ADC
	ADCSRA = (1 << ADEN) | (1 << ADIF) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);	// Enable, prescaler 128 = ~93kHz
	DIDR0 = 0x1F;

	// init data
	for(i = 0; i < 5; i++)
		data.current[i] = 0;
	data.fault = 0;

	uint16_t ch[5];

	adcCalibration();

	sei();

	PORTD |= (1 << LEDR);

	while(1)
	{
		SPIExchange(data.exchange.miso, data.exchange.mosi, MOTOR_DATA_SIZE);

		for(i = 0; i < 5; i++)
			ch[i] = 0;

		uint8_t iterations;
		for(iterations = 0; iterations < 10; iterations++)
		{
			for(i = 0; i < 5; i++)
			{
				ADMUX = i;
				ADCSRA |= (1 << ADSC);
				while(ADCSRA & (1 << ADSC));
				ch[i] += ADC;
			}
		}

		for(i = 0; i < 5; i++)
		{
			ch[i] /= iterations;

			if(ch[i] < internal.offset[i])
				ch[i] = 0;
			else
				ch[i] -= internal.offset[i];
		}

		data.current[0] = ch[1];
		data.current[1] = ch[2];
		data.current[2] = ch[4];
		data.current[3] = ch[0];
		data.current[4] = ch[3];

		uint32_t val = 0;

		for(i = 0; i < 5; i++)
		{
			val = data.current[i];
			val *= 20;	// => [Ve-4]
			val *= 100;
			val /= 40;	// OPV gain
			val /= 5;	// [uV]/[mOhm] = [mA]
			data.current[i] = val;
		}

		data.fault = 0;

		if(!(PINB & (1 << M1F)))
			data.fault |= 0x01;

		if(!(PIND & (1 << M2F)))
			data.fault |= 0x02;

		if(!(PIND & (1 << M3F)))
			data.fault |= 0x04;

		if(!(PINB & (1 << M4F)))
			data.fault |= 0x08;

		if(!(PIND & (1 << M5F)))
			data.fault |= 0x10;

		PORTD ^= (1 << LEDG);
	}

	return 0;
}

void adcCalibration()
{
	uint8_t i;

	// throw away first sample
	for(i = 0; i < 5; i++)
	{
		ADMUX = i;
		ADCSRA |= (1 << ADSC);
		while(ADCSRA & (1 << ADSC));
		internal.offset[i] = 0;
	}

	uint8_t iterations;
	for(iterations = 0; iterations < 10; iterations++)
	{
		for(i = 0; i < 5; i++)
		{
			ADMUX = i;
			ADCSRA |= (1 << ADSC);
			while(ADCSRA & (1 << ADSC));
			internal.offset[i] += ADC;
		}
	}

	for(i = 0; i < 5; i++)
		internal.offset[i] /= iterations;
}
