/*
 * main.c
 *
 *  Created on: 17.03.2011
 *      Author: AndreR
 *
 * Pin Configuration
 *
 * PB0: LED9G
 * PB1: LED10G
 * PB2: CS
 * PB3: MOSI
 * PB4: MISO
 * PB5: SCK
 *
 * PC0: 3V3-A / ADC0
 * PC1: 5V-A / ADC1
 * PC2: C1-A / ADC2
 * PC3: C2-A / ADC3
 * PC4: C3-A / ADC4
 * PC5: C4-A / ADC5
 * ADC7: VCC-A / ADC7
 *
 * PD0: LED1R
 * PD1: LED2R
 * PD2: LED3O
 * PD3: LED4O
 * PD4: LED5Y
 * PD5: LED6Y
 * PD6: LED7G
 * PD7: LED8G
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "power_data.h"
#include "spi.h"

#define LED1R PD0
#define LED2R PD1
#define LED3O PD2
#define LED4O PD3
#define LED5Y PD4
#define LED6Y PD5
#define LED7G PD6
#define LED8G PD7
#define LED9G PB0
#define LED10G PB1

PowerData data;

void setLED(uint16_t level);

int main()
{
	uint8_t i;

	// init ports
	DDRB = (1 << LED9G) | (1 << LED10G);
	DDRD = 0xFF;

	PORTB = (1 << LED9G) | (1 << LED10G);
	PORTD = 0xFF;

	PORTD = 0xFE;

	// init spi
	SPIInit();

	// init ADC
	ADCSRA = (1 << ADEN) | (1 << ADIF) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);	// Enable, prescaler 128 = ~93kHz
	DIDR0 = 0x3F;

	// init data
	for(i = 0; i < 4; i++)
		data.uCell[i] = 4000;
	data.i3V3 = 0;
	data.i5V = 0;
	data.iVCC = 0;

	uint16_t ch[7];

	sei();

	while(1)
	{
		SPIExchange(data.exchange.miso, data.exchange.mosi, POWER_DATA_SIZE);

		for(i = 0; i < 7; i++)
			ch[i] = 0;

		uint8_t iterations;
		for(iterations = 0; iterations < 10; iterations++)
		{
			for(i = 0; i < 6; i++)
			{
				ADMUX = i;
				ADCSRA |= (1 << ADSC);
				while(ADCSRA & (1 << ADSC));
				ch[i] += ADC;
			}

			ADMUX = 7;
			ADCSRA |= (1 << ADSC);
			while(ADCSRA & (1 << ADSC));
			ch[6] += ADC;
		}

		for(i = 0; i < 7; i++)
		{
			ch[i] /= iterations;
			ch[i] *= 20;	// convert to [Ve-4]
		}

		uint32_t val = 0;

		val = ch[5];   		// Cell 1
		val *= 21;      	// [Ve-4] => [Ve-5]
		val /= 100;     	// [Ve-5] => [Ve-3] = [mV]
		data.uCell[0] = val;

		val = ch[4];   		// Cell 2
		val *= 42;
		val /= 100;
		data.uCell[1] = val;

		val = ch[3];    	// Cell 3
		val *= 63;
		val /= 100;
		data.uCell[2] = val;

		val = ch[2];    	// Cell 4
		val *= 84;
		val /= 100;
		data.uCell[3] = val;

		val = ch[0];		// 3V3
		val /= 20;          // Gain
		val *= 100;
		val /= 100;         // [uV]/[mOhm] = [mA]
		data.i3V3 = val;

		val = ch[1];    	// 5V
		val /= 20;          // Gain
		val *= 100;
		val /= 100;         // [uV]/[mOhm] = [mA]
		data.i5V = val;

		val = ch[6];		// VCC
		val *= 100;
		val /= 27;          // Gain
		val /= 2;           // [uV]/[mOhm] = [mA]
		data.iVCC = val;

		uint16_t batLevel = data.uCell[3];

		data.uCell[3] -= data.uCell[2];
		data.uCell[2] -= data.uCell[1];
		data.uCell[1] -= data.uCell[0];

		// bat level 12.8V - 14.8V - 16.8V, steps: 0.4V
		uint16_t leds = 0;
		if(batLevel > 12800)
			leds |= 0x01;
		if(batLevel > 13200)
			leds |= 0x02;
		if(batLevel > 13600)
			leds |= 0x04;
		if(batLevel > 14000)
			leds |= 0x08;
		if(batLevel > 14400)
			leds |= 0x10;
		if(batLevel > 14800)
			leds |= 0x20;
		if(batLevel > 15200)
			leds |= 0x40;
		if(batLevel > 15600)
			leds |= 0x80;
		if(batLevel > 16000)
			leds |= 0x100;
		if(batLevel > 16400)
			leds |= 0x200;

		setLED(leds);
	}

	return 0;
}

void setLED(uint16_t level)
{
	static uint16_t state = 0;

	if(level == state)
		return;

	state = level;

	PORTD = ~(level & 0xFF);
	PORTB |= 0x03;
	PORTB &= ~((level >> 8) & 0x03);
}
