/*
 * spi.c
 *
 *  Created on: 16.03.2011
 *      Author: AndreR
 */

#include <avr/interrupt.h>

#include "power_data.h"
#include "spi.h"

#define SPI_DATA_SIZE POWER_DATA_SIZE/2+3

#define START_FLAG	0x3C
#define END_FLAG	0xA5

uint8_t spiRxBuf1[SPI_DATA_SIZE];
uint8_t spiTxBuf1[SPI_DATA_SIZE];
uint8_t spiRxBuf2[SPI_DATA_SIZE];
uint8_t spiTxBuf2[SPI_DATA_SIZE];

uint8_t* pFrontRxBuf;
uint8_t* pFrontTxBuf;
uint8_t* pBackRxBuf;
uint8_t* pBackTxBuf;

volatile uint8_t spiActive;
volatile uint8_t newData;
uint8_t spiCnt;
uint8_t nextTx;

ISR(SPI_STC_vect)
{
	SPDR = nextTx;	// be fast or be lost :)
	pFrontRxBuf[spiCnt] = SPDR;

	spiCnt++;
	nextTx = pFrontTxBuf[spiCnt];
}

ISR(PCINT0_vect)
{
	if(PINB & (1 << PB2))	// CS
	{
		// CS went high, slave deselected
		pFrontRxBuf[spiCnt] = SPDR;
		spiCnt = 0;
		spiActive = 0;
		newData = 1;
	}
	else
	{
		// CS went low, slave selected
		spiActive = 1;
		SPDR = START_FLAG;
		nextTx = pFrontTxBuf[0];
	}
}

void SPIInit()
{
	// init SPI
	DDRB |= (1 << PB4);

	spiCnt = 0;
	spiActive = 0;
	newData = 0;
	nextTx = 0;
	SPCR = (1 << SPIE) | (1 << SPE);

	// init PCINT for CS pin
	PCMSK0 |= (1 << PCINT2);
	PCICR |= (1 << PCIE0);

	for(uint8_t i = 0; i < SPI_DATA_SIZE; i++)
	{
		spiRxBuf1[i] = 0x00;
		spiTxBuf1[i] = 0x00;
		spiRxBuf2[i] = 0x00;
		spiTxBuf2[i] = 0x00;
	}

	pFrontRxBuf = spiRxBuf1;
	pBackRxBuf = spiRxBuf2;
	pFrontTxBuf = spiTxBuf1;
	pBackTxBuf = spiTxBuf2;
}

void SPISetTx(uint8_t* pData, uint8_t length)
{
	uint8_t i;
	uint8_t sendCRC = 0;

	for(i = 0; i < length; i++)
	{
		sendCRC ^= pData[i];
		pBackTxBuf[i] = pData[i];
	}

	pBackTxBuf[i] = sendCRC;
	pBackTxBuf[i+1] = END_FLAG;
}

uint8_t SPIGetRx(uint8_t* pData, uint8_t length)
{
	uint8_t i;
	uint8_t crc = 0;

	if(pBackRxBuf[0] != START_FLAG || pBackRxBuf[length+2] != END_FLAG)
		return 1;	// Error, FLAG mismatch

	for(i = 1; i < length+1; i++)
	{
		crc ^= pBackRxBuf[i];
	}

	if(crc != pBackRxBuf[i])
		return 1;	// Error, CRC mismatch

	for(i = 0; i < length; i++)
	{
		pData[i] = pBackRxBuf[i+1];
	}

	return 0;
}

uint8_t SPISwitchBuffer()
{
	uint8_t result = 1;

	uint8_t* pTmpRx = pFrontRxBuf;
	uint8_t* pTmpTx = pFrontTxBuf;

	uint8_t sreg = SREG;
	cli();

	if(spiActive == 0)
	{
		// switch buffers
		pFrontRxBuf = pBackRxBuf;
		pBackRxBuf = pTmpRx;

		pFrontTxBuf = pBackTxBuf;
		pBackTxBuf = pTmpTx;

		if(newData)
			result = 0;	// news!
	}

	SREG = sreg;

	return result;
}

void SPIExchange(uint8_t* pMisoData, uint8_t* pMosiData, uint8_t length)
{
	SPISetTx(pMisoData, length/2);
	if(SPISwitchBuffer() == 0)
		SPIGetRx(pMosiData, length/2);
}
