/*
 * cli.c
 *
 *  Created on: 29.05.2010
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"

#include "stm32f10x_rcc.h"

#include "FreeRTOS.h"
#include "task.h"

#include "atmega/ir_prog.h"
#include "atmega/motor_prog.h"
#include "atmega/power_prog.h"
#include "atmega/kicker_prog.h"
#include "hal/buzzer.h"
#include "hal/exti.h"
#include "hal/spi/lis3.h"
#include "ir.h"
#include "hal/motor.h"
#include "usb/usb.h"
#include "movement.h"
#include "power.h"
#include "kicker.h"
#include "hal/led.h"
#include "commands.h"
#include "acceleration.h"
#include "util/fastmath.h"

#include "network/dect_interface.h"
#include "network.h"
#include "network/udp.h"
#include "network/ipv4.h"
#include "network/ethernet.h"
#include "network/arp.h"
#include "network/igmpv3.h"

void processCommand();

void CLITask(void* params)
{
	params = params;

	static uint8_t buf[50];

	portTickType xLastWakeTime;
	portTickType xLastLEDTime;

	xLastWakeTime = xTaskGetTickCount();
	xLastLEDTime = xLastWakeTime;

	LEDMustOffBit(1);

	while(1)
	{
		if(DECTInterfaceGetMode() == DECT_MODE_CONFIG)
		{
			uint16_t bytesRead;
			do
			{
				DECTInterfaceConfigGet(buf, 10, &bytesRead);
				ConsolePut(buf, bytesRead);
			}
			while(bytesRead == 10);
		}

		if(ConsoleIsCmdReady())
		{
			ConsolePrint("\r\n");

			if(DECTInterfaceGetMode() == DECT_MODE_CONFIG)
			{
				if(ConsoleCmpCmd("exit"))
				{
					ConsolePrint("Leaving config mode...");
					DECTInterfaceSetMode(DECT_MODE_DATA);
					ConsolePrint("Done\r\n");
					ConsoleCmdProcessed();
					continue;
				}

				uint16_t len = 0;
				if(ConsoleGetCmd(buf, 50, &len) == 0)
				{
					DECTInterfaceConfigPut(buf, len);
					buf[0] = '\r';
					buf[1] = '\n';
					DECTInterfaceConfigPut(buf, 2);
				}
			}

			if(DECTInterfaceGetMode() == DECT_MODE_DATA)
			{
				if(ConsoleCmpCmd("dect config"))
				{
					ConsolePrint("Entering DECT config mode...");
					DECTInterfaceSetMode(DECT_MODE_CONFIG);
					ConsolePrint("Done\r\n");
					ConsoleCmdProcessed();
					continue;
				}

				processCommand();
			}

			ConsoleCmdProcessed();
		}

		// Heartbeat LED
		if(xLastWakeTime - xLastLEDTime > 500/portTICK_RATE_MS)
		{
			xLastLEDTime = xLastWakeTime;

			LEDToggleBit(0);
		}

		// Buttons
		if(ButtonsIsPressed(0))
		{
			MovementSetAutomatic(0, 0, 0, 0);

			for(uint8_t i = 0; i < 4; i++)
			{
				MovementSetPidSetpoint(i, 0);
				MovementSetManual(i, 0);
			}

			MovementSetDribble(0);

			KickerLoadManual(6, 78, 1);
			vTaskDelay(200/portTICK_RATE_MS);
			KickerFire(10000 | (3 << 14), 240, KICKER_DEVICE_STRAIGHT);
		}

		if(ButtonsIsPressed(1))
		{
			network.enableTimeout = 0;

			BuzzerEnable(1);
			vTaskDelay(200/portTICK_RATE_MS);
			BuzzerEnable(0);
		}

		vTaskDelayUntil(&xLastWakeTime, 50/portTICK_RATE_MS);
	}
}

void processCommand()
{
	int16_t val;
	int16_t val2[2];
	int16_t val3[3];

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 42);
	}

	if(ConsoleCmpCmd("accel get"))
	{
		ConsolePrint("--- LIS3 Inner ---\r\n");
		ConsolePrint("X: %d\r\n", acceleration.raw[0]);
		ConsolePrint("Y: %d\r\n", acceleration.raw[1]);
		ConsolePrint("Z: %d\r\n", acceleration.raw[2]);
		ConsolePrint("--- LIS3 Outer ---\r\n");
		ConsolePrint("X: %d\r\n", acceleration.raw[3]);
		ConsolePrint("Y: %d\r\n", acceleration.raw[4]);
		ConsolePrint("Z: %d\r\n", acceleration.raw[5]);
	}

	if(ConsoleCmpCmd("accel calib"))
	{
		ConsolePrint("Calibrating acceleration sensors\r\n");

		AccelerationCalibrate();
	}

	if(ConsoleCmpCmd("lis3 on"))
	{
		SPILIS3SetConfiguration(&acceleration.inner, 1);
	}

	if(ConsoleCmpCmdI16("led set", 1, &val))
	{
		ConsolePrint("Setting LEDs to 0x%X\r\n", val);

		LEDSet(val);
	}

	if(ConsoleCmpCmdI16("led on", 1, &val))
	{
		ConsolePrint("Setting LED %d on\r\n", val);

		LEDMustOnBit(val);
	}

	if(ConsoleCmpCmdI16("led off", 1, &val))
	{
		ConsolePrint("Setting LED %d off\r\n", val);

		LEDMustOffBit(val);
	}

	if(ConsoleCmpCmd("led get"))
	{
		ConsolePrint("LED State: 0x%X\r\n", LEDGetState());
	}

	if(ConsoleCmpCmd("power get"))
	{
		ConsolePrint("--- Supply infos ---\r\n");
		ConsolePrint("I Vcc:   %d mA\r\n", power.atData.iVCC);
		ConsolePrint("I 5V:    %d mA\r\n", power.atData.i5V);
		ConsolePrint("I 3.3V:  %d mA\r\n", power.atData.i3V3);
		ConsolePrint("U Cell1: %d mV\r\n", power.atData.uCell[0]);
		ConsolePrint("U Cell2: %d mV\r\n", power.atData.uCell[1]);
		ConsolePrint("U Cell3: %d mV\r\n", power.atData.uCell[2]);
		ConsolePrint("U Cell4: %d mV\r\n", power.atData.uCell[3]);
	}

	if(ConsoleCmpCmdI16("power 5", 1, &val))
	{
		ConsolePrint("Setting 5V power to %d\r\n", val);

		PowerEnable5VSupply(val);
	}

	if(ConsoleCmpCmd("motor get"))
	{
		ConsolePrint("--- Motor infos ---\r\n");
		ConsolePrint("M 1: %d mA\r\n", movement.atData.current[0]);
		ConsolePrint("M 2: %d mA\r\n", movement.atData.current[1]);
		ConsolePrint("M 3: %d mA\r\n", movement.atData.current[2]);
		ConsolePrint("M 4: %d mA\r\n", movement.atData.current[3]);
		ConsolePrint("M 5: %d mA\r\n", movement.atData.current[4]);
		ConsolePrint("F  : %X\r\n", movement.atData.fault);
	}

	if(ConsoleCmpCmd("ir status"))
	{
		ConsolePrint("--- IR levels ---\r\n");
		ConsolePrint("Left:    %d mV\r\n", ir.data.level[IR_LEFT]);
		ConsolePrint("Center:  %d mV\r\n", ir.data.level[IR_CENTER]);
		ConsolePrint("Right:   %d mV\r\n", ir.data.level[IR_RIGHT]);
		ConsolePrint("Barrier: %d mV\r\n", ir.data.level[IR_BARRIER]);
	}

	if(ConsoleCmpCmd("ir sig"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&ir.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&ir.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&ir.atmega);
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&ir.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", ir.atmega.signature[0], ir.atmega.signature[1], ir.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", ir.atmega.fuses[0], ir.atmega.fuses[1], ir.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("ir prog"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&ir.atmega);
		if(result == 0)
		{
			result = ATMegaChipErase(&ir.atmega);
			if(result == 0)
			{
				result = ATMegaReadSignature(&ir.atmega);
				if(result == 0)
				{
					result = ATMegaReadFuses(&ir.atmega);
					if(result == 0)
					{
						result = ATMegaLoadProgram(&ir.atmega, ir_prog_data, IR_PROG_SIZE);
						if(result == 0)
						{
							result = ATMegaVerifyProgram(&ir.atmega, ir_prog_data, IR_PROG_SIZE);
						}
					}
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&ir.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", ir.atmega.signature[0], ir.atmega.signature[1], ir.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", ir.atmega.fuses[0], ir.atmega.fuses[1], ir.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("ir fuse"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&ir.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&ir.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&ir.atmega);
				if(result == 0)
				{
					result = ATMegaWriteFuses(&ir.atmega, 0xC0, 0xDF, 0xF9);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&ir.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", ir.atmega.signature[0], ir.atmega.signature[1], ir.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", ir.atmega.fuses[0], ir.atmega.fuses[1], ir.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("ir verify"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&ir.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&ir.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&ir.atmega);
				if(result == 0)
				{
					result = ATMegaVerifyProgram(&ir.atmega, ir_prog_data, IR_PROG_SIZE);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&ir.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
	}

	if(ConsoleCmpCmd("move sig"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&movement.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&movement.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&movement.atmega);
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&movement.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", movement.atmega.signature[0], movement.atmega.signature[1], movement.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", movement.atmega.fuses[0], movement.atmega.fuses[1], movement.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("move prog"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&movement.atmega);
		if(result == 0)
		{
			result = ATMegaChipErase(&movement.atmega);
			if(result == 0)
			{
				result = ATMegaReadSignature(&movement.atmega);
				if(result == 0)
				{
					result = ATMegaReadFuses(&movement.atmega);
					if(result == 0)
					{
						result = ATMegaLoadProgram(&movement.atmega, motor_prog_data, MOTOR_PROG_SIZE);
						if(result == 0)
						{
							result = ATMegaVerifyProgram(&movement.atmega, motor_prog_data, MOTOR_PROG_SIZE);
						}
					}
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&movement.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", movement.atmega.signature[0], movement.atmega.signature[1], movement.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", movement.atmega.fuses[0], movement.atmega.fuses[1], movement.atmega.fuses[2]);
	}


	if(ConsoleCmpCmd("move verify"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&movement.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&movement.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&movement.atmega);
				if(result == 0)
				{
					result = ATMegaVerifyProgram(&movement.atmega, motor_prog_data, MOTOR_PROG_SIZE);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&movement.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", movement.atmega.signature[0], movement.atmega.signature[1], movement.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", movement.atmega.fuses[0], movement.atmega.fuses[1], movement.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("move fuse"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&movement.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&movement.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&movement.atmega);
				if(result == 0)
				{
					result = ATMegaWriteFuses(&movement.atmega, 0xC0, 0xDF, 0xF9);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&movement.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", movement.atmega.signature[0], movement.atmega.signature[1], movement.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", movement.atmega.fuses[0], movement.atmega.fuses[1], movement.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("power sig"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&power.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&power.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&power.atmega);
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&power.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", power.atmega.signature[0], power.atmega.signature[1], power.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", power.atmega.fuses[0], power.atmega.fuses[1], power.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("power prog"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&power.atmega);
		if(result == 0)
		{
			result = ATMegaChipErase(&power.atmega);
			if(result == 0)
			{
				result = ATMegaReadSignature(&power.atmega);
				if(result == 0)
				{
					result = ATMegaReadFuses(&power.atmega);
					if(result == 0)
					{
						result = ATMegaLoadProgram(&power.atmega, power_prog_data, POWER_PROG_SIZE);
						if(result == 0)
						{
							result = ATMegaVerifyProgram(&power.atmega, power_prog_data, POWER_PROG_SIZE);
						}
					}
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&power.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", power.atmega.signature[0], power.atmega.signature[1], power.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", power.atmega.fuses[0], power.atmega.fuses[1], power.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("power fuse"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&power.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&power.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&power.atmega);
				if(result == 0)
				{
					result = ATMegaWriteFuses(&power.atmega, 0xC0, 0xDF, 0xF9);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&power.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", power.atmega.signature[0], power.atmega.signature[1], power.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", power.atmega.fuses[0], power.atmega.fuses[1], power.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("kicker sig"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&kicker.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&kicker.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&kicker.atmega);
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&kicker.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", kicker.atmega.signature[0], kicker.atmega.signature[1], kicker.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", kicker.atmega.fuses[0], kicker.atmega.fuses[1], kicker.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("kicker prog"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&kicker.atmega);
		if(result == 0)
		{
			result = ATMegaChipErase(&kicker.atmega);
			if(result == 0)
			{
				result = ATMegaReadSignature(&kicker.atmega);
				if(result == 0)
				{
					result = ATMegaReadFuses(&kicker.atmega);
					if(result == 0)
					{
						result = ATMegaLoadProgram(&kicker.atmega, kicker_prog_data, KICKER_PROG_SIZE);
						if(result == 0)
						{
							result = ATMegaVerifyProgram(&kicker.atmega, kicker_prog_data, KICKER_PROG_SIZE);
						}
					}
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&kicker.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", kicker.atmega.signature[0], kicker.atmega.signature[1], kicker.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", kicker.atmega.fuses[0], kicker.atmega.fuses[1], kicker.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("kicker fuse"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&kicker.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&kicker.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&kicker.atmega);
				if(result == 0)
				{
					result = ATMegaWriteFuses(&kicker.atmega, 0xC0, 0xDF, 0xF9);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&kicker.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", kicker.atmega.signature[0], kicker.atmega.signature[1], kicker.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", kicker.atmega.fuses[0], kicker.atmega.fuses[1], kicker.atmega.fuses[2]);
	}

	if(ConsoleCmpCmdI16("motor power", 2, val2))
	{
		ConsolePrint("Motor %d power: %d\r\n", val2[0], val2[1]);

		MotorSetPower(val2[0], val2[1]);
	}

	if(ConsoleCmpCmdI16("motor hall", 1, &val))
	{
		int pos;
		switch(val)
		{
			case 0: pos = M1Pos; break;
			case 1: pos = M2Pos; break;
			case 2: pos = M3Pos; break;
			case 3: pos = M4Pos; break;
			case 4: pos = M5Pos; break;
			default: pos = 42; break;
		}

		ConsolePrint("Motor %d hall position: %d\r\n", val, pos);
	}

	if(ConsoleCmpCmd("motor mode"))
	{
		ConsolePrint("Motor mode: %d\r\n", movement.mode);
	}

	if(ConsoleCmpCmd("timeout"))
	{
		ConsolePrint("Timeout state: %d\r\n", network.enableTimeout);
	}

	if(ConsoleCmpCmdI16("enc get", 1, &val))
	{
		if(val < 0 || val > 4)
			return;

		ConsolePrint("Getting encoder count on enc %d\r\n", val);

		MovementSetMode(MOTOR_MODE_MANUAL);

		int32_t n_enc = movement.M[val].pTIM->CNT;
		movement.M[val].pTIM->CNT = 0;

		ConsolePrint("Encoder count %d\r\n", n_enc);
	}

	if(ConsoleCmpCmd("buzz on"))
	{
		ConsolePrint("Enabling buzzer\r\n");

		BuzzerEnable(1);
	}

	if(ConsoleCmpCmd("buzz off"))
	{
		ConsolePrint("Disabling buzzer\r\n");

		BuzzerEnable(0);
	}

	if(ConsoleCmpCmdI16("buzz freq", 1, &val))
	{
		ConsolePrint("Setting buzzer frequency to: %dHz\r\n", val);

		BuzzerSetFrequency(val);
	}

	if(ConsoleCmpCmd("kicker status"))
	{
		ConsolePrint("--- Kicker Status ---\r\n");
		ConsolePrint("Cap: %d\r\n", kicker.latestStatus.cap);
		ConsolePrint("Chg: %d\r\n", kicker.latestStatus.chg);
		ConsolePrint("Q1 T: %d\r\n", kicker.latestStatus.TIGBT[0]);
		ConsolePrint("Q2 T: %d\r\n", kicker.latestStatus.TIGBT[1]);
	}

	if(ConsoleCmpCmdI16("kicker fire1", 2, val2))
	{
		uint8_t level = val2[0];
		uint16_t time = val2[1];
		ConsolePrint("Firing 1, time: %d, level: %d\r\n", time, level);

		KickerFire(time, level, KICKER_DEVICE_STRAIGHT);
	}

	if(ConsoleCmpCmdI16("kicker fire2", 2, val2))
	{
		uint8_t level = val2[0];
		uint16_t time = val2[1];
		ConsolePrint("Firing 2, time: %d, level: %d\r\n", time, level);

		KickerFire(time, level, KICKER_DEVICE_CHIP);
	}

	if(ConsoleCmpCmdI16("kicker fan", 1, &val))
	{
		uint8_t level = val;
		ConsolePrint("Setting FAN speed: %d\r\n", level);

		KickerSetFan(KICKER_FAN_MANUAL, val);
	}

	if(ConsoleCmpCmdI16("kicker charge", 3, val3))
	{
		uint16_t on = val3[0];
		uint16_t off = val3[1];
		uint16_t time = val3[2];
		ConsolePrint("Charging, on: %d, off: %d, time: %d\r\n", on, off, time);

		KickerLoadManual(on, off, time);
	}

	if(ConsoleCmpCmdI16("kicker auto", 1, &val))
	{
		uint16_t maxCap = val;
		ConsolePrint("Enabling Auto-Charge, limit: %d\r\n", maxCap);

		KickerLoadAuto(maxCap);
	}

	if(ConsoleCmpCmdI16("pid set", 2, val2))
	{
		ConsolePrint("Setting PID %d to: %d\r\n", val2[0], val2[1]);

		MovementSetPidSetpoint(val2[0], val2[1]);
	}

	int32_t val32;

	if(ConsoleCmpCmdI32("sin", 1, &val32))
	{
		ConsolePrint("Sine of %d is %d\r\n", val32, fastsin(val32));
	}

	if(ConsoleCmpCmdI32("cos", 1, &val32))
	{
		ConsolePrint("Cosine of %d is %d\r\n", val32, fastcos(val32));
	}

	if(ConsoleCmpCmdI16("power red", 1, &val))
	{
		ConsolePrint("Setting power.blinkRed to %d\r\n", val);

		power.blinkRed = val;
	}

	if(ConsoleCmpCmdI16("power buzz", 1, &val))
	{
		ConsolePrint("Setting power.buzzerOn to %d\r\n", val);

		power.buzzerOn = val;
	}

	// --- Network Commands ---

	uint8_t netBuf[100];
	static uint16_t ident = 0;

	if(ConsoleCmpCmd("udp test"))
	{
		netBuf[0] = 42;
		netBuf[1] = 24;

		IPv4Address src = IPv4AddressSet(192, 168, 7, 42);
		IPv4Address dst = IPv4AddressSet(224, 0, 0, 42);
		MAC srcMAC = MACSet(0x02, 0x00, 0x00, 0x00, 0x00, 0x2A);	// locally administered, 02-00-00-00-00-2A
		MAC dstMAC = MACSet(0x03, 0x00, 0x5E, 0x00, 0x00, 0x2A);	// locally administered, multicast: 224.0.0.42

		uint16_t encLength = 0;

		ConsolePrint("Sending UDP packet\r\n");

		if(UDPEncode(netBuf, 100, 40, &encLength, 7777, 4242))
			return;

		ConsolePrint("Length after UDP: %d\r\n", encLength);

		if(IPv4Encode(netBuf, 100, encLength, &encLength, ident, src, dst, IPV4_PROTOCOL_UDP))
			return;

		ConsolePrint("Length after IPv4: %d\r\n", encLength);

		if(EthernetEncode(netBuf, 100, encLength, &encLength, srcMAC, dstMAC, ETHERNET_PROTOCOL_IPV4))
			return;

		ConsolePrint("Length after Ethernet: %d\r\n", encLength);

		if(DECTInterfacePut(netBuf, encLength))
			return;

		ConsolePrint("UDP packet sent!\r\n");

		ident++;
	}

	if(ConsoleCmpCmd("arp test"))
	{
		netBuf[0] = 42;
		netBuf[1] = 24;

		IPv4Address src = IPv4AddressSet(192, 168, 7, 42);
		IPv4Address dst = IPv4AddressSet(192, 168, 7, 7);
		MAC srcMAC = MACSet(0x02, 0x00, 0x00, 0x00, 0x00, 0x2A);	// locally administered, 02-00-00-00-00-2A
		MAC dstMAC = MACSet(0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF);	// locally administered, multicast: 224.0.0.42

		uint16_t encLength = 0;

		ConsolePrint("Sending ARP request\r\n");

		if(ARPEncode(netBuf, 100, &encLength, 1, srcMAC, src, dstMAC, dst))
			return;

		ConsolePrint("Length after ARP: %d\r\n", encLength);

		if(EthernetEncode(netBuf, 100, encLength, &encLength, srcMAC, dstMAC, ETHERNET_PROTOCOL_ARP))
			return;

		ConsolePrint("Length after Ethernet: %d\r\n", encLength);

		if(DECTInterfacePut(netBuf, encLength))
			return;

		ConsolePrint("ARP packet sent!\r\n");
	}

	if(ConsoleCmpCmd("mcast test"))
	{
		PacketHeader header;
		header.cmd = CMD_SYSTEM_STATUS_MOVEMENT;
		header.len = LEN_SYSTEM_STATUS_MOVEMENT;
		header.section = SECTION_SYSTEM;

		SystemStatusMovement status;
		status.vx = 42;
		status.vy = 4242;
		status.vw = 24;

		int16_t result = NetworkSendPacket(&header, (uint8_t*)&status, 1);

		ConsolePrint("Send packet returned: %d\r\n", result);
	}

	if(ConsoleCmpCmd("igmp test"))
	{
		ConsolePrint("Sending IGMPv3 Membership Report\r\n");

		uint8_t data[IGMPV3_GROUP_RECORD_SIZE];

		IGMPv3GroupRecord* pRecord = (IGMPv3GroupRecord*)data;

		pRecord->auxDataLen = 0;
		pRecord->multicastAddress = IPv4AddressSet(225, 42, 42, 42);
		pRecord->numberOfSources = 0;
		pRecord->recordType = IGMPV3_MODE_IS_EXCLUDE;

		int16_t result = NetworkSendIGMPv3MembershipReport(1, data, IGMPV3_GROUP_RECORD_SIZE);

		ConsolePrint("Result: %d\r\n", result);
	}

	if(ConsoleCmpCmd("netstat"))
	{
		ConsolePrint("--- Network Statistic ---\r\n");
		ConsolePrint("RX bytes:   %d\r\n", network.lastStats.rx.bytes);
		ConsolePrint("RX packets: %d\r\n", network.lastStats.rx.packets);
		ConsolePrint("RX drops:   %d\r\n", network.lastStats.rx.drops);
		ConsolePrint("TX bytes:   %d\r\n", network.lastStats.tx.bytes);
		ConsolePrint("TX packets: %d\r\n", network.lastStats.tx.packets);
		ConsolePrint("TX drops:   %d\r\n", network.lastStats.tx.drops);
	}

	if(ConsoleCmpCmd("announcement"))
	{
		PacketHeader header;
		header.cmd = CMD_SYSTEM_ANNOUNCEMENT;
		header.len = LEN_SYSTEM_ANNOUNCEMENT;
		header.section = SECTION_SYSTEM;

		SystemAnnouncement announce;
		announce.cpuId[0] = network.cpuId[0];
		announce.cpuId[1] = network.cpuId[1];
		announce.cpuId[2] = network.cpuId[2];

		NetworkSendPacket(&header, (uint8_t*)&announce, 1);
	}

	if(ConsoleCmpCmd("version"))
	{
		ConsolePrint("Firmware Version: 1.0.2 - 10.06.2011\r\n");
	}
}
