/*
 * dect.c
 *
 *  Created on: 14.02.2011
 *      Author: AndreR
 */

#include "dect_interface.h"

#include "stm32f10x_usart.h"
#include "stm32f10x_gpio.h"
#include "stm32f10x_rcc.h"
#include "stm32f10x_dma.h"
#include "misc.h"

#include "FreeRTOS.h"
#include "task.h"

#include "util/bit_banding.h"
#include "util/critical.h"
#include "inet.h"

#define RX_STATE_IDLE			1
#define RX_STATE_PARSING		2
#define RX_STATE_PARSING_FIRST	3
#define RX_STATE_MASKED			4

#define DECT_CTS	GPIO_Pin_3
#define DECT_RTS	GPIO_Pin_4
#define DECT_BOOT2	GPIO_Pin_4
#define DECT_TXD	GPIO_Pin_5
#define DECT_RXD	GPIO_Pin_6
#define DECT_RSTB	GPIO_Pin_7
#define DECT_DSR	GPIO_Pin_9
#define DECT_DTR	GPIO_Pin_10
#define DECT_BOOT3	GPIO_Pin_10
#define DECT_DCD	GPIO_Pin_11

#define DMAClearEn() BBPeriphReset((uint32_t)&DMA1_Channel7->CCR, 0)
#define DMASetEn()	BBPeriphSet((uint32_t)&DMA1_Channel7->CCR, 0)
#define DMAGetEn()	BBPeriphGet((uint32_t)&DMA1_Channel7->CCR, 0)

#define DMAClearTCI() BBPeriphReset((uint32_t)&DMA1_Channel7->CCR, 1)
#define DMASetTCI() BBPeriphSet((uint32_t)&DMA1_Channel7->CCR, 1)
#define DMAGetTCI() BBPeriphGet((uint32_t)&DMA1_Channel7->CCR, 1)

#define USART2SetRXI() BBPeriphSet(USART2_BASE + 0x0C, 5)
#define USART2ClearRXI() BBPeriphReset(USART2_BASE + 0x0C, 5)
#define USART2GetRXI() BBPeriphGet(USART2_BASE + 0x0C, 5)

DECTInterface dectInterface;

typedef struct __AtomicLockInfo
{
	uint8_t restoreState;
	uint8_t run;
} _AtomicLockInfo;

static __inline__ void __iRestoreRXI(_AtomicLockInfo* pLock)
{
	if(pLock->restoreState)
		USART2SetRXI();

	__asm__ volatile("" ::: "memory");
}

static __inline__ _AtomicLockInfo __iClearRXI()
{
	_AtomicLockInfo lock;
	lock.run = 1;
	lock.restoreState = USART2GetRXI() ? 1 : 0;

	USART2ClearRXI();

	return lock;
}

#define RXI_CRITICAL_SECTION() \
for(_AtomicLockInfo __infoRXI __attribute((__cleanup__(__iRestoreRXI))) = __iClearRXI(); __infoRXI.run; __infoRXI.run = 0)

static __inline__ void __iRestoreTCI(_AtomicLockInfo* pLock)
{
	if(pLock->restoreState)
		DMASetTCI();

	__asm__ volatile("" ::: "memory");
}

static __inline__ _AtomicLockInfo __iClearTCI()
{
	_AtomicLockInfo lock;
	lock.run = 1;
	lock.restoreState = DMAGetTCI() ? 1 : 0;

	DMAClearTCI();

	return lock;
}

#define TCI_CRITICAL_SECTION() \
for(_AtomicLockInfo __infoTCI __attribute((__cleanup__(__iRestoreTCI))) = __iClearTCI(); __infoTCI.run; __infoTCI.run = 0)

void dectParser(uint8_t data);
void setGPIOToBoot();
void setGPIOToOperational();
void startDMATx();
uint8_t lockRxFifo(void* pData);
void releaseRxFifo(uint8_t state, void* pData);
uint8_t lockTxFifo(void* pData);
void releaseTxFifo(uint8_t state, void* pData);
int16_t HDLCDecode(uint8_t* pData, uint16_t srcLength, uint8_t** ppDecoded, uint16_t* pDecodedLength);
int16_t HDLCEncode(uint8_t* pBuffer, uint16_t bufferLength, uint16_t dataLength, uint16_t* pEncodedLength, uint8_t frameNum);

uint8_t lockRxFifo(void* pData)
{
	pData = pData;

	uint8_t state = USART2GetRXI() ? 1 : 0;

	USART2ClearRXI();

	return state;
}

void releaseRxFifo(uint8_t state, void* pData)
{
	pData = pData;

	if(state)
		USART2SetRXI();
}

uint8_t lockTxFifo(void* pData)
{
	pData = pData;

	uint8_t state = DMAGetTCI() ? 1 : 0;

	DMAClearTCI();

	return state;
}

void releaseTxFifo(uint8_t state, void* pData)
{
	pData = pData;

	if(state)
		DMASetTCI();
}

void USART2_IRQHandler()
{
	uint32_t sr = USART2->SR;

	if(sr & USART_SR_TXE)	// enabled by TXEIE
	{
		// Transmitter empty, cleared by write to DR
	}

	if(sr & USART_SR_TC)	// enabled by TCIE
	{
		// transmission completed, only set when TXE is also set
		// indicates shift register is also empty
		// cleared by: read SR (already done), write DR
	}

	if(sr & USART_SR_ORE)	// enabled by RXNEIE
	{
		// overrun error, shift register received complete data
		// but DR was not read yet
		// cleared by: read SR (already done), read DR
		// contents of DR will not be lost, instead shift
		// register is overwritten if new data arrives
	}

	if(sr & USART_SR_RXNE)	// enabled by RXNEIE
	{
		// read data register not empty, cleared by: read DR
		uint8_t data = USART2->DR;

		if(dectInterface.mode == DECT_MODE_DATA)
		{
			dectParser(data);
		}

		if(dectInterface.mode == DECT_MODE_CONFIG)
		{
			if(dectInterface.rxConfig.EmptyBytes > 0)
				FifoPush(&dectInterface.rxConfig, data);
		}
	}
}

void DMA1_Channel7_IRQHandler()
{
	if(DMA1->ISR & DMA1_IT_TC7)
	{
		// transfer complete, disable DMA
		DMAClearEn();
		DMA1->IFCR = DMA1_IT_GL7;	// clear flags

		uint16_t packetLength;
		int16_t result = FifoPacketSyncedPop(&dectInterface.HDLC.tx, &packetLength, dectInterface.HDLC.dmaTxBuf, DECT_DMA_TX_SIZE);
		if(result == 0)	// OK
		{
			// popped a packet (hrrrr)
			dectInterface.stats.tx.packets++;
			dectInterface.stats.tx.bytes += packetLength;
			// reconfigure DMA

			DMA1_Channel7->CNDTR = packetLength;
			DMA1_Channel7->CMAR = (uint32_t)dectInterface.HDLC.dmaTxBuf;
			DMASetEn();	// enable DMA
		}

		if(result == ERROR_NOT_ENOUGH_MEMORY)
		{
			FifoPacketSyncedDrop(&dectInterface.HDLC.tx);
			dectInterface.stats.tx.drops++;
		}
	}
	else
	{
		DMA1->IFCR = DMA1_IT_GL7;	// clear flags
	}
}

void dectParser(uint8_t data)
{
	switch(dectInterface.rxState)
	{
		case RX_STATE_IDLE:
		{
			if(data == 0x7E)	// start flag?
				dectInterface.rxState = RX_STATE_PARSING_FIRST;
		}
		break;
		case RX_STATE_PARSING_FIRST:
		{
			if(data != 0x7E)
			{
				FifoPacketSyncedBegin(&dectInterface.HDLC.rx);
				FifoPacketSyncedPushByte(&dectInterface.HDLC.rx, data);

				dectInterface.stats.rx.bytes++;

				dectInterface.rxState = RX_STATE_PARSING;
			}
			else
			{
				// we received a flag after a flag,
				// that means the first flag was an end flag, not a start flag
				// stay in PARSING_FIRST
			}
		}
		break;
		case RX_STATE_PARSING:
		{
			if(data == 0x7E)	// end flag?
			{
				FifoPacketSyncedComplete(&dectInterface.HDLC.rx);

				dectInterface.stats.rx.packets++;

				dectInterface.rxState = RX_STATE_IDLE;

				if(dectInterface.rxCallback)
					(*dectInterface.rxCallback)();
			}
			else
			{
				if(data == 0x7D)	// mask flag
				{
					dectInterface.rxState = RX_STATE_MASKED;
				}
				else	// normal data
				{
					if(FifoPacketSyncedPushByte(&dectInterface.HDLC.rx, data))
					{
						FifoPacketSyncedCancel(&dectInterface.HDLC.rx);
						dectInterface.rxState = RX_STATE_IDLE;	// packet too big => lost
						dectInterface.stats.rx.drops++;
					}
					else
					{
						dectInterface.stats.rx.bytes++;
					}
				}
			}
		}
		break;
		case RX_STATE_MASKED:
		{
			data ^= 0x20;

			if(FifoPacketSyncedPushByte(&dectInterface.HDLC.rx, data))
			{
				FifoPacketSyncedCancel(&dectInterface.HDLC.rx);
				dectInterface.rxState = RX_STATE_IDLE;	// packet too big => lost
				dectInterface.stats.rx.drops++;
			}
			else
			{
				dectInterface.stats.rx.bytes++;
				dectInterface.rxState = RX_STATE_PARSING;
			}
		}
		break;
		default: break;
	}
}

void DECTInterfaceInit()
{
	dectInterface.mode = DECT_MODE_DATA;

	dectInterface.txDataInUse = xSemaphoreCreateMutex();
	dectInterface.rxDataInUse = xSemaphoreCreateMutex();

	DECTResetStats();

	dectInterface.rxState = RX_STATE_IDLE;
	dectInterface.rxCallback = 0;

	FifoInit(&dectInterface.rxConfig, dectInterface.HDLC.rxData, DECT_HDLC_RX_SIZE);

	FifoPacketSyncedInit(&dectInterface.HDLC.rx, dectInterface.HDLC.rxData, DECT_HDLC_RX_SIZE, lockRxFifo, releaseRxFifo, 0);
	FifoPacketSyncedInit(&dectInterface.HDLC.tx, dectInterface.HDLC.txData, DECT_HDLC_TX_SIZE, lockTxFifo, releaseTxFifo, 0);

	// DMA init
	DMA1_Channel7->CCR = DMA_CCR7_PL | DMA_CCR7_MINC | DMA_CCR7_DIR | DMA_CCR7_TCIE;
	DMA1_Channel7->CNDTR = 0;	// bytes to transmit
	DMA1_Channel7->CPAR = (uint32_t)&USART2->DR;	// peripheral address
	DMA1_Channel7->CMAR = (uint32_t)dectInterface.HDLC.dmaTxBuf;

	// init port
	GPIO_InitTypeDef gpioInit;

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART2, ENABLE);

	GPIO_PinRemapConfig(GPIO_Remap_USART2, ENABLE);

	gpioInit.GPIO_Mode = GPIO_Mode_AF_PP;
	gpioInit.GPIO_Pin = DECT_RTS | DECT_TXD;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOD, &gpioInit);

	gpioInit.GPIO_Mode = GPIO_Mode_IN_FLOATING;
	gpioInit.GPIO_Pin = DECT_CTS | DECT_RXD | DECT_DSR | DECT_DCD;
	GPIO_Init(GPIOD, &gpioInit);

	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
	gpioInit.GPIO_Pin = DECT_RSTB | DECT_DTR;
	GPIO_Init(GPIOD, &gpioInit);

	GPIOD->BRR = DECT_DTR;

	NVIC_InitTypeDef NVICInit;
	NVICInit.NVIC_IRQChannel = USART2_IRQn;
	NVICInit.NVIC_IRQChannelCmd = ENABLE;
	NVICInit.NVIC_IRQChannelPreemptionPriority = IRQL_DECT_USART2;
	NVICInit.NVIC_IRQChannelSubPriority = 0x00;
	NVIC_Init(&NVICInit);

	NVICInit.NVIC_IRQChannel = DMA1_Channel7_IRQn;
	NVICInit.NVIC_IRQChannelPreemptionPriority = IRQL_DECT_DMA;
	NVIC_Init(&NVICInit);
}

uint8_t DECTInterfaceGetMode()
{
	return dectInterface.mode;
}

void DECTResetStats()
{
	dectInterface.stats.rx.bytes = 0;
	dectInterface.stats.rx.packets = 0;
	dectInterface.stats.rx.drops = 0;

	dectInterface.stats.tx.bytes = 0;
	dectInterface.stats.tx.packets = 0;
	dectInterface.stats.tx.drops = 0;
}

// Uses task delay, call from task only
// not re-entrant safe!
void DECTInterfaceSetMode(uint8_t mode)
{
	uint8_t boot2 = 1;
	uint8_t boot3 = 1;

	if(mode == DECT_MODE_TRANSITION)
		return;	// manual transition to transition mode is not allowed!

	USART_InitTypeDef USARTInit;
	USARTInit.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
	USARTInit.USART_Parity = USART_Parity_No;
	USARTInit.USART_StopBits = USART_StopBits_1;
	USARTInit.USART_WordLength = USART_WordLength_8b;

	switch(mode)
	{
		case DECT_MODE_CONFIG:
		{
			USARTInit.USART_BaudRate = 9600;
			USARTInit.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
		}
		break;
		case DECT_MODE_DATA:
		{
			boot2 = 0;
			USARTInit.USART_BaudRate = 230400;
			USARTInit.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
		}
		break;
		default:
		{
			return;
		}
	}

	setGPIOToBoot();

	RXI_CRITICAL_SECTION()
	{
		TCI_CRITICAL_SECTION()
		{
			dectInterface.mode = DECT_MODE_TRANSITION;
			DMAClearEn();	// stop DMA

			FifoClear(&dectInterface.rxConfig);
			dectInterface.rxState = RX_STATE_IDLE;

			FifoPacketSyncedClear(&dectInterface.HDLC.tx);
			FifoPacketSyncedClear(&dectInterface.HDLC.rx);

			DECTResetStats();

			GPIOD->BRR = DECT_RSTB;	//initiate reset

			if(boot2)
				GPIOD->BSRR = DECT_BOOT2;
			else
				GPIOD->BRR = DECT_BOOT2;

			if(boot3)
				GPIOD->BSRR = DECT_BOOT3;
			else
				GPIOD->BRR = DECT_BOOT3;

			vTaskDelay(1000/portTICK_RATE_MS);

			GPIOD->BSRR = DECT_RSTB;

			setGPIOToOperational();

			USART_Init(USART2, &USARTInit);
			USART2->CR3 |= USART_CR3_DMAT;
			USART2->CR1 |= USART_CR1_UE;

			dectInterface.mode = mode;
		}
	}

	USART2SetRXI();
}

void setGPIOToBoot()
{
	BBPeriphReset(USART2_BASE + 0x14, 8);	//disable RTS
	BBPeriphReset(USART2_BASE + 0x0C, 13);	//disable UART

	//RTS = OUT PP, DTR = OUT PP
	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
	gpioInit.GPIO_Pin = DECT_RTS;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOD, &gpioInit);
}

void setGPIOToOperational()
{
	//RTS = AF PP, DTR = OUT PP
	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_AF_PP;
	gpioInit.GPIO_Pin = DECT_RTS;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOD, &gpioInit);

	BBPeriphSet(USART2_BASE + 0x0C, 13);	//enable UART
	BBPeriphSet(USART2_BASE + 0x14, 8);		//enable RTS
}

void DECTInterfaceSetRxCallback(DectRxCallback cb)
{
	dectInterface.rxCallback = cb;
}

int16_t	DECTInterfaceConfigPut(uint8_t* pSrc, uint16_t len)
{
	if(dectInterface.mode != DECT_MODE_CONFIG)
		return ERROR_DECT_INVALID_MODE;

	int16_t result = 0;

	TASK_MUTEX(dectInterface.txDataInUse)
	{
		result = FifoPacketSyncedBegin(&dectInterface.HDLC.tx);
		if(result)
			return result;

		result = FifoPacketSyncedPush(&dectInterface.HDLC.tx, len, pSrc);
		if(result)
		{
			FifoPacketSyncedCancel(&dectInterface.HDLC.tx);
			return result;
		}

		result = FifoPacketSyncedComplete(&dectInterface.HDLC.tx);

		startDMATx();
	}

	return result;
}

int16_t	DECTInterfaceConfigGet(uint8_t* pDst, uint16_t len, uint16_t* pBytesRead)
{
	if(dectInterface.mode != DECT_MODE_CONFIG)
		return ERROR_DECT_INVALID_MODE;

	RXI_CRITICAL_SECTION()
	{
		TASK_MUTEX(dectInterface.rxDataInUse)
		{
			uint16_t bytesUsed = dectInterface.rxConfig.Size - dectInterface.rxConfig.EmptyBytes;

			len = len < bytesUsed ? len : bytesUsed;

			for(uint16_t i = 0; i < len; i++)
				FifoPop(&dectInterface.rxConfig, pDst+i);
		}
	}

	*pBytesRead = len;

	return 0;
}

int16_t	DECTInterfacePut(uint8_t* pSrc, uint16_t len)
{
	if(dectInterface.mode != DECT_MODE_DATA)
		return ERROR_DECT_INVALID_MODE;

	int16_t result = 0;

	// HDLC stuff
	HDLCHeader header;

	header.first = 1;
	header._unused1 = 2;
	header.frameNum = dectInterface.frameNum;
	header.last = 1;
	header._unused2 = 1;
	header.ctrl = 3;

	uint16_t crc = pppfcs16(header._data, HDLC_HEADER_SIZE, PPPINITFCS16);
	crc = pppfcs16(pSrc, len, crc);

	TASK_MUTEX(dectInterface.txDataInUse)
	{
		if(dectInterface.HDLC.tx.fifo.EmptyBytes < 6)
			return ERROR_DECT_BUFFER_TOO_SMALL;

		result = FifoPacketSyncedBegin(&dectInterface.HDLC.tx);
		if(result)
			return result;

		FifoPacketSyncedPushByte(&dectInterface.HDLC.tx, 0x7E);	// start flag
		FifoPacketSyncedPush(&dectInterface.HDLC.tx, HDLC_HEADER_SIZE, header._data);

		for(uint16_t i = 0; i < len; i++)
		{
			if(dectInterface.HDLC.tx.fifo.EmptyBytes < 2)
			{
				FifoPacketSyncedCancel(&dectInterface.HDLC.tx);
				return ERROR_DECT_BUFFER_TOO_SMALL;
			}

			switch(pSrc[i])
			{
				case 0x7E:
				{
					FifoPacketSyncedPushByte(&dectInterface.HDLC.tx, 0x7D);
					FifoPacketSyncedPushByte(&dectInterface.HDLC.tx, 0x5E);
				}
				break;
				case 0x7D:
				{
					FifoPacketSyncedPushByte(&dectInterface.HDLC.tx, 0x7D);
					FifoPacketSyncedPushByte(&dectInterface.HDLC.tx, 0x5D);
				}
				break;
				default:
				{
					FifoPacketSyncedPushByte(&dectInterface.HDLC.tx, pSrc[i]);
				}
				break;
			}
		}

		if(dectInterface.HDLC.tx.fifo.EmptyBytes < 3)
		{
			FifoPacketSyncedCancel(&dectInterface.HDLC.tx);
			return ERROR_DECT_BUFFER_TOO_SMALL;
		}

		FifoPacketSyncedPush(&dectInterface.HDLC.tx, 2, (uint8_t*)&crc);
		FifoPacketSyncedPushByte(&dectInterface.HDLC.tx, 0x7E);

		FifoPacketSyncedComplete(&dectInterface.HDLC.tx);

		dectInterface.frameNum++;
		dectInterface.frameNum %= 4;

		startDMATx();
	}

	return result;
}

int16_t DECTInterfaceGet(uint8_t* pData, uint16_t srcLength, uint8_t** ppDecoded, uint16_t* pDecodedLength)
{
	if(dectInterface.mode != DECT_MODE_DATA)
		return ERROR_DECT_INVALID_MODE;

	int16_t result = 0;
	uint16_t bytesRead;

//	TASK_CRITICAL_SECTION()	// TODO: use semaphore
	{
		result = FifoPacketSyncedPop(&dectInterface.HDLC.rx, &bytesRead, pData, srcLength);
		if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
			FifoPacketSyncedDrop(&dectInterface.HDLC.rx);
		if(result)
			return result;
	}

	if(bytesRead < HDLC_HEADER_SIZE + HDLC_TRAILER_SIZE)
		return ERROR_HDLC_INVALID_PACKET;

	HDLCTrailer* pTrailer = (HDLCTrailer*)(pData + bytesRead - HDLC_TRAILER_SIZE);

	uint8_t* pCRCStart = pData;
	uint16_t crcLength = bytesRead;

	pTrailer->crc ^= 0xFFFF;

	uint16_t crc = pppfcs16(pCRCStart, crcLength, PPPINITFCS16);

	if(crc != PPPGOODFCS16)
		return ERROR_HDLC_BAD_CHECKSUM;

	*pDecodedLength = bytesRead - HDLC_HEADER_SIZE - HDLC_TRAILER_SIZE;

	*ppDecoded = pData + HDLC_HEADER_SIZE;

	return 0;
}

// call only within TCI_CRITICAL_SECTION and TASK_CRITICAL_SECTION! DMA interrupt must be disabled!
void startDMATx()
{
	if(dectInterface.mode == DECT_MODE_TRANSITION)
		return;	// USART not configured during transition

	if(DMAGetEn() == 0)
	{
		// dma not enabled => not running
		uint16_t packetLength;
		int16_t result = FifoPacketSyncedPop(&dectInterface.HDLC.tx, &packetLength, dectInterface.HDLC.dmaTxBuf, DECT_DMA_TX_SIZE);
		if(result == 0)	// OK
		{
			// popped a packet (hrrrr)
			dectInterface.stats.tx.packets++;
			dectInterface.stats.tx.bytes += packetLength;
			// reconfigure DMA

			DMA1_Channel7->CNDTR = packetLength;
			DMA1_Channel7->CMAR = (uint32_t)dectInterface.HDLC.dmaTxBuf;
			DMASetEn();	// enable DMA, start TX
		}

		if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
		{
			FifoPacketSyncedDrop(&dectInterface.HDLC.tx);
			dectInterface.stats.tx.drops++;
		}
	}
}
