/*
 * power.c
 *
 * Power control and overcurrent watchdog.
 *
 *  Created on: 25.10.2010
 *      Author: AndreR
 */

#include "power.h"

#include "stm32f10x.h"
#include "stm32f10x_gpio.h"
#include "stm32f10x_iwdg.h"

#include "FreeRTOS.h"
#include "task.h"

#include "commands.h"
#include "network.h"
#include "hal/motor.h"
#include "atmega/power_prog.h"
#include "hal/buzzer.h"
#include "hal/led.h"
#include "kicker.h"
#include "util/console.h"

#include "spi1_2.h"

#define POWER_5V_ENABLE (1 << 8)	// GPIOD
#define POWER_ATMEGA_CS (1 << 12)	// GPIOB
#define POWER_ATMEGA_RST (1 << 12)	// GPIOE

PowerGlobal power;

void PowerPreInit()
{
	GPIO_InitTypeDef gpioInit;

	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
	gpioInit.GPIO_Pin = POWER_5V_ENABLE;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	GPIO_Init(GPIOD, &gpioInit);
}

void PowerInit()
{
	GPIO_InitTypeDef gpioInit;

	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
	gpioInit.GPIO_Pin = POWER_ATMEGA_CS;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	GPIO_Init(GPIOB, &gpioInit);

	gpioInit.GPIO_Pin = POWER_ATMEGA_RST;
	GPIO_Init(GPIOE, &gpioInit);

	GPIOB->BSRR = POWER_ATMEGA_CS;
	GPIOE->BSRR = POWER_ATMEGA_RST;

	ATMegaInit(&power.atmega, &spi2, GPIOB, POWER_ATMEGA_CS, GPIOE, POWER_ATMEGA_RST);

	power.state = POWER_STATE_GOOD;
	power.atData.uCell[0] = 4000;
	power.atData.uCell[1] = 4000;
	power.atData.uCell[2] = 4000;
	power.atData.uCell[3] = 4000;

	power.blinkRed = 0;
	power.buzzerOn = 0;
}

void PowerTask(void* params)
{
	params = params;

	portTickType xLastWakeTime = xTaskGetTickCount();
	portTickType xLogTime = xLastWakeTime;
	portTickType xBeepTime = xLastWakeTime;
	portTickType xLEDTime = xLastWakeTime;
	portTickType xLowBatTimeStart = 0;
	portTickType xOffBatTimeStart = 0;

	uint8_t newState = POWER_STATE_GOOD;

	uint16_t beepDelay = 100;
	uint8_t beepIter = 0;
	uint16_t beepCycles[4] = {100, 250, 100, 2000};

	uint16_t ledDelay = 100;
	uint8_t ledIter = 0;
	uint16_t ledCycles[6] = {100, 100, 100, 100, 200, 800};

	int16_t result = 0;

	while(1)
	{
		IWDG_ReloadCounter();

		if(xOffBatTimeStart != 0 && xLastWakeTime - xOffBatTimeStart > 10000/portTICK_RATE_MS)
			newState = POWER_STATE_OFF;
		else if(xLowBatTimeStart != 0 && xLastWakeTime - xLowBatTimeStart > 10000/portTICK_RATE_MS)
			newState = POWER_STATE_LOW;
		else
			newState = POWER_STATE_GOOD;

		// Power level state machine
		if(power.state == POWER_STATE_GOOD && newState == POWER_STATE_LOW)
		{
			// enable buzzer
			power.buzzerOn = 1;

			power.state = newState;
		}

		if(power.state != POWER_STATE_OFF && newState == POWER_STATE_OFF)
		{
			power.buzzerOn = 0;
			power.blinkRed = 1;

			power.state = newState;

			MotorEmergencyShutdown();

			KickerLoadManual(6, 78, 1);
			vTaskDelay(200/portTICK_RATE_MS);
			KickerFire(10000 | (3 << 14), 240, KICKER_DEVICE_STRAIGHT);

			PowerEnable5VSupply(0);
		}

		if(power.buzzerOn)
		{
			if(xLastWakeTime - xBeepTime > beepDelay/portTICK_RATE_MS)
			{
				xBeepTime = xLastWakeTime;

				beepDelay = beepCycles[beepIter++];
				beepIter %= 4;
				if(beepIter % 2 == 0)
					BuzzerEnable(0);
				else
					BuzzerEnable(1);
			}
		}
		else
		{
			BuzzerEnable(0);
		}

		if(power.blinkRed)
		{
			if(xLastWakeTime - xLEDTime > ledDelay/portTICK_RATE_MS)
			{
				xLEDTime = xLastWakeTime;

				ledDelay = ledCycles[ledIter++];
				ledIter %= 6;
				if(ledIter % 2 == 0)
					LEDMustOff(LED_MAIN_RED);
				else
					LEDMustOn(LED_MAIN_RED);
			}
		}
		else
		{
			LEDMustOff(LED_MAIN_RED);
		}

		if(xLastWakeTime - xLogTime > 500/portTICK_RATE_MS)
		{
			xLogTime = xLastWakeTime;

			SystemPowerLog log;
			log.iVCC = power.atData.iVCC;
			log.i5V = power.atData.i5V;
			log.i3V3 = power.atData.i3V3;
			log.u1 = power.atData.uCell[0];
			log.u2 = power.atData.uCell[1];
			log.u3 = power.atData.uCell[2];
			log.u4 = power.atData.uCell[3];

			PacketHeader header;
			header.section = SECTION_SYSTEM;
			header.cmd = CMD_SYSTEM_POWER_LOG;
			header.len = LEN_SYSTEM_POWER_LOG;

			NetworkSendPacket(&header, (uint8_t*)&log, 0);

			result = ATMegaExchange(&power.atmega, power.atData.exchange.mosi, power.atData.exchange.miso, POWER_DATA_SIZE);
			if(result)
				ConsolePrint("Power ATMegaExchange code: 0x%X\r\n", result);

			if(power.atData.uCell[0] > 5000)
				ConsolePrint("Cell 0 abnormal: %d\r\n", power.atData.uCell[0]);
			if(power.atData.uCell[1] > 5000)
				ConsolePrint("Cell 1 abnormal: %d\r\n", power.atData.uCell[1]);
			if(power.atData.uCell[2] > 5000)
				ConsolePrint("Cell 2 abnormal: %d\r\n", power.atData.uCell[2]);
			if(power.atData.uCell[3] > 16000)
				ConsolePrint("Cell 3 abnormal: %d\r\n", power.atData.uCell[3]);
			if(power.atData.iVCC > 10000)
				ConsolePrint("iVCC abnormal: %d\r\n", power.atData.iVCC);
			if(power.atData.i3V3 > 10000)
				ConsolePrint("i3V3 abnormal: %d\r\n", power.atData.i3V3);
			if(power.atData.i5V > 10000)
				ConsolePrint("i5V abnormal: %d\r\n", power.atData.i5V);

			/* LOG
			 * Cell 2 abnormal: 48128
iVCC abnormal: 14336 (0x3800)
i3V3 abnormal: 13312 (0x3400)
i5V abnormal: 35328  (0x8A00)

Cell 2 abnormal: 48128 (0xBC00)
iVCC abnormal: 16896   (0x4200)
i3V3 abnormal: 13312   (0x3400)
i5V abnormal: 61440    (0xF000)

Cell 2 abnormal: 48128 (0xBC00)
i5V abnormal: 34304    (0x8600)

assumption: 1 byte shift => ATmega TX late!

Cell 2 abnormal: 48128
i5V abnormal: 34304

Cell 2 abnormal: 48128
i5V abnormal: 34304

Cell 2 abnormal: 48128
i5V abnormal: 34304
			 *
			 */

			uint8_t i;
			uint8_t low = 0;
			uint8_t off = 0;
			for(i = 0; i < 4; i++)
			{
				if(power.atData.uCell[i] < 3200)
					low = 1;
				if(power.atData.uCell[i] < 3000)
					off = 1;
			}

			if(power.atData.uCell[0] < 100 && power.atData.uCell[1] < 100 && power.atData.uCell[2] < 100 && power.atData.uCell[3] > 10000)
			{
				// Stationary power supply connected
			}
			else
			{
				if(low)
				{
					if(xLowBatTimeStart == 0)
						xLowBatTimeStart = xLastWakeTime;
				}
				else
					xLowBatTimeStart = 0;

				if(off)
				{
					if(xOffBatTimeStart == 0)
						xOffBatTimeStart = xLastWakeTime;
				}
				else
					xOffBatTimeStart = 0;
			}
		}

		vTaskDelayUntil(&xLastWakeTime, 10/portTICK_RATE_MS);
	}
}

void PowerEnable5VSupply(uint8_t enable)
{
	if(enable)
		GPIOD->BRR = POWER_5V_ENABLE;
	else
		GPIOD->BSRR = POWER_5V_ENABLE;
}

uint8_t PowerIs5VSupplyEnabled()
{
	if(GPIOD->ODR & POWER_5V_ENABLE)
		return 0;
	else
		return 1;
}
