/*
 * fifo8_synced.c
 *
 *  Created on: 25.05.2011
 *      Author: AndreR
 */

#include "fifo8_synced.h"
#include "errors.h"

// Atomic lock and critical section stuff
typedef struct __AtomicLockInfo
{
	uint8_t restoreState;
	uint8_t run;
	FifoSynced* pFifo;
} _AtomicLockInfo;

static __inline__ void __iRelease(_AtomicLockInfo* pLock)
{
	(*pLock->pFifo->releaseCallback)(pLock->restoreState, pLock->pFifo->syncData);

	__asm__ volatile("" ::: "memory");
}

static __inline__ _AtomicLockInfo __iLock(FifoSynced* pFifo)
{
	_AtomicLockInfo lock;
	lock.run = 1;
	lock.pFifo = pFifo;
	lock.restoreState = (*pFifo->lockCallback)(pFifo->syncData);

	return lock;
}

#define FIFO_CRITICAL_SECTION(pFifo) \
for(_AtomicLockInfo __info __attribute((__cleanup__(__iRelease))) = __iLock(pFifo); __info.run; __info.run = 0)


void FifoSyncedInit(FifoSynced* pFifo, uint8_t* pData, uint16_t Size, FifoSyncedLockCallback lockCb, FifoSyncedReleaseCallback releaseCb)
{
	pFifo->pData = pData;
	pFifo->Size = Size;
	pFifo->pDataEnd = (pFifo->pData + pFifo->Size - 1);

	pFifo->syncData = 0;
	pFifo->lockCallback = lockCb;
	pFifo->releaseCallback = releaseCb;

	FifoSyncedClear(pFifo);
}

void FifoSyncedClear(FifoSynced* pFifo)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		pFifo->pWrite = pFifo->pData;
		pFifo->pRead = pFifo->pData;
		pFifo->EmptyBytes = pFifo->Size;
	}
}

int16_t FifoSyncedPushByte(FifoSynced* pFifo, uint8_t Byte)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->EmptyBytes == 0)
			return ERROR_FIFO_SYNCED_NOT_ENOUGH_MEMORY;

		*pFifo->pWrite = Byte;
		if(pFifo->pWrite == pFifo->pDataEnd)
			pFifo->pWrite = pFifo->pData;
		else
			pFifo->pWrite++;

		pFifo->EmptyBytes--;
	}

	return 0;
}

int16_t FifoSyncedPopByte(FifoSynced* pFifo, uint8_t* pByte)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->EmptyBytes == pFifo->Size)
			return ERROR_FIFO_SYNCED_EMPTY;

		*pByte = *pFifo->pRead;

		if(pFifo->pRead == pFifo->pDataEnd)
			pFifo->pRead = pFifo->pData;
		else
			pFifo->pRead++;

		pFifo->EmptyBytes++;
	}

	return 0;
}

int16_t	FifoSyncedPush(FifoSynced* pFifo, uint16_t length, uint8_t* pData)
{
	int16_t result = 0;

	for(uint16_t i = 0; i < length; i++)
	{
		result = FifoSyncedPushByte(pFifo, pData[i]);
		if(result)
			break;
	}

	return result;
}

int16_t	FifoSyncedPop(FifoSynced* pFifo, uint16_t length, uint8_t* pData, uint16_t* pBytesRead)
{
	int16_t result = 0;
	uint16_t i;

	for(i = 0; i < length; i++)
	{
		result = FifoSyncedPopByte(pFifo, pData+i);
		if(result)
			break;
	}

	*pBytesRead = i;

	return result;
}
