package edu.dhbw.mannheim.tigers.sim.engine;


import org.apache.log4j.Logger;

import com.jme.app.AbstractGame;
import com.jme.app.FixedFramerateGame;
import com.jme.input.InputSystem;
import com.jme.util.NanoTimer;
import com.jme.util.Timer;

import edu.dhbw.mannheim.tigers.sim.util.Dormouse;
import edu.dhbw.mannheim.tigers.sim.util.ThreadUtil;


/**
 * An implementation of AbstractGame which controls the number of game-cycles per second.
 * This is necessary because an uncontrolled AbstractGame tends to consume 100% of the cpu cycles.
 * Idea from and veeery similar to {@link FixedFramerateGame} ;-)
 * 
 * @author Gero
 */
public abstract class ACycleControlledEngine extends AbstractGame {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger				log					= Logger.getLogger(ACycleControlledEngine.class);
	

	// Cycle-rate managing stuff
	protected final Timer			timer					= new NanoTimer();
	private long						cycles				= 0;
	
	private long						startTime;
	private long						preferredTicksPerCycle;
	private long						cycleStartTick;
	private long						cycleDurationTicks;
	

	/** {@link Dormouse} */
	private Thread						dormouse;
	
	/** Default cycles per second */
	public static final int			DEFAULT_CPS			= 500;
	public static final int			DEFAULT_FPS			= 30;
	
	/** Boolean value to check whether last cycle has been completed (so that cleanup may start) */
	private boolean					lastCycleComplete	= false;
	

	// CPS/FPS managing stuff
	private Object						frameSync			= new Object();
	
	private int							cyclesPerSecond	= 0;
	private int							framesPerSecond	= 0;
	
	/** Indicates whether the scene is supposed to be rendered this cycle */
	private boolean					isRenderCycle		= true;

	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	public ACycleControlledEngine() {
		// Make sure settings are loaded
		getAttributes();
	}
	

	// --------------------------------------------------------------------------
	// --- simulation cycle -----------------------------------------------------
	// --------------------------------------------------------------------------
	float dump = 0;
	long cycleEnd = 0;
	/**
	 * <code>startCycle</code> begin monitoring the current cycle. This method
	 * should be called every cycle before update and drawing code.
	 */
	private void startCycle() {
		// Timer
		timer.update();
		
		// Cycle-time control
		cycleStartTick = timer.getTime();
		
		
		// Frame-Cycle separation
		dump += framesPerSecond;
		
		if (dump >= cyclesPerSecond) {
			dump = dump % cyclesPerSecond;
			
			isRenderCycle = true;
		} else {
			isRenderCycle = false;
		}
//		long diff = cycleStartTick - cycleEnd;
//		log.debug("START CYCLE (" + diff + ", " + isRenderCycle + ", " + getCurrentCycleRate()
//				+ ")");
	}
	
	
	/**
	 * <code>endCycle</code> ends the current cycle. Pads any excess time in
	 * the cycle by sleep()-ing the thread in order to maintain the desired
	 * cycle rate. No attempt is made to rectify cycles which have taken too
	 * much time.
	 */
	private void endCycle() {
		cycles++;
		cycleDurationTicks = timer.getTime() - cycleStartTick;
		while (cycleDurationTicks < preferredTicksPerCycle) {
			long sleepNanos = preferredTicksPerCycle - cycleDurationTicks;
			
//			log.debug("SLEEP START (" + sleepTimeMillis + "ms)");
			
			if (sleepNanos > 0) {
				ThreadUtil.parkNanosSafe(sleepNanos);
			}
			
//			log.debug("SLEEP END");
			
			cycleDurationTicks = timer.getTime() - cycleStartTick;
		}
	}
	
	
	/**
	 * Render and update logic at a specified fixed rate.
	 * <b>!!!Must call super method if overridden!!!</b>
	 */
	public void start() {
		log.info("Simulation started.");
		
		try {
			initSystem();
			
			assertDisplayCreated();
			
			// Prepare time measurement
			timer.reset();			//Timer.getTimer();
			dormouse = new Thread(Dormouse.getInstance());
			dormouse.start();
			
			// Assure that CPS and FPS are set
			if (getCyclesPerSecond() == 0) {
				setCyclesPerSecond(DEFAULT_CPS);
			}
			if (getFramesPerSecond() == 0) {
				setFramesPerSecond(DEFAULT_FPS);
			}
			
			
			initGame();
			
			// main loop
			while (!finished && !display.isClosing()) {
				
				startCycle();
				
				// handle input events prior to updating the scene
				// - some applications may want to put this into update of the game state
				InputSystem.update();
				
				// update game state, do not use interpolation parameter
				update(-1.0f);
				
				// render only if cycle-to-frame-rate says so! (Actual separation of frames and cycles!)
				if (isRenderCycle) {	
					// render, do not use interpolation parameter
					render(-1.0f);
					
					// swap buffers
					display.getRenderer().displayBackBuffer();
					
					// Give any implementations the chance to do something after the renderer is finished
					afterRenderCallback(-1.0f);
				}
				
				cycleEnd = timer.getTime();
//				log.debug("END CYCLE");
				endCycle();
//				log.debug("END CYCLE 2nd");
				Thread.yield();
			}
		} catch (Throwable t) {
			log.error("Exception in game loop", t);
			handleError(t);
		} finally {
			lastCycleComplete = true;
			
			log.info("Closing Simulation...");
			
			if (dormouse != null) {
				dormouse.interrupt();
			}
			
			quit();
			
			log.info("Simulation closed.");
		}
	}
	

	/**
	 * Called after the renderer has done its job
	 * 
	 * @param interpolation
	 */
	protected abstract void afterRenderCallback(float interpolation);


	/**
	 * @param interpolation
	 *           unused in this implementation
	 * @see AbstractGame#update(float interpolation)
	 */
	protected abstract void update(float interpolation);
	

	/**
	 * @param interpolation
	 *           unused in this implementation
	 * @see AbstractGame#render(float interpolation)
	 */
	protected abstract void render(float interpolation);
	

	/**
	 * @see AbstractGame#initSystem()
	 */
	protected abstract void initSystem();
	

	/**
	 * @see AbstractGame#initGame()
	 */
	protected abstract void initGame();
	

	/**
	 * @see AbstractGame#reinit()
	 */
	protected abstract void reinit();
	

	/**
	 * @see AbstractGame#cleanup()
	 */
	protected abstract void cleanup();
	

	protected abstract void handleError(Throwable t);
	

	protected boolean isLastCycleComplete() {
		return lastCycleComplete;
	}

	
	// --------------------------------------------------------------------------
	// --- frame/cycle rate-control ---------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Set preferred cycle rate. The main loop will make every attempt to
	 * maintain the given cycle rate. This should not be called prior to the
	 * application being <code>start()</code> -ed.
	 * 
	 * @param newCPS
	 *           the desired cycle rate in cycles per second
	 */
	public final void setCyclesPerSecond(int newCPS) {
		
		synchronized (frameSync) {
			if (newCPS < framesPerSecond) {
				cyclesPerSecond = framesPerSecond;
				log.info("Cycle rate must be >= frame rate, set it to " + cyclesPerSecond);
			} else {
				cyclesPerSecond = newCPS;
				
				if (newCPS <= 0) {
					throw new IllegalArgumentException("Cycles per second cannot be less than one.");
				}
				
				long newPreferredTicksPC = timer.getResolution() / newCPS;
				if (preferredTicksPerCycle != newPreferredTicksPC) {
					preferredTicksPerCycle = newPreferredTicksPC;
					log.info("Attempting to run at " + newCPS + " cycles/second.");
				}
			}
			// recalculateCyclesPerFrame();
		}
	}
	

	public final void setFramesPerSecond(int newFrameRate) {
		
		synchronized (frameSync) {
			if (newFrameRate > cyclesPerSecond) {
				framesPerSecond = cyclesPerSecond;
				log.info("Frame rate must be <= cycle rate, set it to " + cyclesPerSecond);
			} else {
				framesPerSecond = newFrameRate;
				log.info("Attempting to render " + framesPerSecond + " frames/second.");
			}
			// recalculateCyclesPerFrame();
		}
	}
	

	// private void recalculateCyclesPerFrame() {
	// synchronized (frameSync) {
	// cyclesPerFrame = cyclesPerSecond / framesPerSecond;
	// }
	// }
	

	/**
	 * Gets the current cycle rate.
	 * 
	 * @return the current number of cycles rendering per second
	 */
	public final float getCurrentCycleRate() {
		float time = (timer.getTime() - startTime) / (float) timer.getResolution();
		float cps = cycles / time;
		startTime = timer.getTime();
		cycles = 0;
		return cps;
	}
	
	
	public int getFramesPerSecond() {
		
		return framesPerSecond;
	}
	

	public int getCyclesPerSecond() {
		return cyclesPerSecond;
	}
	
	
	/**
	 * @return	Whether the scene is supposed to be rendered this cycle
	 */
	public boolean isRenderCycle() {
		return isRenderCycle;
	}

	
	public long getCurrentCycle() {
		return cycles;
	}
}
