package edu.dhbw.mannheim.tigers.sim.engine.com.bot;


import java.io.IOException;
import java.net.SocketException;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.engine.com.IStartableInterface;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.BotID;


/**
 * This is the main communication structure. It is responsible for its two parts, the receiving and
 * transmitting, and handles the data-flow between them and the simulation.
 * 
 * These three classes are also referred to as "botcom-trio".
 * 
 * @see ABotCommandReceiver
 * @see ABotStatusTransmitter
 * 
 * @author Gero, Malte
 */
public abstract class ABotCommunicator<C, S> implements IStartableInterface, ICommunicator<C>, Runnable {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger						log						= Logger.getLogger(getClass());
	
	/** [ms] */
	private static final int				STOP_TIMEOUT			= 500;
	protected final IBotController<C>	botController;
	
	private final Object						sleep						= new Object();
	private volatile boolean				closeConn				= false;
	/** [ms] */
	private final int							timeout;
	private Thread								thread;
	private final String						name;
	private final String						id;
	
	/** Whether this comm is in its shutdown progress */
	private volatile boolean				shutdownInProgress	= false;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see ABotCommunicator
	 * @param botController
	 * @param timeout
	 */
	public ABotCommunicator(IBotController<C> botController, int timeout) {
		super();
		
		this.botController = botController;
		this.id = botController.getID().toString();
		this.name = "Comm " + id;
		this.timeout = timeout;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public synchronized void start() {
		thread = new Thread(this, name);
		thread.start();
	}
	

	@Override
	public void run() {
		while (!Thread.currentThread().isInterrupted() && !shutdownInProgress) {
			try {
				closeConn = false;
				
				boolean startupWorked = startTransceiving();
				if (!startupWorked) {
					break; // Shutdown in progress, exit...
				}
				
				botController.onConnect();
				
				// Maintain state: Sleep regularly until on (or both) of them got killed and calls notifyShutdown() or comm
				// is shut down...
				while (!shutdownInProgress && !closeConn) {
					
					// Check; allow sub-classes to use timeout
					if (!checkConnection()) {
						break;
					}
					
					// Control timeout
					synchronized (sleep) {
						sleep.wait(timeout);
					}
				}
				
				halt(); // ...then kill both, and start again! =)
				
			} catch (SocketException err) {
				if (shutdownInProgress) {
					log.debug(id + " interrupted...");
				} else {
					log.error(err.getMessage(), err);
				}
				break;
			} catch (IOException err) {
				log.error(id + " ServerSocket error!", err);
				break;
			} catch (InterruptedException err) {
				if (!shutdownInProgress)
				{
					log.error(id + " Error while waiting on start/stop of Transmitter/Receiver!", err);
				}
				break;
			}
		}
		
		cleanup();
	}


	@Override
	public void newCommand(C cmd) {
		botController.processNewCommand(cmd);
	}
	
	
	/**
	 * Sends the given message to the connected bot
	 * 
	 * @param msg
	 * @throws IOException 
	 */
	public abstract void send(S msg) throws IOException;
	

	public synchronized void setStatusInterval(int newInterval) {
		IBotCommandTransmitter<S> trans = getTransmitter();
		if (trans != null) {
			trans.setStatusMessageInterval(newInterval);
		}
	}
	

	private void halt() {
		stopTransceiving();
		
		botController.onDisconnect();
	}
	

	private void cleanup() {
		halt();
		deinitialize();
		closeConn = false;
	}
	

	@Override
	public synchronized void stop() {
		shutdownInProgress = true;
		
		cleanup();
		
		if (thread != null) {
			thread.interrupt();
			notifyShutdown();
			
			try {
				thread.join(STOP_TIMEOUT);
			} catch (InterruptedException err) {
				log.debug("Interrupted while waiting for shutdown...");
			} finally {
				thread = null;
			}
		}
		
		shutdownInProgress = false;
	}
	

	@Override
	public void notifyShutdown() {
		synchronized (sleep) {
			closeConn = true;
			sleep.notifyAll();
		}
	}
	

	/**
	 * This got called every {@link #timeout} ms (approx.)
	 * 
	 * @return Whether the connection is still existing
	 */
	protected boolean checkConnection() {
		return true;
	}
	
	
	/**
	 * @return The timeout for {@link #checkConnection()} [ms]
	 */
	protected int getTimeout() {
		return timeout;
	}
	

	/**
	 * Called in the constructor, sets up the socket, e.g.
	 */
	protected abstract void initialize();
	

	/**
	 * Start both parts: the receiving and the transmitting
	 * 
	 * @param in
	 * @param out
	 * @return If starting the comm has been successful
	 * 
	 * @throws IOException
	 * @throws InterruptedException If it is waiting for incoming connection and interrupted from outside
	 */
	protected abstract boolean startTransceiving() throws IOException, InterruptedException;
	

	/**
	 * Stops both communicating parts
	 */
	protected abstract void stopTransceiving();
	

	/**
	 * Called when the {@link ICommunicator} is {@link ICommunicator#stop()}ed
	 */
	protected abstract void deinitialize();
	

	/**
	 * @return The {@link IBotCommandTransmitter}-implementation used
	 */
	protected abstract IBotCommandTransmitter<S> getTransmitter();
	

	@Override
	public BotID getBotId() {
		return botController.getID();
	}
}
