package edu.dhbw.mannheim.tigers.sim.engine.com.bot.tiger;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.NetworkInterface;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.engine.com.bot.ABotCommunicator;
import edu.dhbw.mannheim.tigers.sim.settings.data.ITeamProperties;
import edu.dhbw.mannheim.tigers.sim.util.network.MulticastUDPTransceiver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.CommandConstants;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerSystemAnnouncement;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerSystemSetIdentity;


/**
 * This class handles the multicast-messaging the TigerBot is capable of. This includes the commands:
 * <ul>
 * <li>{@link TigerSystemSetIdentity} - Sent by the CS to set the bots ports, SSL-ID etc..</li>
 * <li>{@link TigerSystemAnnouncement} - Sent by the bot to notify its there (once on startup)</li>
 * </ul>
 * Unlike the unicast-messaging it is running as long as the TigerBot is online. This special behavior is responsible
 * for a few more methods in the {@link TigerBotCommunicator} then specified in {@link ABotCommunicator}, which does not
 * cover this behavior.
 * 
 * @author Gero
 */
public class TigerBotMulticastTransceiver implements Runnable {
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger							log						= Logger.getLogger(getClass());
	
	/** [B] */
	private static final int					BUFFER_SIZE				= 1500;
	/** [ms] */
	private static final int					STOP_TIMEOUT			= 500;
	
	private final String							name;
	private final MulticastUDPTransceiver	transceiver;
	private Thread									thread;
	private volatile boolean					shutdownInProgress	= false;
	
	private volatile long						lastPacketReceived = 0;
	
	private final TigerCommandParser			parser;
	
	private final TigerBotController			controller;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see TigerBotMulticastTransceiver
	 * @param controller
	 * @param pairs
	 */
	public TigerBotMulticastTransceiver(TigerBotController controller, NetworkInterface nif) {
		this.controller = controller;
		this.name = "McastTxRx " + controller.getID();
		

		// Get connect params
		final ITeamProperties props = controller.getProperties();
		
		transceiver = new MulticastUDPTransceiver(props.getMcastListenPort(), props.getMcastDiscoverGroup(),
				props.getMcastServerPort(), nif);
		

		parser = new TigerCommandParser();
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// -------------------------------------------------------------------------
	/**
	 * Starts the thread which is listening on the multicast socket.
	 */
	public synchronized void start() {
		thread = new Thread(this, name);
		thread.start();
		
		log.info("Initialized.");
	}
	

	@Override
	public void run() {
		while (!shutdownInProgress) { // No Thread.currentThread().isInterrupted(), as this won't receive interrupt()
												// signals while waiting for new data in receive(...)
			
			// Wait for incoming packets...
			byte[] buffer = new byte[BUFFER_SIZE];
			DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
			
			try {
				packet = transceiver.receive(packet);
				if (packet == null) {
					break; // This shows us that we have an error: Jump off!
				}
				
				lastPacketReceived = System.nanoTime();
//				System.out.println("TxRx:" + lastPacketReceived);
				
			} catch (IOException err) {
				if (!shutdownInProgress) {
					log.error("Error while receiving packet!");
				}
				break;
			}
			
			// Interpret it:
			ACommand cmd = parser.readCommand(buffer);
			if (cmd == null) {
				log.debug("Unable to parse message!"); // : " + Arrays.toString(buffer));
				continue;
			}
			
			switch (cmd.getCommand()) {
				
				case CommandConstants.CMD_MULTICAST_UPDATE_ALL_V2:
					controller.processNewCommand(cmd);	// Pass incoming multicast
					break;
					
				
				case CommandConstants.CMD_SYSTEM_SET_IDENTITY:
					TigerSystemSetIdentity newIdentity = (TigerSystemSetIdentity) cmd;
					
					if (newIdentity.getBotId() == TigerBotController.DISCONNECT_BOT_ID) {
						controller.processDisconnect(newIdentity);
					} else {
						controller.passNewIdentity(new TigerIdentityInfo(packet.getAddress(), newIdentity));
					}
					break;
				
				default:
					log.warn("Received a message of type '" + cmd.getCommand() + "' per multicast, unexpected and dropped!");
					break;
			}
		}
		
		cleanup();
		
		log.info("Shut down.");
	}
	
	
	long getLastPacketReceived() {
		return lastPacketReceived;
	}
	
	
	void resetLastPacketReceived() {
		lastPacketReceived = 0;
	}
	

	private void cleanup() {
		if (transceiver.isReady()) {
			try {
				transceiver.cleanup();
			} catch (IOException err) {
				log.debug("An error occurred while closing:" + err.getMessage());
			}
		}
		
		resetLastPacketReceived();
	}
	

	public synchronized void sendAnnouncement() {
		if (transceiver.isReady()) {
			final String cpuId = controller.getCpuId();
			
			TigerSystemAnnouncement announcement = new TigerSystemAnnouncement();
			announcement.setCpuId(cpuId);
			
			transceiver.send(announcement.getTransferData());
//			log.debug("Sent announcement with CPU-ID '" + cpuId + "'...");
		} else {
			log.debug("Transceiver not ready to send announcement!");
		}
	}
	

	/**
	 * Stops the listening-thread and closes the multicast socket
	 */
	public synchronized void stop() {
		shutdownInProgress = true;
		
		if (thread != null) {
			cleanup();
			
			try {
				thread.join(STOP_TIMEOUT); // Avoid race-conditions if restarting the communicator...
			} catch (InterruptedException err) {
				log.debug("Interrupted while waiting for TigerBotTxRx to shut down...");
			}
			thread = null;
		}
		
		shutdownInProgress = false;
	}
}
