package edu.dhbw.mannheim.tigers.sim.engine.com.cam;


import java.net.BindException;
import java.net.NetworkInterface;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.locks.LockSupport;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sim.engine.com.ERoboCupTeam;
import edu.dhbw.mannheim.tigers.sim.engine.com.IRoboCupControlProvider;
import edu.dhbw.mannheim.tigers.sim.engine.com.IStartableInterface;
import edu.dhbw.mannheim.tigers.sim.engine.model.TimedModelLocation;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.BotID;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.IBasicBotModelControl;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.observer.ABotModelObserver;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.observer.BotLocation;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.observer.IBotModelObservable;
import edu.dhbw.mannheim.tigers.sim.engine.model.observer.AModelStateObserver;
import edu.dhbw.mannheim.tigers.sim.engine.model.observer.IModelStateObservable;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.EPluginType;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.IPluginProvider;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.PluginID;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.cam.ACamOutputPlugin;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.ICamProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ISimProperties;
import edu.dhbw.mannheim.tigers.sim.util.ThreadUtil;
import edu.dhbw.mannheim.tigers.sim.util.network.ITransmitter;
import edu.dhbw.mannheim.tigers.sim.util.network.MulticastUDPTransmitter;
import edu.dhbw.mannheim.tigers.sim.util.network.NetworkUtility;
import edu.dhbw.mannheim.tigers.sim.util.network.UnicastUDPTransmitter;
import edu.dhbw.mannheim.tigers.sumatra.model.data.MessagesRobocupSslWrapper.SSL_WrapperPacket;


/**
 * This class provides information about the current positions of the models to the CentralSoftware
 * via multicast. On the one hand it decouples the information-objects from the simulation. On the
 * other hand it passes them to the chosen builder ({@link ACamOutputPlugin}), which creates a data
 * structure that is finally sent.
 * 
 * @author Gero
 */
public class SimCamInputProvider implements IStartableInterface, Runnable {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger											log							= Logger.getLogger(getClass());
	
	/**
	 * Thread that sends the packages to the specified multicast address
	 */
	private Thread													cam							= null;
	private final AtomicBoolean								cancelSwitch				= new AtomicBoolean(false);
	
	// Control over the simulation
	private final IRoboCupControlProvider					controlProvider;
	

	// Model observer
	private AModelStateObserver								ballObserver				= null;
	private ABotModelObserver									blueBotObserver			= null;
	private ABotModelObserver									yellowBotObserver			= null;
	

	// Buffer
	private final HashMap<Integer, TimedModelLocation>	blueBotBuffer				= new HashMap<Integer, TimedModelLocation>();
	private final ReentrantLock								blueBotLock					= new ReentrantLock();
	private final HashMap<Integer, TimedModelLocation>	yellowBotBuffer			= new HashMap<Integer, TimedModelLocation>();
	private final ReentrantLock								yellowBotLock				= new ReentrantLock();
	private TimedModelLocation									ballBuffer					= TimedModelLocation.EMPTY_MODEL_LOCATION;
	private final ReentrantLock								ballLock						= new ReentrantLock();
	
	/** Temporary, from simulation completely separate HashMap-variable, used for frame-building */
	private final HashMap<Integer, TimedModelLocation>	blueBotBuilderBuffer		= new HashMap<Integer, TimedModelLocation>();
	
	/** Temporary, from simulation completely separate HashMap-variable, used for frame-building */
	private final HashMap<Integer, TimedModelLocation>	yellowBotBuilderBuffer	= new HashMap<Integer, TimedModelLocation>();
	
	/** Temporary, from simulation completely separate variable, used for frame-building */
	private TimedModelLocation									ballBuilderBuffer;
	

	// Settings
	private final ICamProperties								props;
	private boolean												multicastMode				= false;
	

	// Communication
	private final NetworkUtility								networkUtil					= new NetworkUtility();
	private ITransmitter<byte[]>								transmitter					= null;
	private boolean												blocked						= false;
	
	// Cam
	private ACamOutputPlugin									frameBuilder				= null;
	private final long											packetCount					= 0;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see SimCamInputProvider
	 * @param newControlProvider
	 */
	public SimCamInputProvider(IRoboCupControlProvider newControlProvider) {
		
		this.controlProvider = newControlProvider;
		
		// Load properties
		ISimProperties simProps = ISettingsProvider.INSTANCE.getSimProperties();
		multicastMode = simProps.isMulticastMode();
		props = simProps.getCamProperties();
		
		// ACamOutputPlugin
		PluginID pluginID = new PluginID(props.getPluginID());
		ACamOutputPlugin plugin = IPluginProvider.INSTANCE.getPlugin(pluginID, EPluginType.CAM_OUTPUT);
		props.setPluginID(plugin.getID().base);
		
		frameBuilder = plugin;
		log.info("Using plugin " + plugin + ".");
		
		
		// Initialize cam-plugin
		frameBuilder.initialize(packetCount, newControlProvider.getFieldData());
		

		// Register ballObserver
		ballObserver = new AModelStateObserver(null, false, "BallObserver") {
			
			@Override
			protected void onUpdate(IModelStateObservable observable, TimedModelLocation event) {
				try {
					ballLock.lock();
					ballBuffer = event;
				} finally {
					ballLock.unlock();
				}
			}
		};
		
		controlProvider.getBall().addObservers(ballObserver);
		

		// Register blue botObserver
		blueBotObserver = new ABotModelObserver(null, false, "BotObserver") {
			
			@Override
			protected void onUpdate(IBotModelObservable observable, BotLocation event) {
				
				try {
					blueBotLock.lock();
					
					// Dispatch
					TimedModelLocation loc = blueBotBuffer.get(event.getId());
					if (loc == null) {
						blueBotBuffer.put(event.getId(), loc = new TimedModelLocation(event));
					} else {
						loc.set(event);
					}
					
				} finally {
					blueBotLock.unlock();
				}
			}
		};
		
		for (BotID botID : controlProvider.getTeamIDs(ERoboCupTeam.BLUE_TEAM)) {
			IBasicBotModelControl bot = controlProvider.getBot(botID);
			bot.addObservers(blueBotObserver);
		}
		

		// Register yellow botObserver
		yellowBotObserver = new ABotModelObserver(null, false, "BotObserver") {
			
			@Override
			protected void onUpdate(IBotModelObservable observable, BotLocation event) {
				
				try {
					yellowBotLock.lock();
					
					// Dispatch
					TimedModelLocation loc = yellowBotBuffer.get(event.getId());
					if (loc == null) {
						yellowBotBuffer.put(event.getId(), loc = new TimedModelLocation(event));
					} else {
						loc.set(event);
					}
					
				} finally {
					yellowBotLock.unlock();
				}
			}
		};
		
		for (BotID botID : controlProvider.getTeamIDs(ERoboCupTeam.YELLOW_TEAM)) {
			IBasicBotModelControl bot = controlProvider.getBot(botID);
			bot.addObservers(yellowBotObserver);
		}
		
		
		// Choose network interface
		final String ifaceAddr = props.getMcastInterface();
		final NetworkInterface nif = networkUtil.chooseNetworkInterface(ifaceAddr, 3);
		if (nif == null)
		{
			log.error("No proper nif for cam-output in network '" + ifaceAddr + "' found!");
		} else
		{
			log.info("Chose nif for cam-output: " + nif.getDisplayName() + ".");
		}
		

		// Prepare communication
		if (multicastMode) {
			transmitter = new MulticastUDPTransmitter(props.getLocalPort(), props.getMcastAddress(), props.getTargetPort(), nif);
		} else {
			try {
				transmitter = new UnicastUDPTransmitter(props.getLocalPort(), props.getMcastAddress(), props.getTargetPort());
			} catch (BindException err) {
				log.error("Error establishing unicast UDP transmitter: ", err);
			}
		}
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void start() {
		cam = new Thread(this, "SimCam");
		cam.start();
		log.info("Started SimCamInputProvider...");
	}
	

	@Override
	public void run() {
		
		final long totalSleepNanos = TimeUnit.MILLISECONDS.toNanos(1000 / props.getFPS());
		
		while (!Thread.currentThread().isInterrupted()) {
			
			if (!blocked) {
				List<SSL_WrapperPacket> packets = buildCamFrame();
				
				for (SSL_WrapperPacket packet : packets) {
					// log.debug(packet);
					transmitter.send(packet.toByteArray());
				}
			}
			
			cancelSwitch.set(false);
			ThreadUtil.parkNanosSafe(totalSleepNanos, cancelSwitch);
		}
		
		cleanup();
		log.info("Stopped SimCamInputProvider.");
	}
	
	
	private void cleanup() {
		// Clear buffers
		if (blueBotBuffer != null) {
			try {
				blueBotLock.lock();
				blueBotBuffer.clear();
			} finally {
				blueBotLock.unlock();
			}
		}
		
		if (yellowBotBuffer != null) {
			try {
				yellowBotLock.lock();
				yellowBotBuffer.clear();
			} finally {
				yellowBotLock.unlock();
			}
		}
		
		if (blueBotBuilderBuffer != null) {
			blueBotBuilderBuffer.clear();
		}
		
		if (yellowBotBuilderBuffer != null) {
			yellowBotBuilderBuffer.clear();
		}
		
		// Unregister observers...
		for (BotID id : controlProvider.getTeamIDs(ERoboCupTeam.YELLOW_TEAM)) {
			IBasicBotModelControl bot = controlProvider.getBot(id);
			bot.removeObserver(yellowBotObserver);
		}
		
		for (BotID id : controlProvider.getTeamIDs(ERoboCupTeam.BLUE_TEAM)) {
			IBasicBotModelControl bot = controlProvider.getBot(id);
			bot.removeObserver(blueBotObserver);
		}
		
		controlProvider.getBall().removeObserver(ballObserver);
		

		// ... and close them.
		yellowBotObserver.shutdown();
		blueBotObserver.shutdown();
		ballObserver.shutdown();
	}
	

	@Override
	public void stop() {
		// Close thread
		if (cam != null) {
			cam.interrupt();
			cancelSwitch.set(true);
			LockSupport.unpark(cam);
		
			try {
				cam.join();
			} catch (InterruptedException err) {
				log.error("Error while waiting for cam to stop!", err);
			}
			
			cam = null;
		}
	}
	

	private List<SSL_WrapperPacket> buildCamFrame() {
		
		// Dereferencing old objects
		ballBuilderBuffer = null;
		blueBotBuilderBuffer.clear();
		yellowBotBuilderBuffer.clear();
		
		try {
			ballLock.lock();
			ballBuilderBuffer = ballBuffer.clone();
		} finally {
			ballLock.unlock();
		}
		
		try {
			blueBotLock.lock();
			blueBotBuilderBuffer.putAll(blueBotBuffer);
		} finally {
			blueBotLock.unlock();
		}
		
		try {
			yellowBotLock.lock();
			yellowBotBuilderBuffer.putAll(yellowBotBuffer);
		} finally {
			yellowBotLock.unlock();
		}
		
		return frameBuilder.buildCamInputs(ballBuilderBuffer, blueBotBuilderBuffer, yellowBotBuilderBuffer);
	}
	

	public boolean isBlocked() {
		return blocked;
	}
	

	public void setBlocked(boolean blocked) {
		this.blocked = blocked;
	}
}