package edu.dhbw.mannheim.tigers.sim.engine.delegates;

import java.awt.Point;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.PixelInterleavedSampleModel;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.lwjgl.BufferUtils;

import com.bric.qt.io.JPEGMovieAnimation;
import com.jme.image.Image;
import com.jme.input.InputHandler;
import com.jme.input.KeyInput;
import com.jme.input.action.InputAction;
import com.jme.input.action.InputActionEvent;
import com.jme.renderer.Renderer;

import edu.dhbw.mannheim.tigers.sim.Sim;
import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.ISimEngineProperties;


/**
 * This {@link ASimDelegate} captures videos between two hits in 'F8'
 * 
 * @author Gero
 */
public class VideoCaptureDelegate extends ASimDelegate {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// private static final VideoFormat VIDEO_FORMAT = VideoFormat.RAW;
	private static final String			PATH					= Sim.VIDEOS_PATH;
	private static final String			FILE_EXTENSION		= ".mov";
	// private static final float COMPRESSION_QUALITY = 0.5f;
	
	/** Whether a video should be produced or not */
	private boolean							captureState		= false;
	/** Helper-variable for detection of {@link #captureState}-switch */
	private boolean							lastCaptureState	= false;
	
	// /** The buffer for captured frames! */
	private int									width					= -1;
	private int									height				= -1;
	private File								outFile				= null;
	private JPEGMovieAnimation				movOut				= null;
	
	private final ISimEngineProperties	props;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see VideoCaptureDelegate
	 * @param core
	 */
	public VideoCaptureDelegate(ICore core) {
		super(core);
		
		props = ISettingsProvider.INSTANCE.getSimProperties().getSimEngineProperties();
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void initSimulation() {
		core.getInputHandler().addAction(new CaptureAction(), InputHandler.DEVICE_KEYBOARD, KeyInput.KEY_F8,
				InputHandler.AXIS_NONE, false);
	}
	

	@Override
	public void afterRenderCallback(float interpolation, Renderer renderer) {
		
		// ### If captureState changed, handle it
		if (captureState != lastCaptureState) {
			if (captureState) {
				
				// ### Start capture
				startCapture(renderer);
			} else {
				
				// ### Already capturing, stop it!
				stopCapture();
			}
			
			lastCaptureState = captureState;
		}
		

		// ### If capture is ON: Do it!
		if (captureState) {
			// ### Grab content from renderer's buffer
			final ByteBuffer dBuffer = BufferUtils.createByteBuffer(width * height * 3);
			renderer.grabScreenContents(dBuffer, Image.Format.RGB8, 0, 0, width, height);
			
			// As the DirectByteBuffer needed for OpenGL is not backed by an Array we need to copy its data.
			// And not just that: Our image is upside down, too. And as PixelInterLeavedSampleModel can not be convinced to
			// access its model backwards, we have to do reverse it manually.
			byte[] arrBuff = new byte[dBuffer.capacity()];
			for (int y = height; y > 0; y--) { // Reverse image, ...
				int index = ((y - 1) * width) * 3; // ... starting at the beginning of every row, ...
				dBuffer.get(arrBuff, index, width * 3); // ...copying it line by line!
			}
			

			// ### Create a BufferedImage
			// The advantage of having the data in an array is, that we can directly use if for out
			// BufferedImage-representation. For that we need a DataBufferByte with our array in it, and a SampleModel,
			// which holds the information how to access this DataBuffer.
			DataBufferByte dbb = new DataBufferByte(arrBuff, arrBuff.length);
			PixelInterleavedSampleModel pilSampleModel = new PixelInterleavedSampleModel(DataBuffer.TYPE_BYTE, width,
					height, 3, 3 * width, new int[] { 0, 1, 2 }); // r, g, b
			WritableRaster wra = WritableRaster.createWritableRaster(pilSampleModel, dbb, new Point(0, 0));
			
			// Then we create the BufferedImage from out Raster and ColorModel
			ColorSpace cs = ColorSpace.getInstance(1000); // default sRGB
			int[] bits = new int[] { 8, 8, 8 }; // 24 bit
			boolean isAlphaPremultiplied = false;
			ColorModel cm = new ComponentColorModel(cs, bits, false, isAlphaPremultiplied, 1, 0);
			BufferedImage img = new BufferedImage(cm, wra, isAlphaPremultiplied, null);
			
			// BufferedImage img = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
			// // Grab each pixel information and set it to the BufferedImage info.
			// for (int x = 0; x < width; x++) {
			// for (int y = 0; y < height; y++) {
			//
			// int index = 3 * ((height - y - 1) * width + x);
			// int argb = (((int) (buffer.get(index + 0)) & 0xFF) << 16) // r
			// | (((int) (buffer.get(index + 1)) & 0xFF) << 8) // g
			// | (((int) (buffer.get(index + 2)) & 0xFF)); // b
			//
			// img.setRGB(x, y, argb);
			// }
			// }
			
			// ### Write it to output
			try {
				float q = props.getVideoQuality();
				if (q > 1 || !(q > 0)) {
					q = 1;
				}
				movOut.addFrame(1f / core.getFramesPerSecond(), img, q);
			} catch (IOException err) {
				log.error("Error while writing BufferedImage to output!", err);
			}
		}
	}
	

	/**
	 * Prepares the delegates internal state for recording
	 * 
	 * @param renderer
	 */
	private void startCapture(Renderer renderer) {
		try {
			// Name
			SimpleDateFormat sd = new SimpleDateFormat("kk-mm-ss");
			String dateString = sd.format(new Date());
			String filename = "video_" + dateString + FILE_EXTENSION;
			
			// File...
			outFile = new File(PATH + filename);
			
			// ... and Stream
			// out = new AVIOutputStream(outFile, VIDEO_FORMAT);
			// out.setVideoCompressionQuality(COMPRESSION_QUALITY);
			// out.setTimeScale(1);
			// out.setFrameRate(30);
			movOut = new JPEGMovieAnimation(outFile);
			
			// Sample size
			height = renderer.getHeight();
			width = renderer.getWidth();
			
			log.info("Started video capture to: " + filename);
			
		} catch (IOException err) {
			log.warn("Unable to open AVIOutputStream for video capture!");
		}
	}
	

	/**
	 * Stops the active capture-process.
	 */
	private void stopCapture() {
		try {
			// if (out != null) {
			// out.close();
			// }
			if (movOut != null) {
				movOut.close();
			}
		} catch (IOException err) {
			log.warn("Error while closing AVIOutputStream!");
		} finally {
			log.info("Finished video capture to: " + outFile.getName());
			
			// out = null;
			movOut = null;
			outFile = null;
			
			width = -1;
			height = -1;
		}
	}
	

	@Override
	public void cleanup() {
		if (captureState) {
			captureState = false;
			
			stopCapture(); // Ensure that the AVIOutputStream is closed
		}
	}
	
	
	class CaptureAction extends InputAction {
		@Override
		public void performAction(InputActionEvent evt) {
			if (!evt.getTriggerPressed()) { // Trigger released
				captureState = !captureState; // Switch state
			}
		}
	}
}
