package edu.dhbw.mannheim.tigers.sim.engine.delegates;


import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.apache.log4j.Logger;

import com.jme.scene.Node;
import com.jme.system.DisplaySystem;

import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.engine.com.visualization.IVisualizationDataInterpreter;
import edu.dhbw.mannheim.tigers.sim.engine.modes.IMatchControlProvider;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.EPluginType;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.IPluginProvider;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.PluginID;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization.AVisualizationPlugin;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization.IDrawer;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization.SimVisMessages.SimVisData;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.IVisualizationProperties;


/**
 * This class accepts the given {@link VisualizationData} and display appropriate information in the
 * simulation. The default {@link IVisualizationDataInterpreter}-implementation.
 * 
 * @author Gero
 */
public class VisualizationDataDelegate extends ASimDelegate implements IVisualizationDataInterpreter {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected final Logger					log			= Logger.getLogger(getClass());
	
	// Data
	private SimVisData						data;
	private final Object						dataSync		= new Object();
	private boolean							changed		= false;
	
	// Visualization
	private DisplaySystem					display;
	private Node								visRoot;
	
	private PluginID							pluginID		= null;
	private AVisualizationPlugin			plugin		= null;
	private final List<IDrawer>			drawers		= new ArrayList<IDrawer>();
	
	private final IMatchControlProvider	matchControl;
	
	/** [ns], {@link System#nanoTime()} */
	private long								lastUpdate	= 0;
	private final long						MAX_HOLD_TIME;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see VisualizationDataDelegate
	 * @param core
	 */
	public VisualizationDataDelegate(ICore core, IMatchControlProvider matchControl) {
		super(core);
		
		this.matchControl = matchControl;
		
		IVisualizationProperties props = ISettingsProvider.INSTANCE.getSimProperties().getVisualizationProperties();
		MAX_HOLD_TIME = TimeUnit.SECONDS.toNanos(props.getMaxHoldTime());
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void receive(SimVisData msg) {
		synchronized (dataSync) {
			data = msg;
			changed = true;
		}
	}
	

	@Override
	public void initSimulation() {
		
		display = core.getDisplay();
		visRoot = core.getVisualizationRoot();
		

		// Get properties
		IVisualizationProperties props = ISettingsProvider.INSTANCE.getSimProperties().getVisualizationProperties();
		pluginID = new PluginID(props.getPluginID());
		
		// Get plugin
		plugin = IPluginProvider.INSTANCE.getPlugin(pluginID, EPluginType.VISUALIZATION);
		props.setPluginID(plugin.getID().base);
		log.info("Using plugin " + plugin + ".");
		
		// Get drawers!
		drawers.clear();
		drawers.addAll(plugin.getPoiDrawers(visRoot, display, matchControl));
	}
	

	@Override
	public void update(float interpolation, float tpc, float cps) {
		synchronized (dataSync) {
			if (!changed) {
				if (lastUpdate < System.nanoTime() - MAX_HOLD_TIME) {
					for (IDrawer drawer : drawers) {
						drawer.undisplayInfo();
					}
				}
				return;
			}
			
			for (IDrawer drawer : drawers) {
				drawer.displayInfo(data);
			}
			
			changed = false;
			lastUpdate = System.nanoTime();
		}
	}
	

	/*
	 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	 * Action
	 */
	// private class SwitchAction extends InputAction {
	//
	// protected final APOIDrawer drawer;
	// private Boolean display = true;
	//
	//
	// public SwitchAction(APOIDrawer drawer) {
	// this(drawer, true);
	// }
	//
	//
	// public SwitchAction(APOIDrawer drawer, boolean displayAtStartup) {
	// this.drawer = drawer;
	// this.display = displayAtStartup;
	// }
	//
	//
	// @Override
	// public void performAction(InputActionEvent evt) {
	//
	// if (!evt.getTriggerPressed()) { // Trigger released
	// if (display) {
	// display = false;
	// drawer.undisplayInfo();
	// } else {
	// display = true;
	// }
	// }
	// }
	//
	// }
}
