package edu.dhbw.mannheim.tigers.sim.engine.model;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.ExtendedMetaData;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.emf.ecore.xmi.impl.XMLResourceImpl;

import com.jmex.physics.material.Material;

import edu.dhbw.mannheim.tigers.sim.Sim;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.IDataFactory;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.IDataPackage;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.IModelDataStorage;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.ball.IBallFactory;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.ball.IBallModelData;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.ball.impl.BallPackageImpl;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.bot.IBotFactory;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.bot.ICTBotModelData;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.bot.ITigerBotModelData;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.bot.impl.BotPackageImpl;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.field.IFieldFactory;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.field.IRoboCupFieldData;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.field.impl.FieldPackageImpl;
import edu.dhbw.mannheim.tigers.sim.resources.AResourceProvider;


/**
 * <p>
 * Provides data sets to size the models. Maybe it'll load it from a file later on...? - In deed, it already does! =D
 * </p>
 * 
 * @author Gero
 */
public class ModelDataProvider extends AResourceProvider<String, IModelDataStorage> implements IModelDataProvider {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	public static final String									MODEL_PATH		= Sim.MODEL_PROPERTIES_PATH;
	

	// Util
	private final IDataFactory									factory			= IDataFactory.eINSTANCE;
	

	// Storages
	private final HashMap<String, IRoboCupFieldData>	fieldData		= new HashMap<String, IRoboCupFieldData>();
	// private HashMap<String, Material> matData = new HashMap<String, Material>();
	private final HashMap<String, IBallModelData>		ballData			= new HashMap<String, IBallModelData>();
	
	private final HashMap<String, ICTBotModelData>		ctBotData		= new HashMap<String, ICTBotModelData>();
	private final HashMap<String, ITigerBotModelData>	tigerBotData	= new HashMap<String, ITigerBotModelData>();
	

	// Stans
	public static final String									STAN_FIELD		= FieldPackageImpl.eINSTANCE
																									.getRoboCupFieldData_Name()
																									.getDefaultValueLiteral();
	// public static final String STAN_MATERIAL = Material.DEFAULT.getName();
	public static final String									STAN_BALL		= BallPackageImpl.eINSTANCE
																									.getBallModelData_Name()
																									.getDefaultValueLiteral();
	
	public static final String									STAN_CT_BOT		= BotPackageImpl.eINSTANCE
																									.getCTBotModelData_Name()
																									.getDefaultValueLiteral();
	public static final String									STAN_TIGER_BOT	= BotPackageImpl.eINSTANCE
																									.getTigerBotModelData_Name()
																									.getDefaultValueLiteral();
	

	private static ModelDataProvider							instance			= null;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	private ModelDataProvider() {
		super(ModelDataProvider.class);
	}
	

	public static synchronized ModelDataProvider getInstance() {
		if (instance == null) {
			instance = new ModelDataProvider();
		}
		return instance;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void doInitialize() {
		@SuppressWarnings("unused")
		IDataPackage pack = IDataPackage.eINSTANCE; // Register package (done as side effect), as this
																	// is no RCP environment
	}
	

	@Override
	protected void doLoad() {
		// Load settings from file using EMF
		Resource resource = new XMLResourceImpl();
		
		Map<String, ExtendedMetaData> options = new HashMap<String, ExtendedMetaData>();
		options.put(XMLResource.OPTION_EXTENDED_META_DATA, ExtendedMetaData.INSTANCE);
		

		IModelDataStorage data = null;
		File settingsFile = new File(MODEL_PATH);
		
		try {
			resource.load(new FileInputStream(settingsFile), options);
			EObject result = resource.getContents().get(0);
			
			if (result instanceof IModelDataStorage) {
				data = (IModelDataStorage) result;
			} else {
				log.warn("Could not load the ModelDataStorage, generating a default ModelDataStorage");
			}
			
		} catch (FileNotFoundException exc) {
			log.warn("Couldn't locate model file '" + settingsFile.getAbsolutePath()
					+ "'; generating new, default ModelDataStorage!");
			
		} catch (IOException exc) {
			log.warn("Couldn't load model from file '" + settingsFile.getAbsolutePath()
					+ "'; generating new, default ModelDataStorage!");
		} finally {
			data = checkConsistency(data);
		}
		

		// Put data into HashMaps to grant faster access to it
		for (IRoboCupFieldData field : data.getFieldData()) {
			
			fieldData.put(field.getName(), field);
		}
		
		// for (Material mat : data.getMaterials()) {
		// matData.put(mat.getName(), mat);
		// }
		
		for (IBallModelData ball : data.getBallData()) {
			
			ballData.put(ball.getName(), ball);
		}
		

		for (ICTBotModelData bot : data.getCTBotData()) {
			
			ctBotData.put(bot.getName(), bot);
		}
		

		for (ITigerBotModelData bot : data.getTigerBotData()) {
			
			tigerBotData.put(bot.getName(), bot);
		}
	}
	

	private IModelDataStorage checkConsistency(IModelDataStorage data) {
		
		if (data == null) {
			data = factory.createModelDataStorage();
		}
		
		// Field
		if (data.getFieldData().isEmpty()) {
			IRoboCupFieldData fieldData = IFieldFactory.eINSTANCE.createRoboCupFieldData();
			data.getFieldData().add(fieldData);
		}
		
		// Ball
		if (data.getBallData().isEmpty()) {
			IBallModelData ballData = IBallFactory.eINSTANCE.createBallModelData();
			data.getBallData().add(ballData);
		}
		
		// CT-Bot
		if (data.getCTBotData().isEmpty()) {
			ICTBotModelData botData = IBotFactory.eINSTANCE.createCTBotModelData();
			data.getCTBotData().add(botData);
		}
		
		// Tiger-Bot
		if (data.getTigerBotData().isEmpty()) {
			ITigerBotModelData botData = IBotFactory.eINSTANCE.createTigerBotModelData();
			data.getTigerBotData().add(botData);
		}
		
		return data;
	}
	

	@Override
	public void cleanup() {
		
		synchronized (storeLock) {
			
			save();
			
			super.cleanup();
			
			if (fieldData != null) {
				fieldData.clear();
			}
			
			if (ballData != null) {
				ballData.clear();
			}
			
			if (ctBotData != null) {
				ctBotData.clear();
			}
			
			if (tigerBotData != null) {
				tigerBotData.clear();
			}
			
			// if (matData != null) {
			// matData.clear();
			// matData = null;
			// }
		}
	}
	

	private void save() {
		
		synchronized (storeLock) {
			
			IModelDataStorage data = IDataFactory.eINSTANCE.createModelDataStorage();
			

			data.getFieldData().clear();
			for (IRoboCupFieldData field : fieldData.values()) {
				
				data.getFieldData().add(field);
			}
			// data.getMaterials().addAll(matData.values());
			
			data.getBallData().clear();
			for (IBallModelData ball : ballData.values()) {
				
				data.getBallData().add(ball);
			}
			

			data.getCTBotData().clear();
			for (ICTBotModelData bot : ctBotData.values()) {
				
				data.getCTBotData().add(bot);
			}
			
			data.getTigerBotData().clear();
			for (ITigerBotModelData bot : tigerBotData.values()) {
				
				data.getTigerBotData().add(bot);
			}
			

			// Save SimProperties using EMF
			Resource resource = new XMLResourceImpl();
			resource.getContents().add(data);
			
			// Options
			Map<String, Object> options = new HashMap<String, Object>();
			options.put(XMLResource.OPTION_KEEP_DEFAULT_CONTENT, Boolean.TRUE); // Write default data to file
			options.put(XMLResource.OPTION_SAVE_ONLY_IF_CHANGED, Boolean.TRUE);
			options.put(XMLResource.OPTION_FORMATTED, Boolean.TRUE);
			options.put(XMLResource.OPTION_LINE_WIDTH, 1);
			
			File file = new File(MODEL_PATH);
			
			try {
				
				if (!file.exists()) {
					if (!file.createNewFile()) {
						log.error("Couldn't create model-data-file to '" + file.getAbsolutePath() + "'; they get lost...");
					}
				}
				
				FileOutputStream fos = new FileOutputStream(file);
				resource.save(fos, options);
				fos.flush();
				fos.close();
				log.info("Successfully saved model data!");
				
			} catch (FileNotFoundException exc) {
				log.error("Couldn't write model data to '" + file.getAbsolutePath() + "'; they get lost...", exc);
			} catch (IOException exc) {
				log.error("Couldn't write model data to '" + file.getAbsolutePath() + "'; they get lost...", exc);
			}
		}
	}
	

	@Override
	public IRoboCupFieldData getFieldData(String type) {
		
		synchronized (storeLock) {
			
			IRoboCupFieldData newData = fieldData.get(type);
			
			if (newData == null) {
				newData = fieldData.get(STAN_FIELD);
				log.info("No such FieldData ('" + type + "') found, took '" + STAN_FIELD + "'");
			}
			
			return newData;
		}
	}
	

	@Override
	public Material getMaterial(String type) {
		
		synchronized (storeLock) {
			
			Material newData = null; // matData.get(type);
			
			// if (newData == null) {
			// newData = matData.get(STAN_MATERIAL);
			// log.info("Now such MaterialData found, took '" + STAN_MATERIAL + "'");
			// }
			
			return newData;
		}
	}
	

	@Override
	public IBallModelData getRoboCupBallData(String type) {
		
		synchronized (storeLock) {
			
			IBallModelData newData = ballData.get(type);
			
			if (newData == null) {
				newData = ballData.get(STAN_BALL);
				log.info("No such BallData ('" + type + "') found, took '" + STAN_BALL + "'");
			}
			
			return newData;
		}
	}
	

	@Override
	public ICTBotModelData getCTBotModelData(String type) {
		
		synchronized (storeLock) {
			
			ICTBotModelData newData = ctBotData.get(type);
			
			if (newData == null) {
				newData = ctBotData.get(STAN_CT_BOT);
				log.info("No such BotData ('" + type + "') found, took '" + STAN_CT_BOT + "'");
			}
			
			return newData;
		}
	}
	

	@Override
	public ITigerBotModelData getTigerBotModelData(String type) {
		
		synchronized (storeLock) {
			
			ITigerBotModelData newData = tigerBotData.get(type);
			
			if (newData == null) {
				newData = tigerBotData.get(STAN_TIGER_BOT);
				log.info("No such BotData ('" + type + "') found, took '" + STAN_TIGER_BOT + "'");
			}
			
			return newData;
		}
	}
	

	@Override
	public List<String> getAvailableFieldTypes() {
		final List<String> types = new ArrayList<String>();
		for (IRoboCupFieldData data : fieldData.values()) {
			types.add(data.getName());
		}
		return Collections.unmodifiableList(types);
	}
	

	@Override
	public List<String> getAvailableBallTypes() {
		final List<String> types = new ArrayList<String>();
		for (IBallModelData data : ballData.values()) {
			types.add(data.getName());
		}
		return Collections.unmodifiableList(types);
	}
}
