package edu.dhbw.mannheim.tigers.sim.engine.model;


import java.util.ArrayList;

import org.apache.log4j.Logger;

import com.jme.bounding.BoundingBox;
import com.jme.math.Quaternion;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.scene.Node;
import com.jme.system.DisplaySystem;
import com.jmex.physics.CollisionGroup;
import com.jmex.physics.PhysicsNode;
import com.jmex.physics.PhysicsSpace;
import com.jmex.physics.StaticPhysicsNode;
import com.jmex.physics.impl.ode.OdePhysicsSpace;

import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.engine.com.ERoboCupTeam;
import edu.dhbw.mannheim.tigers.sim.engine.model.ball.IRoboCupBallFactory;
import edu.dhbw.mannheim.tigers.sim.engine.model.ball.RoboCupBall;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.ABotModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.EBotModelType;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.IBotModelFactory;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.PseudoBot;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.ct.CTBotModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger.TigerBotModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.ball.IBallModelData;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.bot.ICTBotModelData;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.bot.ITigerBotModelData;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.field.IRoboCupFieldData;
import edu.dhbw.mannheim.tigers.sim.engine.model.field.FieldGround;
import edu.dhbw.mannheim.tigers.sim.engine.model.field.Goal;
import edu.dhbw.mannheim.tigers.sim.engine.model.field.IRoboCupFieldFactory;
import edu.dhbw.mannheim.tigers.sim.engine.model.field.RoboCupField;
import edu.dhbw.mannheim.tigers.sim.engine.model.field.SurrWall;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelException;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelResourceException;


/**
 * <p>
 * This is an implementation of the Factory-Pattern for the model-elements of the simulator.
 * </p>
 * <p>
 * <b>Conventions:</b>
 * 
 * <ul>
 * 
 * <li>
 * <u>Scale:</u> The models manufactured by this factory already have the correct size to fit into this simulation (If
 * you want to move/translate them however, remember your units! See {@link UnitConversion}</li>
 * 
 * <li>
 * <u>Physical:</u> The <b>dynamic</b> models manufactured by this factory <b>are not active</b> parts of the physical
 * space, you have to activate them by using {@link PhysicsNode#setActive(boolean)} or
 * {@link ABotModel#setActive(boolean)}! Why? Because if the models are created at runtime, they have already fallen
 * through the ground before they have a correct physical representation!</li>
 * 
 * </ul>
 * </p>
 * 
 * @author Gero
 */
public class ModelFactory implements IBotModelFactory, IRoboCupFieldFactory, IRoboCupBallFactory {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final Logger								log					= Logger.getLogger(getClass());
	
	private static ModelFactory					instance				= null;
	
	private final ModelDataProvider				dataProvider		= ModelDataProvider.getInstance();
	private final ArrayList<CollisionGroup>	collisionGroups	= new ArrayList<CollisionGroup>();
	
	private ICore										core					= null;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	private ModelFactory() {
	}
	

	public static synchronized ModelFactory getInstance() {
		if (instance == null) {
			instance = new ModelFactory();
		}
		return instance;
	}
	
	
	@Override
	public void setCore(ICore newCore) {
		this.core = newCore;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public ABotModel createBotModel(PhysicsSpace space, DisplaySystem display, EBotModelType modelType, ERoboCupTeam team) {
		return createBotModel(space, display, modelType.toString(), team, null);
	}
	

	@Override
	public ABotModel createBotModel(PhysicsSpace space, DisplaySystem display, String modelType, ERoboCupTeam team) {
		return createBotModel(space, display, modelType, team, null);
	}
	

	@Override
	public ABotModel createBotModel(PhysicsSpace space, DisplaySystem display, EBotModelType modelType, ERoboCupTeam team, ColorRGBA color) {
		return createBotModel(space, display, modelType.toString(), team, color);
	}
	

	@Override
	public ABotModel createBotModel(PhysicsSpace space, DisplaySystem display, String modelType, ERoboCupTeam team, ColorRGBA color) {
		
		ABotModel newModel = null;
		
		if (color == null) {
			color = ColorRGBA.red;
		}
		
		EBotModelType type = EBotModelType.valueOf(modelType);
		if (type == null) {
			type = EBotModelType.CT_MODEL;
			log.info("'" + modelType + "' doesn't represent any valid EBotModelType, took '" + type + "'!");
		}
		
		try {
			switch (type) {
				
				case PSEUDO_BOT:
					ITigerBotModelData pseudoData = dataProvider.getTigerBotModelData(""); // Make ModelDataProvider give _anything
					
					newModel = new PseudoBot((OdePhysicsSpace) space, display, pseudoData, color, team);
					newModel.setName("Pseudo-Bot " + newModel.getID());
					break;
				
				case TIGER_MODEL:
					ITigerBotModelData tigerData = dataProvider.getTigerBotModelData(modelType.toString());
					
					newModel = new TigerBotModel((OdePhysicsSpace) space, display, tigerData, core.getBalls(), color.clone(), team);
					newModel.setName("TIGER " + newModel.getID());
					break;
				
				default:
				case CT_MODEL:
					ICTBotModelData data = dataProvider.getCTBotModelData(modelType.toString());
					
					newModel = new CTBotModel((OdePhysicsSpace) space, display, data, color.clone(), team);
					newModel.setName("CT-Bot " + newModel.getID());
					break;
			}
			
		} catch (SimModelResourceException err) {
			log.error(err);
		} catch (SimModelException err) {
			log.error(err);
		}
		
		newModel.setLocalScale(getModelScale());
		

		// Attach CollisionGroups to each other
		for (CollisionGroup group : collisionGroups) {
			newModel.getCollisionGroup().collidesWith(group, true);
		}
		collisionGroups.add(newModel.getCollisionGroup());
		

		return newModel;
	}
	

	@Override
	public RoboCupField createField(PhysicsSpace space, DisplaySystem display, String type) throws SimModelException,
			SimModelResourceException {
		
		IRoboCupFieldData data = dataProvider.getFieldData(type.toString());
		
		final RoboCupField field = new RoboCupField(data);
		

		// Create ground ################################
		FieldGround fieldGround = new FieldGround((OdePhysicsSpace) space, display, data.getTotalFieldWidth(),
				data.getTotalFieldLength(), data.getFieldDepth());
		

		// Create the goals ################################
		float goalToFieldCenterDistance = data.getFieldLength() / 2;
		
		StaticPhysicsNode goals = space.createStaticNode();
		goals.setName("goals");
		Node goalA = new Goal((OdePhysicsSpace) space, display, data);
		goalA.setLocalTranslation(0, 0, goalToFieldCenterDistance);
		
		// Make two of them
		Node goalB = new Goal((OdePhysicsSpace) space, display, data);
		
		Quaternion turnaroundGoalB = new Quaternion();
		turnaroundGoalB.fromAngleAxis((float) Math.toRadians(180), Vector3f.UNIT_Y);
		goalB.setLocalRotation(turnaroundGoalB);
		goalB.setLocalTranslation(0, 0, -goalToFieldCenterDistance);
		
		// Finish goals
		goals.attachChild(goalA);
		goals.attachChild(goalB);
		

		// Surroundings ################################
		StaticPhysicsNode surrGroup = space.createStaticNode();
		surrGroup.setName("surroundings");
		
		Node surrWall = new SurrWall((OdePhysicsSpace) space, display, data, data.getFieldDepth());
		
		surrGroup.attachChild(surrWall);
		

		// Attach all groups
		field.attachChild(fieldGround);
		field.attachChild(goals);
		field.attachChild(surrGroup);
		
		// Prepare field itself
		field.setModelBound(new BoundingBox());
		field.updateModelBound();
		

		// Model scale ################################
		if (field != null) {
			field.setLocalScale(getModelScale());
		}
		
		return field;
	}
	

	@Override
	public RoboCupBall createRoboCupBall(OdePhysicsSpace space, DisplaySystem display) {
		
		IBallModelData data = dataProvider.getRoboCupBallData("");
		
		RoboCupBall ball = new RoboCupBall(space, display, data);
		
		ball.setMass(data.getMass());
		ball.setLocalScale(getModelScale());
		
		return ball;
	}
	

	/** Get World-To-Sim-scale */
	private static float getModelScale() {
		return UnitConversion.getModelScale();
	}
}
