package edu.dhbw.mannheim.tigers.sim.engine.model;


import org.apache.log4j.Logger;

import com.jme.input.InputHandler;
import com.jme.input.action.InputAction;
import com.jme.input.action.InputActionEvent;
import com.jme.intersection.PickData;
import com.jme.intersection.TrianglePickResults;
import com.jme.math.Ray;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.scene.Node;
import com.jme.scene.Spatial;
import com.jme.system.DisplaySystem;
import com.jme.util.export.Savable;
import com.jmex.physics.DynamicPhysicsNode;
import com.jmex.physics.Joint;
import com.jmex.physics.PhysicsCollisionGeometry;
import com.jmex.physics.PhysicsSpace;
import com.jmex.physics.StaticPhysicsNode;
import com.jmex.physics.contact.ContactInfo;
import com.jmex.physics.geometry.PhysicsRay;
import com.jmex.physics.util.PhysicsPicker;

import edu.dhbw.mannheim.tigers.sim.engine.model.UserData.Value;


/**
 * A modification of jmePhysics2 {@link PhysicsPicker}, which moves objects in a certain height over
 * the y0-Plane instead of the viewport-plane, and allows further influence on the pickedNode.
 * 
 * @author Gero
 */
public class SimPhysicsPicker {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	protected final Logger					log				= Logger.getLogger(getClass());
	
	// Nodes and joints
	/** root node of the scene - used for picking */
	private final Node						rootNode;
	
	/** helper no the picked node is joined to */
	private final DynamicPhysicsNode		myNode;
	
	/** joint to link myNode and picked node */
	private final Joint						joint;
	
	/** joint to fix myNode in the world */
	private final Joint						worldJoint;
	
	private final Ray							pickRay			= new Ray();
	
	private final TrianglePickResults	pickResults		= new TrianglePickResults();
	

	// Actions and their handler
	private final InputHandler				rootInput;
	private final InputHandler				inputHandler;
	private InputHandler						pickHandler;
	private SimPhysicsPicker.PickAction	pickAction;
	private SimPhysicsPicker.MoveAction	moveAction;
	
	// Picking
	private final Vector2f					mousePosition	= new Vector2f();
	private final boolean					onlyPick;
	private DynamicPhysicsNode				picked;
	
	private PhysicsRay						physicsRay;
	private PhysicsCollisionGeometry		nearestPickedGeom;
	private float								minPickDistance;
	
	private final Vector3f					anchor			= new Vector3f();
	
	private boolean							pickModeVisual	= true;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Constructor of the class.
	 * 
	 * @param input where {@link #getInputHandler()} is attached to
	 * @param rootNode root node of the scene - used for picking
	 * @param physicsSpace physics space to create joints in (picked nodes must reside in there)
	 * @param onlyPick The picked node is only returned, no move-action is applied
	 */
	public SimPhysicsPicker(InputHandler input, Node rootNode, PhysicsSpace physicsSpace, boolean onlyPick) {
		this(input, rootNode, physicsSpace, false, onlyPick);
	}
	

	/**
	 * Constructor of the class.
	 * 
	 * @param input where {@link #getInputHandler()} is attached to
	 * @param rootNode root node of the scene - used for picking
	 * @param physicsSpace physics space to create joints in (picked nodes must reside in there)
	 * @param allowRotation true to allow rotation of the picked object
	 * @param onlyPick The picked node is only returned, no move-action is applied
	 */
	private SimPhysicsPicker(InputHandler input, Node rootNode, PhysicsSpace physicsSpace, boolean allowRotation,
			boolean onlyPick) {
		
		this.rootInput = input;
		this.onlyPick = onlyPick;
		
		this.inputHandler = new InputHandler();
		input.addToAttachedHandlers(this.inputHandler);
		this.rootNode = rootNode;
		joint = physicsSpace.createJoint();
		if (allowRotation) {
			joint.createRotationalAxis().setDirection(Vector3f.UNIT_X);
			joint.createRotationalAxis().setDirection(Vector3f.UNIT_Y);
			joint.createRotationalAxis().setDirection(Vector3f.UNIT_Z);
		}
		joint.setSpring(2000, 200);
		myNode = physicsSpace.createDynamicNode();
		myNode.setName("Physics Picker Helper Node");
		myNode.setAffectedByGravity(false);
		worldJoint = physicsSpace.createJoint();
		activatePhysicsPicker();
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @return the input handler for this picker
	 */
	public InputHandler getInputHandler() {
		return inputHandler;
	}
	

	/**
	 * Should be overridden by subclasses (default: mouse)
	 * @return Should return the device the pick-action is activated by
	 */
	protected String getDevice() {
		return InputHandler.DEVICE_MOUSE;
	}
	

	/**
	 * Should be overridden by subclasses (default: 0/left-click)
	 * @return Should return the key on which the pick-action is activated
	 */
	protected int getKey() {
		return 0;	// left-click
	}
	

	private void activatePhysicsPicker() {
		pickAction = new PickAction();
		inputHandler.addAction(pickAction, getDevice(), getKey(), InputHandler.AXIS_NONE, false);
		
		moveAction = new MoveAction();
		inputHandler.addAction(moveAction, InputHandler.DEVICE_MOUSE, InputHandler.BUTTON_NONE, InputHandler.AXIS_ALL,
				false);
		
		if (!onlyPick) {
			// right-click
			inputHandler.addAction(new InputAction() {
				
				@Override
				public void performAction(InputActionEvent evt) {
					if (!evt.getTriggerPressed()) {
						upright(getPickedNode());
					}
				}
			}, InputHandler.DEVICE_MOUSE, 1, InputHandler.AXIS_NONE, false);
			
			// middle-click
			inputHandler.addAction(new InputAction() {
				
				@Override
				public void performAction(InputActionEvent evt) {
					if (!evt.getTriggerPressed()) {
						heightOffset = HEIGHT_OFFSET_DEFAULT;
						updateMove();
					}
				}
			}, InputHandler.DEVICE_MOUSE, 2, InputHandler.AXIS_NONE, false);
		}
	}
	

	private void release() {
		picked = null;
		joint.detach();
		worldJoint.detach();
		myNode.setActive(false);
		
		heightOffset = HEIGHT_OFFSET_DEFAULT;
		
		onRelease();
	}
	

	/** Does nothing on default, but provides subclasses access to the picked node! */
	public void onRelease() {
	}
	

	// private final Vector3f pickedScreenPos = new Vector3f();
	// private final Vector3f pickedWorldOffset = new Vector3f();
	

	private void attach(DynamicPhysicsNode node) {
		// DisplaySystem.getDisplaySystem().getScreenCoordinates(node.getWorldTranslation(),
		// pickedScreenPos);
		// DisplaySystem.getDisplaySystem().getWorldCoordinates(mousePosition, pickedScreenPos.z,
		// pickedWorldOffset);
		
		picked = node;
		// log.debug("##### New Pick #####");
		// log.debug("Node L: " + node.getLocalTranslation());
		// log.debug("Node W: " + node.getWorldTranslation());
		node.updateGeometricState(0, true);
		node.localToWorld(node.getCenterOfMass(myNode.getLocalTranslation()), myNode.getLocalTranslation());
		// log.debug("myNode L: " + myNode.getLocalTranslation());
		// log.debug("myNode W: " + myNode.getWorldTranslation());
		myNode.setActive(true);
		worldJoint.setAnchor(myNode.getLocalTranslation());
		worldJoint.attach(myNode);
		joint.attach(myNode, node);
		joint.setAnchor(new Vector3f());
	}
	

	/** Does nothing on default, but provides subclasses access to the picked node! */
	public void onAttach(DynamicPhysicsNode picked) {
	}
	

	public void delete() {
		// inputHandler.removeAction( pickAction );
		// inputHandler.removeAction( moveAction );
		inputHandler.removeAllActions();
		rootInput.removeFromAttachedHandlers(inputHandler);
		
		myNode.setActive(false);
		myNode.removeFromParent();
		joint.detach();
		joint.setActive(false);
		worldJoint.detach();
		worldJoint.setActive(false);
		picked = null;
	}
	

	public DynamicPhysicsNode getPickedNode() {
		return picked;
	}
	

	/**
	 * @return true if picking is done with the visuals, false if done with the physics
	 *         representation
	 * @see #setPickModeVisual(boolean)
	 */
	public boolean isPickModeVisual() {
		return pickModeVisual;
	}
	

	/**
	 * The mode used for picking: visual picking uses the bounding volumes and visual meshes of scene
	 * elements to
	 * determine what was clicked by the user, the physical mode uses the actual physical
	 * representations.
	 * 
	 * @param pickModeVisual true to switch to visual mode, false to switch to physical mode
	 */
	public void setPickModeVisual(boolean pickModeVisual) {
		this.pickModeVisual = pickModeVisual;
	}
	
	private class PickAction extends InputAction {
		
		@Override
		public void performAction(InputActionEvent evt) {
			if (evt.getTriggerPressed()) {
				DisplaySystem.getDisplaySystem().getWorldCoordinates(mousePosition, 0, pickRay.origin);
				DisplaySystem.getDisplaySystem().getWorldCoordinates(mousePosition, 0.3f, pickRay.direction);
				pickRay.direction.subtractLocal(pickRay.origin).normalizeLocal();
				
				if (pickModeVisual) {
					pickVisual();
				} else {
					pickPhysical();
				}
			} else {
				release();
			}
		}
	}
	

	private boolean isPickable(DynamicPhysicsNode pickedNode) {
		boolean result = false;
		Savable sav = pickedNode.getUserData(UserData.PICKABLE);
		if (sav instanceof UserData) {
			UserData ud = (UserData) sav;
			result = ud.getValue().equals(Value.TRUE);
		}
		
		return result;
	}
	
	
	// --------------------------------------------------------------------------
	// --- visual pick ----------------------------------------------------------
	// --------------------------------------------------------------------------
	private void pickVisual() {
		pickResults.clear();
		pickResults.setCheckDistance(true);
		
		rootNode.findPick(pickRay, pickResults);
		
		loopResults: for (int i = 0; i < pickResults.getNumber(); i++) {
			PickData data = pickResults.getPickData(i);
			if (data.getTargetTris() != null && data.getTargetTris().size() > 0) {
				Spatial target = data.getTargetMesh();
				while (target != null) {
					if (target instanceof DynamicPhysicsNode) {
						DynamicPhysicsNode picked = (DynamicPhysicsNode) target;
//						attach(picked);
						
						attachPickableNode(picked);
						
						break loopResults;
					}
					target = target.getParent();
				}
			}
		}
	}
	

	// --------------------------------------------------------------------------
	// --- physic pick ----------------------------------------------------------
	// --------------------------------------------------------------------------
	private void pickPhysical() {
		if (physicsRay == null) {
			final StaticPhysicsNode pickNode = myNode.getSpace().createStaticNode();
			pickNode.setActive(false);
			physicsRay = pickNode.createRay("pickRay");
			pickHandler = new InputHandler();
			pickHandler.addAction(new PhysicsPickAction(), physicsRay.getCollisionEventHandler(), false);
		}
		physicsRay.getLocalTranslation().set(pickRay.getOrigin());
		physicsRay.getLocalScale().set(pickRay.getDirection()).multLocal(1000);
		physicsRay.updateWorldVectors();
		
		nearestPickedGeom = null;
		minPickDistance = Float.POSITIVE_INFINITY;
		pickHandler.setEnabled(true);
		myNode.getSpace().pick(physicsRay);
		pickHandler.update(0);
		pickHandler.setEnabled(false);
		physicsRay.getLocalScale().set(1, 1, 1);
		
		if (nearestPickedGeom != null) {
			DynamicPhysicsNode pickedNode = (DynamicPhysicsNode) nearestPickedGeom.getPhysicsNode();
			attachPickableNode(pickedNode);
		}
	}
	
	
	private void attachPickableNode(DynamicPhysicsNode selectedNode) {
		if (selectedNode != null && isPickable(selectedNode)) {
			if (!onlyPick) {
				attach(selectedNode);
			}
			onAttach(selectedNode);
		}
	}
	
	
	private class PhysicsPickAction extends InputAction {
		@Override
		public void performAction(InputActionEvent evt) {
			final ContactInfo info = (ContactInfo) evt.getTriggerData();
			PhysicsCollisionGeometry other = info.getGeometry1() == physicsRay ? info.getGeometry2() : info.getGeometry1();
			if (!other.getPhysicsNode().isStatic() && info.getPenetrationDepth() < minPickDistance) {
				minPickDistance = info.getPenetrationDepth();
				nearestPickedGeom = other;
			}
		}
	}
	

	// --------------------------------------------------------------------------
	// --- move the picked object -----------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Uprights the pickedNode
	 * 
	 * @param pickedNode
	 */
	private void upright(DynamicPhysicsNode pickedNode) {
		if (pickedNode != null) {
			release();
			
			// log.debug("##### Upright #####");
			// log.debug("Local Rotation: "+ pickedNode.getLocalRotation());
			// log.debug("Local Rotation World: "+ pickedNode.getWorldRotation());
			pickedNode.setLocalRotation(ModelHelper.x90.clone());
			// log.debug("Local Rotation: "+ pickedNode.getLocalRotation());
			// log.debug("Local Rotation World: "+ pickedNode.getWorldRotation());
			pickedNode.updateGeometricState(0, true);
			
			attach(pickedNode);
		}
	}
	
	
	private static final float	HEIGHT_OFFSET_DEFAULT	= UnitConversion.cm(10);
	private float					heightOffset				= HEIGHT_OFFSET_DEFAULT;
	private static final float	HEIGHT_MOD					= UnitConversion.cm(5);
	
	private class MoveAction extends InputAction {
		
		@Override
		public void performAction(InputActionEvent evt) {
			
			switch (evt.getTriggerIndex()) {
				case 0:
					mousePosition.x = evt.getTriggerPosition() * DisplaySystem.getDisplaySystem().getWidth();
					break;
				case 1:
					mousePosition.y = evt.getTriggerPosition() * DisplaySystem.getDisplaySystem().getHeight();
					break;
				case 2:
					// move into z direction with the wheel
					if (evt.getTriggerDelta() > 0) {
						heightOffset += HEIGHT_MOD;
					} else {
						heightOffset -= HEIGHT_MOD;
					}
					break;
			}
			
			// Update with mouse-position
			if (!onlyPick) {
				updateMove();
			}
		}
	}
	
	
	private void updateMove() {
		
		if (picked != null) {
			Vector3f intersection = CamHelper.calculateMouseOnGround(mousePosition);
			
			// log.debug("Position:" + intersection);
			
			anchor.set(intersection);
			// anchor.subtractLocal(pickedWorldOffset);
			anchor.y = heightOffset;
			
			myNode.getLocalTranslation().set(anchor);
			worldJoint.setAnchor(myNode.getLocalTranslation());
			worldJoint.attach(myNode);
			
			// joint.getAnchor(jAnchor).set(world)
			
			// log.debug("Intersection: " + intersection);
			// log.debug("Anchor: " + anchor);
			// log.debug("myNode Local: " + myNode.getLocalTranslation());
			// log.debug("myNode World: " + myNode.getWorldTranslation());
			// log.debug("Picked Local: " + picked.getLocalTranslation());
			// log.debug("Picked World: " + picked.getWorldTranslation());
			// log.debug("Joint: " + joint.getAnchor(null));
		}
	}
}
