package edu.dhbw.mannheim.tigers.sim.engine.model.bot.ct;


import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.data;

import com.jme.bounding.BoundingBox;
import com.jme.bounding.BoundingSphere;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.renderer.Renderer;
import com.jme.scene.Node;
import com.jme.scene.shape.Cylinder;
import com.jme.scene.shape.Sphere;
import com.jme.system.DisplaySystem;
import com.jmex.physics.DynamicPhysicsNode;
import com.jmex.physics.Joint;
import com.jmex.physics.RotationalJointAxis;
import com.jmex.physics.geometry.PhysicsCylinder;
import com.jmex.physics.impl.ode.OdePhysicsSpace;

import edu.dhbw.mannheim.tigers.sim.engine.com.ERoboCupTeam;
import edu.dhbw.mannheim.tigers.sim.engine.model.ModelHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.ABotModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.EBotModelType;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.bot.ICTBotModelData;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.RoboCupMaterial;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelException;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelResourceException;


/**
 * The {@link ABotModel} implementation of the CT-Bot.
 * It has two wheels and should use them to drive around!
 * 
 * @author Gero
 */
public class CTBotModel extends ABotModel implements ICTBotControl {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long			serialVersionUID	= -5609996649395053527L;

	// Constants
	private static final float			MOTOR_FORCE			= 0.15f;

	// Size
	private final ICTBotModelData		data;

	// Parts
	private final DynamicPhysicsNode	body;
	private final DynamicPhysicsNode	rWheel;
	private final DynamicPhysicsNode	lWheel;
	private final DynamicPhysicsNode	frontKnob;
	private final DynamicPhysicsNode	rearKnob;
	// private final DynamicPhysicsNode knockBlock;
	
	private RotationalJointAxis		rWheelRotation;
	private RotationalJointAxis		lWheelRotation;
	

	private final ColorRGBA				color;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @param space
	 * @param display
	 * @param newData
	 * @param color May be null!!!
	 * @param team
	 * @throws SimModelResourceException
	 * @throws SimModelException
	 */
	public CTBotModel(OdePhysicsSpace space, DisplaySystem display, ICTBotModelData newData,
			ColorRGBA color, ERoboCupTeam team) throws SimModelResourceException, SimModelException {
		super(space, display, EBotModelType.CT_MODEL, newData.getName(), team);
		
		this.data = newData;
		this.color = color;
		
		addPhysicNode(body = space.createDynamicNode());
		
		addPhysicNode(rWheel = space.createDynamicNode());
		addPhysicNode(lWheel = space.createDynamicNode());
		
		addPhysicNode(frontKnob = space.createDynamicNode());
		addPhysicNode(rearKnob = space.createDynamicNode());
		
		// addPhysicNode(knockBlock = space.createDynamicNode());
		
		createPhysicalNode();
		
		// Make the body pickable!
		makePickable(body);
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void createPhysicalNode() throws SimModelResourceException, SimModelException {
		
		/*
		 * Here we want to build a "simple" CT-bot consisting of the following parts:
		 * - a cylindric body
		 * - two wheels
		 * - and two small knobs preventing the bot from falling forth or back
		 * 
		 * It is looking towards X+, right is Z+, left Z-
		 */

		// ##### The Body
		final float bodyRadius = data.getRadius();
		final float bodyHeight = data.getHeight();
		final float bodyToGroundDist = data.getBodyToGroundDist();
		
		// Visual
		body.setName("CT-Bot [body]");
		final Cylinder vBody = new Cylinder("CT-Bot [body] Vis", 5, 25, bodyRadius, bodyHeight, true);
		vBody.setModelBound(new BoundingBox());
		vBody.updateModelBound();
		
		ModelHelper.color(vBody, display, color); // Color!
		
		body.attachChild(vBody);
		
		// Physic
		final PhysicsCylinder phBody = body.createCylinder("Bot-Body");
		phBody.setMaterial(RoboCupMaterial.BOT_BODY);
		phBody.setLocalScale(new Vector3f(bodyRadius, bodyRadius, bodyHeight));
		
		// Location
		body.setLocalTranslation(0, data(bodyHeight / 2 + bodyToGroundDist), 0); // Lift it a bit
																											// above the floor
		
		// Rotation
		body.setLocalRotation(ModelHelper.x90.clone()); // and turn it up!
		
		this.attachChild(body);
		

		// // ##### KnockBlock
		// // Sits in front of the bots to knock balls!
		// // Visual
		// final float bX = 10;
		// final float bY = 20;
		// final float bZ = data.getRadius();
		//
		// final Box vKnockBlock = new Box("knockBlock", Vector3f.ZERO.clone(), bX / 2, bY / 2, bZ /
		// 2);
		// vKnockBlock.setModelBound(new BoundingBox());
		// vKnockBlock.updateModelBound();
		//
		// ModelHelper.color(vKnockBlock, display, ColorRGBA.cyan);
		//
		// knockBlock.attachChild(vKnockBlock);
		//
		// // Physic
		// final PhysicsBox phKnockBlock = knockBlock.createBox("phKnockBlock");
		// phKnockBlock.setMaterial(RoboCupMaterial.BOT_BODY);
		// phKnockBlock.setLocalScale(new Vector3f(bX, bY, bZ));
		//
		// // Location
		// knockBlock.setLocalTranslation(data(data.getRadius()), data(data.getBodyToGroundDist() +
		// (bY / 2)), 0);
		//
		// // Connect
		// final Joint kbJoint = space.createJoint();
		// kbJoint.attach(knockBlock, body);
		//
		// this.attachChild(knockBlock);
		

		// ##### Wheels
		final float wheelRadius = data.getWheelRadius();
		final float wheelWidth = data.getWheelWidth();
		final float wheelToBodyDist = data.getBodyToWheelDist();
		
		// The right wheel...
		rWheel.setName("CT-Bot [right wheel]");
		adjustWheel(rWheel, "right", wheelRadius, wheelWidth);
		rWheel.setLocalTranslation(0, data(wheelRadius), data(bodyRadius + wheelToBodyDist)); // Move
																															// into
																															// correct
																															// position
		this.attachChild(rWheel);
		
		// ...and the left
		lWheel.setName("CT-Bot [left wheel]");
		adjustWheel(lWheel, "left", wheelRadius, wheelWidth);
		lWheel.setLocalTranslation(0, data(wheelRadius), -data(bodyRadius + wheelToBodyDist));
		this.attachChild(lWheel);
		

		// Now we add the joints...
		rWheelRotation = attachWheel(rWheel);
		lWheelRotation = attachWheel(lWheel);
		

		// ##### Knobs
		final float knobRadius = data.getBodyToGroundDist() / 2;
		// final float knobOffset = knobRadius / 2;
		
		// Front...
		adjustKnob(frontKnob, knobRadius);
		frontKnob.setLocalTranslation(data(data.getRadius() - knobRadius), data(knobRadius), 0);
		this.attachChild(frontKnob);
		
		final Joint frontKnobJoint = space.createJoint();
		frontKnobJoint.attach(frontKnob, body);
		
		// ...and rear!
		adjustKnob(rearKnob, knobRadius);
		rearKnob.setLocalTranslation(-data(data.getRadius() - knobRadius), data(knobRadius), 0);
		this.attachChild(rearKnob);
		
		final Joint rearKnobJoint = space.createJoint();
		rearKnobJoint.attach(rearKnob, body);
		

		// Change center of mass (at the end for easier attaching of other objects to the body (as
		// their attached relative to the CoM)
		body.setCenterOfMass(new Vector3f(0, 0, -data(data.getHeight() / 2
				+ data.getBodyToGroundDist())));
		body.setMass(data.getMass()); // Set mass after adjusting center of mass to make it
												// _happen!;-))
		

		// Finish...
		this.setRenderQueueMode(Renderer.QUEUE_OPAQUE);
	}
	

	private void adjustWheel(DynamicPhysicsNode wheel, String whichWheel, final float radius,
			final float width) {
		
		// Visual
		final Cylinder visualWheel = new Cylinder(whichWheel + "visualWheel", 5, 25, radius, width,
				true);
		visualWheel.setModelBound(new BoundingSphere());
		visualWheel.updateModelBound();
		
		// Colorize
		ModelHelper.color(visualWheel, display, ColorRGBA.lightGray);
		
		wheel.attachChild(visualWheel);
		
		// Physic
		final PhysicsCylinder cylinder = wheel.createCylinder(whichWheel + "physicalWheel");
		cylinder.setMaterial(RoboCupMaterial.BOT_WHEEL);
		cylinder.setLocalScale(new Vector3f(radius, radius, width));
		// wheel.generatePhysicsGeometry();
		
		wheel.setMass(0.05f);
	}
	

	private RotationalJointAxis attachWheel(DynamicPhysicsNode wheel) {
		
		final Joint joint = space.createJoint();
		final RotationalJointAxis rotationAxis = joint.createRotationalAxis(); // Add some freedom to
																										// the joint
		rotationAxis.setDirection(new Vector3f(0, 0, -1f));
		rotationAxis.setAvailableAcceleration(MOTOR_FORCE);
		joint.attach(wheel, body);
		
		return rotationAxis;
	}
	

	private void adjustKnob(DynamicPhysicsNode knob, final float radius) {
		// Visual knob
		final Sphere visualSphere = new Sphere("Knob", 10, 10, radius);
		visualSphere.setModelBound(new BoundingSphere());
		visualSphere.updateModelBound();
		
		ColorRGBA color = ColorRGBA.black;
		ModelHelper.color(visualSphere, display, color);
		
		knob.attachChild(visualSphere);
		knob.setMaterial(RoboCupMaterial.BOT_BODY_KNOB);
		knob.generatePhysicsGeometry();
		knob.setMass(0.01f);
	}
	

	/**
	 * As the CT-Bot-Controller sends the wanted desired velocity, so we have to convert it to the
	 * angular velocity.
	 * 
	 * wheelPerimeter = 2*Pi * r;
	 * v(desired) = (2*Pi) * x / wheelPerimeter;
	 * 
	 * => v(desired) = x / r;
	 */
	@Override
	public void setSpeed(int left, int right) {
		rWheelRotation.setDesiredVelocity((float) (right / data.getWheelRadius()));
		lWheelRotation.setDesiredVelocity((float) (left / data.getWheelRadius()));
	}
	

	// --------------------------------------------------------------------------
	// --- ABotModel ------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void clearVelocity() {
		super.clearVelocity();
		
		rWheelRotation.setDesiredVelocity(0);
		lWheelRotation.setDesiredVelocity(0);
	}
	

	// --------------------------------------------------------------------------
	// --- IUpdateable ----------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void doUpdate(long cycleNumber, float tpc) {
		
		body.updateGeometricState(0f, true);
		notifyObservers(getCurrentBotStatus(body.getWorldTranslation(), body.getWorldRotation()));
		
		// Vector3f vec = body.getWorldTranslation();
		// if (vec.x < 0.0001f && vec.x > -0.0001f &&
		// vec.z < 0.0001f && vec.z > -0.0001f) {
		// log.debug("##### " + getID() + " 0:0!");
		// }
		
		// log.debug(DefaultCamFrameBuilder.shiftRadians(getCurrentBotStatus(body.getWorldTranslation(),
		// body.getWorldRotation()).orientation, FastMath.HALF_PI));
	}
	

	// --------------------------------------------------------------------------
	// --- getter ---------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public float getHeight() {
		return data.getHeight() + data.getBodyToGroundDist();
	}
	

	@Override
	public float getRadius() {
		return data.getRadius();
	}
	

	@Override
	public Node getNode() {
		return this;
	}
	

	@Override
	public CTBotStatus getBotStatus() {
		// 2*pi*r * v(w) / (2*pi)
		// = r * v(w)
		float curSpeedL = lWheelRotation.getVelocity() * data.getWheelRadius();
		float curSpeedR = rWheelRotation.getVelocity() * data.getWheelRadius();
		
		return new CTBotStatus(curSpeedL, curSpeedR, lWheelRotation.getDesiredVelocity(),
				rWheelRotation.getDesiredVelocity());
	}
}
