package edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts;


/**
 * Abstract PID-controller implementation. Parameters can be set with
 * {@link #configure(float, float, float, float, float)}, the control-target can be set with
 * {@link #updateSetpoint(float)} and calculation is done in {@link #update()}. Feedback must be implemented in
 * {@link #getFeedback()}, output can be retrieved by overriding {@link #setOutput(float)} or calling
 * {@link #getLastOutput()}. (NOT thread-safe)
 * 
 * @author Gero
 */
public abstract class APid {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// Config
	private float	kp				= 0;
	private float	ki				= 0;
	private float	kd				= 0;
	
	private float	slewMax		= 0;
	private float	max			= 0;
	
	// State
	private float	lastErrorT0	= 0;
	private float	lastErrorT1	= 0;
	private float	lastErrorT2	= 0;
	private float	lastOutput	= 0;
//	private float	lastFeedback = 0;
	
	private float	setpoint		= 0;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see APid
	 */
	public APid() {
		
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see APid
	 * 
	 * @param ki
	 * @param kp
	 * @param kd
	 * @param max
	 * @param slewMax
	 */
	public void configure(float kp, float ki, float kd, float max, float slewMax) {
		this.kp = kp;
		this.ki = ki;
		this.kd = kd;
		
		this.max = max;
		this.slewMax = slewMax;
	}
	
//	private static int cnt = 0;
	/**
	 * @see APid
	 */
	public void update() {
		
		lastErrorT2 = lastErrorT1;
		lastErrorT1 = lastErrorT0;
		float rawFeedback = getFeedback();
//		final float feedback = (rawFeedback + lastFeedback)/2;
		final float feedback = rawFeedback;
//		lastFeedback = rawFeedback;
		lastErrorT0 = setpoint - feedback;

//		if (cnt % 4 == 0){
//			System.out.println("#######################");
//			System.out.println("setpoint: " + setpoint);
//			System.out.println("feedback: " + feedback);
////			System.out.println("lastfeed: " + lastFeedback);
//		}
//		cnt++;
				
		final float pError = lastErrorT0 - lastErrorT1;
		final float iError = lastErrorT1;
		final float dError = lastErrorT0 - 2 * lastErrorT1 + lastErrorT2;
		
		float newOutput = lastOutput + (kp * pError + ki * iError + kd * dError);
		
		// prevent large output steps
		if ((newOutput - lastOutput) > slewMax) {
			newOutput = lastOutput + slewMax;
		} else if ((lastOutput - newOutput) > slewMax) {
			newOutput = lastOutput - slewMax;
		}
		
		// prevent outputs greater than maximum
		// overload = 0;
		if (newOutput > max) {
			newOutput = max;
			// overload = 1;
		} else if (newOutput < -max){
			newOutput= -max;
		}
		
		// log.current = current;
		// log.pError = pError;
		// log.iError = iError;
		// log.dError = dError;
		lastOutput = newOutput;
		
		setOutput(newOutput);
	}
	

	/**
	 * Sets the new setpoint of this PID (Checks bounds: 0 - {@link #max})
	 * 
	 * @param newSetpoint
	 * @see APid
	 */
	public void updateSetpoint(float newSetpoint) {
		if (newSetpoint > max) {
			this.setpoint = max;
		} else if (newSetpoint < -max) {
			this.setpoint = -max;
		} else {
			this.setpoint = newSetpoint;
		}
	}
	

	/**
	 * Must be implemented by subclasses!
	 * 
	 * @return The feedback of the system to control
	 */
	protected abstract float getFeedback();
	

	/**
	 * Callback in {@link #update()}, may be implemented by subclasses to retrieve the output!
	 * 
	 * @param newOutput
	 * @see #getLastOutput()
	 */
	protected void setOutput(float newOutput) {
	}
	

	/**
	 * @return The last calculated output
	 */
	public float getLastOutput() {
		return lastOutput;
	}
}
