package edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts;

import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.data;
import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.mm;

import java.util.ArrayList;
import java.util.List;

import com.jme.bounding.BoundingBox;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.scene.Spatial;
import com.jme.scene.shape.Cylinder;
import com.jme.system.DisplaySystem;
import com.jmex.physics.DynamicPhysicsNode;
import com.jmex.physics.Joint;
import com.jmex.physics.RotationalJointAxis;
import com.jmex.physics.geometry.PhysicsCylinder;

import edu.dhbw.mannheim.tigers.sim.engine.model.ModelHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.IBotPartsOwner;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger.TigerBotModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.RoboCupMaterial;


/**
 * This {@link ABotModelPart}-implementation represents the dribbler-device of the {@link TigerBotModel}.
 * 
 * This models dimensions are based on the following assumptions:
 * <ol>
 * <li>The model has a total height
 * 
 * @author Gero
 */
public class Dribbler extends ABotModelPart {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
//	private final Logger						log									= Logger.getLogger(getClass());
	
	// Size
	private static final float				ROLLER_RADIUS						= 4;
	private static final float				ROLLER_LENGTH						= 55;
	private static final float				ROLLER_MASS							= 0.010f;	// [kg], needs some mass to be consistent!

	private static final float				BASE_PLATE_DEPTH					= mm(4);
	private static final float				ROLLER_OVER_BOTTOM				= mm(25.5f) + BASE_PLATE_DEPTH;
	/** front-border + guessed (because of concave roller) - {@link #ROLLER_RADIUS} */
	private static final float				ROLLER_DISTANCE_FROM_CENTER	= mm(66 + 5 - ROLLER_RADIUS);

	private static final float				ROLLER_ACC							= 0.15f;	// May not be greater, else the
	private static final float				ROLLER_RPM							= 5000;
	

	// Physicals
	private final DynamicPhysicsNode		dribbler;
	private final RotationalJointAxis	dribbleMotor;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see Dribbler
	 * @param owner
	 * @param display
	 * @param body
	 * @param bodyFromGroundDist
	 * @param com
	 */
	public Dribbler(IBotPartsOwner owner, DisplaySystem display, DynamicPhysicsNode body, float bodyFromGroundDist,
			Vector3f com) {
		super(owner, display, body, com);
		
		dribbler = space.createDynamicNode();
		
		// Size
		final float distFromGround = ROLLER_OVER_BOTTOM + data(bodyFromGroundDist);
		
		// Visuals
		Spatial vRoller = new Cylinder("vRoller", 10, 25, ROLLER_RADIUS, ROLLER_LENGTH, true);
		vRoller.setModelBound(new BoundingBox());
		vRoller.updateModelBound();
		
		ModelHelper.color(vRoller, display, ColorRGBA.lightGray.clone());
		
		dribbler.attachChild(vRoller);
		
		// Physics
		final PhysicsCylinder phRoller = dribbler.createCylinder("phRoller");
		phRoller.setMaterial(RoboCupMaterial.BOT_DRIBBLER);
		phRoller.setLocalScale(new Vector3f(ROLLER_RADIUS, ROLLER_RADIUS, ROLLER_LENGTH));
		dribbler.setMass(ROLLER_MASS);
		
		
		// Position
		dribbler.setLocalTranslation(ROLLER_DISTANCE_FROM_CENTER, distFromGround, 0);
		dribbler.getLocalTranslation().subtractLocal(com); // Offset
		owner.attachNode(dribbler);
		
		// Attach it to the bot
		final Joint dribblerPivot = space.createJoint();
		dribblerPivot.setSpring(0, 100);
//		OdeJoint odeJoint = (OdeJoint) dribblerPivot;
//		odeJoint.setERP(0);	//.2f);
//		odeJoint.setCFM(0);	//.0001f);
		dribbleMotor = dribblerPivot.createRotationalAxis();
		dribbleMotor.setDirection(Vector3f.UNIT_Z.clone());
		
		dribbleMotor.setAvailableAcceleration(ROLLER_ACC);
		
		dribblerPivot.attach(dribbler, body);
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @param rpm Wished rpm (max: {@link #ROLLER_RPM}: {@value #ROLLER_RPM})
	 */
	public void fire(float rpm) {
		if (rpm < 0) {
			rpm = 0;
		} else if (rpm > ROLLER_RPM) {
			rpm = ROLLER_RPM;
		}
		
		// float radV = rpm * FastMath.TWO_PI;
		
		dribbleMotor.setDesiredVelocity(-rpm); // Invert, as the dribbler is hung up the other way round
	}
	

	@Override
	public void doUpdate(long cycleNumber, float tpc) {
		
	}
	

	@Override
	public List<DynamicPhysicsNode> getPhysicNodes() {
		final ArrayList<DynamicPhysicsNode> result = new ArrayList<DynamicPhysicsNode>();
		result.add(dribbler);
		return result;
	}
}
