package edu.dhbw.mannheim.tigers.sim.engine.model.bot.tiger;


import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.MODEL_SCALE_FACTOR;
import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.data;
import static edu.dhbw.mannheim.tigers.sim.engine.model.UnitConversion.mm;

import java.io.IOException;
import java.text.DecimalFormat;

import com.jme.bounding.BoundingBox;
import com.jme.bounding.BoundingSphere;
import com.jme.math.FastMath;
import com.jme.math.Quaternion;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.renderer.Renderer;
import com.jme.scene.Node;
import com.jme.scene.Spatial;
import com.jme.scene.shape.Cylinder;
import com.jme.system.DisplaySystem;
import com.jme.util.export.JMEExporter;
import com.jme.util.export.JMEImporter;
import com.jme.util.export.Savable;
import com.jmex.physics.DynamicPhysicsNode;
import com.jmex.physics.Joint;
import com.jmex.physics.RotationalJointAxis;
import com.jmex.physics.geometry.PhysicsCylinder;
import com.jmex.physics.impl.ode.OdePhysicsSpace;

import edu.dhbw.mannheim.tigers.sim.engine.com.ERoboCupTeam;
import edu.dhbw.mannheim.tigers.sim.engine.model.ModelHelper;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.ABotModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.EBotModelType;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.Dribbler;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.IKickerOwner;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.IRSensor;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.KickMode;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.Lobber;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.Plunger;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.parts.SimpleCapModel;
import edu.dhbw.mannheim.tigers.sim.engine.model.data.bot.ITigerBotModelData;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.MaterialContactCallbackXT;
import edu.dhbw.mannheim.tigers.sim.engine.model.material.RoboCupMaterial;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelException;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelResourceException;


/**
 * The {@link ABotModel} implementation of the Tiger-Bot.
 * 
 * @author Gero, Marco
 */
public class TigerBotModel extends ABotModel implements ITigerBotControl, IKickerOwner {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long			serialVersionUID		= -5609996649395053527L;
	
	private final ITigerBotModelData	data;
	private final Node					ballsRoot;
	
	private static final float			MODEL_OBJECT_HEIGHT	= 150;
	
	// ID
	private final DecimalFormat		format24Digits			= new DecimalFormat("000000000000000000000000");
	

	// Parts
	private final DynamicPhysicsNode	body;
	/** Radius of the wheel [mm] */
	private final float					wRadius;
	
	/** 1st Wheel, front-right */
	private final DynamicPhysicsNode	wheel1;
	/** 2nd Wheel, front-left */
	private final DynamicPhysicsNode	wheel2;
	/** 3rd Wheel, rear-left */
	private final DynamicPhysicsNode	wheel3;
	/** 4th Wheel, rear-left */
	private final DynamicPhysicsNode	wheel4;
	
	/** 1 front right */
	private RotationalJointAxis		axis1;
	/** 2 rear right */
	private RotationalJointAxis		axis2;
	/** 3 front left */
	private RotationalJointAxis		axis3;
	/** 4 rear left */
	private RotationalJointAxis		axis4;
	
	private final ColorRGBA				color;
	
	private TigerMotorController		motorController;
	private Plunger						plunger;
	private Lobber							lobber;
	private Dribbler						dribbler;
	private IRSensor						irSensor;
	
	private SimpleCapModel				cap;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see TigerBotModel
	 * @param space
	 * @param display
	 * @param newData
	 * @param balls
	 * @param color May be null!
	 * @throws SimModelResourceException
	 * @throws SimModelException
	 */
	public TigerBotModel(OdePhysicsSpace space, DisplaySystem display, ITigerBotModelData newData, Node balls,
			ColorRGBA color, ERoboCupTeam team)
			throws SimModelResourceException, SimModelException {
		super(space, display, EBotModelType.TIGER_MODEL, newData.getName(), team);
		
		this.data = newData;
		this.ballsRoot = balls;
		this.color = color;
		
		wRadius = data.getWheelRadius();
		
		addPhysicNode(body = space.createDynamicNode());
		
		addPhysicNode(wheel1 = space.createDynamicNode());
		addPhysicNode(wheel2 = space.createDynamicNode());
		addPhysicNode(wheel3 = space.createDynamicNode());
		addPhysicNode(wheel4 = space.createDynamicNode());
		
		createPhysicalNode();
		
		makePickable(body);
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void createPhysicalNode() throws SimModelResourceException, SimModelException {
		
		/*
		 * Here we want to build a "simple" Tiger-bot consisting of the following parts:
		 * - a cylindric body
		 * - four wheels
		 * - the kicking-devices
		 * (- sensors)
		 */

		// ##### The Body
		final float bodyRadius = data.getRadius();
		final float bodyHeight = data.getHeight();
		final float bodyToGroundDist = data.getBodyToGroundDist();
		
		body.setName("Tiger-Bot [body]");
		
		createBody(bodyRadius, bodyHeight);
		
		// Location
		// Lift it a bit above the floor
		body.setLocalTranslation(0, data(bodyHeight / 2 + bodyToGroundDist), 0);
		
		// Rotation
		body.setLocalRotation(ModelHelper.x90.clone()); // and turn it up!
		
		this.attachChild(body);
		
		// Define the Center of Mass (local offset to center)
		final Vector3f com = new Vector3f(-mm(18), -mm(52), 0);
		// Rotation of x90 (body to world)
		final Vector3f comBody = new Vector3f(com.x, com.z, -com.y);
		// Remember to adjust length, as model will be scaled down later on
		comBody.multLocal(1 / MODEL_SCALE_FACTOR.floatValue());
		

		// ##### Wheels
		final float frontWheelAngleDist = data.getFrontWheelAngleDist();
		final float frontRightAngle = 90 - frontWheelAngleDist / 2;
		final float frontLeftAngle = 90 + frontWheelAngleDist / 2;
		
		// First wheel
		wheel1.setName("Tiger Wheel [1fr]");
		adjustWheel(wheel1, "1fr", rad(frontRightAngle), com);
		this.attachChild(wheel1);
		
		// MaterialContactCallbackXT.getInstance(null).wheelUnderTest = firstWheel;
		
		// Second wheel
		wheel2.setName("Tiger Wheel [2fl]");
		adjustWheel(wheel2, "2fl", rad(frontLeftAngle), com);
		this.attachChild(wheel2);
		
		// Third wheel
		wheel3.setName("Tiger Wheel [3rl]);");
		adjustWheel(wheel3, "3rl", rad(-135), com);
		this.attachChild(wheel3);
		
		wheel4.setName("Tiger Wheel [4rr]");
		adjustWheel(wheel4, "4rr", rad(-45), com);
		this.attachChild(wheel4);
		

		// Now we add the joints...
		axis1 = attachWheel(wheel1, rad(frontRightAngle));
		axis2 = attachWheel(wheel2, rad(frontLeftAngle));
		axis3 = attachWheel(wheel3, rad(-135));
		axis4 = attachWheel(wheel4, rad(-45));
		
		// Motor
		motorController = new TigerMotorController(getID(), axis1, axis2, axis3, axis4, wRadius);
		

		// ##### Kicking-device
		cap = new SimpleCapModel(200, 0.65f);
		addModelPart(cap);
		
		addBotModelPart(plunger = new Plunger(this, this, display, body, data.getBodyToGroundDist(), data.getRadius(),
				com));
		addBotModelPart(lobber = new Lobber(this, this, display, body, data.getHeight(), data.getBodyToGroundDist(), com));
		addBotModelPart(dribbler = new Dribbler(this, display, body, data.getBodyToGroundDist(), com));
		
		// Apply IR-sensor: A bit in front of the bots edge, a bit above the ground...
		// As the body has already been turned by x90:
		Vector3f offset = new Vector3f(66 + 5, -35f, 75 - (13 + 8.5f));
		irSensor = new IRSensor(body, offset, new Vector3f(0, 70, 0), ballsRoot, 0.025f);
		addModelPart(irSensor);
		

		// ##### Finish
		// Change CoM beeeefore setting the mass... makes it actually happen! ;-)
		// Vector3f comDiff = new Vector3f(0, 0, 45);
		body.setCenterOfMass(comBody);
		
//		 // Debug CoM
//		 Sphere vCom = new Sphere("CoM", new Vector3f(), 10, 10, 25f);
//		 ColorRGBA comColor = ColorRGBA.cyan.clone();
//		 ModelHelper.color(vCom, display, comColor);
//		
//		 vCom.setModelBound(new BoundingSphere());
//		 vCom.updateModelBound();
//		 vCom.setLocalTranslation(comBody); // Down, as the Z-axis points into the ground
//		 vCom.setCullHint(CullHint.Never);
//		 body.attachChild(vCom);
		
		// System.out.println("CoM: " + body.getCenterOfMass(null));
		
		// Aaaafter adjusting the CoM!
		body.setMass(data.getMass());
		

		// Finish...
		this.setRenderQueueMode(Renderer.QUEUE_OPAQUE);
	}
	

	private void createBody(float bodyRadius, float bodyHeight) {
		
		Spatial vBody = null;
		
		if (data.getHeight() > 0) {
			try {
				
				// Visual
				vBody = ModelHelper.loadModel("bot/", "TIGER_v2c3.obj");
				
				// Scale height (as the loaded model normally meets the league-criteria (150x180), but
				// we have wheels etc to put under it...)
				float objectTargetHeight = data.getHeight() - data.getBodyToGroundDist();
				float factor = objectTargetHeight / MODEL_OBJECT_HEIGHT;
				vBody.getLocalScale().multLocal(new Vector3f(1, factor, 1));
				
				// Adjust imported 'world' coordinate system
				vBody.setLocalRotation(ModelHelper.xDeg(-90));
				vBody.getLocalRotation().multLocal(ModelHelper.y180.clone());
				
				body.attachChild(vBody);
				
				// Physics
				final Vector3f oldScale = new Vector3f(body.getLocalScale());
				body.setLocalScale(0.001f);
				body.generatePhysicsGeometry(true);
				body.setLocalScale(oldScale);
				
			} catch (IOException err) {
				log.debug("Unable to load model!" + err.getMessage());
			}
		}
		
		if (vBody == null) {
			
			// Visual
			vBody = new Cylinder("Tiger-Bot [body] Vis", 5, 25, bodyRadius, bodyHeight, true);
			vBody.setModelBound(new BoundingBox());
			vBody.updateModelBound();
			
			// Physic
			final PhysicsCylinder phBody = body.createCylinder("Bot-Body");
			phBody.setLocalScale(new Vector3f(bodyRadius, bodyRadius, bodyHeight));
			
			body.attachChild(vBody);
		}

		body.setMaterial(RoboCupMaterial.BOT_BODY);
		
		// Visual Common
		ModelHelper.color(vBody, display, color); // Color!
	}
	

	/**
	 * @param wheel
	 * @param whichWheel
	 * @param angle IN RADIANS!
	 * @param com The offset for the Center of Mass (from the center of the object; needs to be regarded!)
	 */
	private void adjustWheel(DynamicPhysicsNode wheel, String whichWheel, final float angle, Vector3f com) {
		
		// Size
		final float wheelWidth = data.getWheelWidth();
		final float bodyRadius = data.getRadius();
		final float wheelToBodyDist = data.getBodyToWheelDist();
		

		// Visual
		final Cylinder visualWheel = new Cylinder(whichWheel + "visualWheel", 5, 15, wRadius, wheelWidth, true);
		visualWheel.setModelBound(new BoundingSphere());
		visualWheel.updateModelBound();
		
		// Colorize
		ModelHelper.color(visualWheel, display, ColorRGBA.lightGray);
		// ModelHelper.makeTransparent(wheelBase, display);
		
		wheel.attachChild(visualWheel);
		
		// Physic
		final PhysicsCylinder cylinder = wheel.createCylinder(whichWheel + "physicalWheel");
		cylinder.setMaterial(RoboCupMaterial.BOT_WHEEL);
		cylinder.setLocalScale(new Vector3f(wRadius, wRadius, wheelWidth));
		wheel.setMass(0.060f); // 10g TODO Parameterize! (before: 0.010 17.2.2011)
		

		// Position
		final float dist = data(bodyRadius + wheelToBodyDist);
		wheel.setLocalTranslation(FastMath.sin(angle) * dist, data(wRadius), FastMath.cos(angle) * dist);
		wheel.getLocalTranslation().subtractLocal(com);
		
		// Rotation
		Quaternion rotQuat = new Quaternion();
		rotQuat.fromAngleAxis(angle, Vector3f.UNIT_Y);
		wheel.setLocalRotation(rotQuat);
	}
	

	/**
	 * @param wheel
	 * @param angle IN RADIANS!
	 * @return
	 */
	private RotationalJointAxis attachWheel(DynamicPhysicsNode wheel, float angle) {
		
		final Joint joint = space.createJoint();
//		joint.setSpring(0.000001f, 1.0f); // best: (0.0065f, 0.05f);
//		OdeJoint odeJoint = (OdeJoint) joint;
//		odeJoint.setCFM(0.5f); // 100f);
//		odeJoint.setERP(0.8f);
		final RotationalJointAxis rotationalAxis = joint.createRotationalAxis(); // Add some freedom
																											// to the joint
		
		joint.attach(wheel, body);
		
		wheel.setUserData(WheelAxis.WHEEL_AXIS, new WheelAxis(rotationalAxis));
		
		// Determine the direction of the rotation-axis
		float z = FastMath.cos(angle);
		float x = FastMath.sin(angle);
		Vector3f vec = new Vector3f(x, 0, z).normalizeLocal();
		rotationalAxis.setDirection(vec);
		
		return rotationalAxis;
	}
	

	// --------------------------------------------------------------------------
	// --- Shoot ----------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void lob(int percent, KickMode mode) {
		lobber.fire(percent, mode);
	}
	

	@Override
	public boolean isBallDetected() {
		return irSensor.isBlocked();
	}
	

	/**
	 * @param rpm
	 */
	@Override
	public void dribble(int rpm) {
		dribbler.fire(rpm);
	}
	

	@Override
	public void charge(boolean charge) {
		cap.charge(charge);
	}
	

	@Override
	public void kickStraight(int percent, KickMode mode) {
		plunger.fire(percent, mode);
	}
	

	// --------------------------------------------------------------------------
	// --- ABotModel ------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void clearVelocity() {
		super.clearVelocity();
		
		axis1.setDesiredVelocity(0);
		axis2.setDesiredVelocity(0);
		axis3.setDesiredVelocity(0);
		axis4.setDesiredVelocity(0);
	}
	

	// --------------------------------------------------------------------------
	// --- IUpdateable ----------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void doUpdate(long cycleNumber, float tpc) {
		
		// Motors (set speed etc.)
		motorController.doUpdate(cycleNumber, tpc);
		
		// Position
		body.addForce(new Vector3f(0, -500, 0));
		body.updateGeometricState(0f, true);
		notifyObservers(getCurrentBotStatus(body.getWorldTranslation(), body.getWorldRotation()));
	}
	

	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public SimpleCapModel getCapModel() {
		return cap;
	}
	

	@Override
	public float getHeight() {
		return data.getHeight() + data.getBodyToGroundDist();
	}
	

	@Override
	public float getRadius() {
		return data.getRadius();
	}
	

	@Override
	public Node getNode() {
		return body;
	}
	

	@Override
	public TigerBotStatus getBotStatus() {
		return new TigerBotStatus(motorController.calcTigerMove(), plunger.getCurrentCharge(), plunger.isCapLoading());
	}
	

	@Override
	public String getCpuId() {
		return format24Digits.format(getID().getIntValue());
	}
	

	@Override
	public ITigerMotorControl getMotorControl() {
		return motorController;
	}
	
	
	/**
	 * TODO Find a better way to populate wheels orientation to {@link MaterialContactCallbackXT}
	 * @author Gero
	 */
	public class WheelAxis implements Savable {
		
		public static final String			WHEEL_AXIS	= "WHEEL_AXIS";
		
		public final RotationalJointAxis	axis;
		
		
		public WheelAxis(RotationalJointAxis axis) {
			this.axis = axis;
		}
		

		@Override
		public void write(JMEExporter ex) throws IOException {
			throw new UnsupportedOperationException();
		}
		

		@Override
		public void read(JMEImporter im) throws IOException {
			throw new UnsupportedOperationException();
		}
		

		@Override
		public Class<?> getClassTag() {
			return WheelAxis.class;
		}
	}
}
