package edu.dhbw.mannheim.tigers.sim.engine.modes.display;

import java.util.ArrayList;
import java.util.List;

import com.jme.scene.Node;

import edu.dhbw.mannheim.tigers.sim.engine.ASimEngineState;
import edu.dhbw.mannheim.tigers.sim.engine.ESimEngineState;
import edu.dhbw.mannheim.tigers.sim.engine.ICore;
import edu.dhbw.mannheim.tigers.sim.engine.adapter.DisplayModeAdapter;
import edu.dhbw.mannheim.tigers.sim.engine.com.visualization.IVisualizationDataInterpreter;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.ASimDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.BasicInputDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.BasicInputDelegate.InputType;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.FreeCameraDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.ScreenshotDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.SimPhysicsDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.TigersBannerDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.VideoCaptureDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.delegates.VisualizationDataDelegate;
import edu.dhbw.mannheim.tigers.sim.engine.model.ball.IBallModelControl;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.IBasicBotModelControl;
import edu.dhbw.mannheim.tigers.sim.engine.modes.ASimEngineMode;
import edu.dhbw.mannheim.tigers.sim.engine.modes.ESimEngineMode;
import edu.dhbw.mannheim.tigers.sim.engine.modes.IMatchControlProvider;
import edu.dhbw.mannheim.tigers.sim.engine.modes.IVisDataInterpreterProvider;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelException;
import edu.dhbw.mannheim.tigers.sim.exceptions.SimModelResourceException;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.IMatchModeProperties;


/**
 * {@link ESimEngineMode#DISPLAY}
 * 
 * @author Gero
 */
public class DisplayMode extends ASimEngineMode implements IVisDataInterpreterProvider, IMatchControlProvider {
   
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// Properties
	private final IMatchModeProperties	props;
	

	// Delegates
	private SimPhysicsDelegate				physicsDelegate;
	private FreeCameraDelegate				camDelegate;
	private TigersBannerDelegate			skyBoxDelegate;
	private BasicInputDelegate				basicInputDelegate;
	private ScreenshotDelegate				screenshotDelegate;
	private VisualizationDataDelegate	visualizationDataDelegate;
	@SuppressWarnings("unused")	// Needs no explicit update!
	private VideoCaptureDelegate			videoDelegate;
  
   
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
   /**
     * @see DisplayMode
     * @param core
     */
   public DisplayMode(ICore core) {
      super(core, ESimEngineMode.DISPLAY);
      
      props = ISettingsProvider.INSTANCE.getSimProperties().getMatchModeProperties();
   }
   
   
	// --------------------------------------------------------------------------
	// --- current ASimEngineState ----------------------------------------------
	// --------------------------------------------------------------------------
   @Override
   public void changeStateTo(ESimEngineState newState) {
      
      switch (newState) {
         case PAUSED_SIMULATION:
            currentState = paused;
            break;
         
         case RUNNING:
            currentState = running;
            break;
         
         default:
            super.changeStateTo(newState);
            break;
      }
   }
   
   private ASimEngineState paused   = new ASimEngineState() {
                                       
      public void update(float interpolation, float tpc, float cps) {
         camDelegate.update(interpolation, tpc, cps);
         skyBoxDelegate.update(interpolation, tpc, cps);
         basicInputDelegate.update(interpolation, tpc, cps);
         screenshotDelegate.update(interpolation, tpc, cps);
      };
   };
   
   private ASimEngineState running  = new ASimEngineState() {
      
      public void update(float interpolation, float tpc, float cps) {
         physicsDelegate.update(interpolation, tpc, cps);
         camDelegate.update(interpolation, tpc, cps);
         skyBoxDelegate.update(interpolation, tpc, cps);
         displayDelegate.update(interpolation, tpc, cps);
         basicInputDelegate.update(interpolation, tpc, cps);
         screenshotDelegate.update(interpolation, tpc, cps);
         visualizationDataDelegate.update(interpolation, tpc, cps);
      };
   };
   
   
	// --------------------------------------------------------------------------
	// --- delegates ------------------------------------------------------------
	// --------------------------------------------------------------------------
   @Override
   public void setupDelegates() {
      addDelegate(physicsDelegate = new SimPhysicsDelegate(core));
      addDelegate(camDelegate = new FreeCameraDelegate(core));
      addDelegate(skyBoxDelegate = new TigersBannerDelegate(core));
      addDelegate(basicInputDelegate = new BasicInputDelegate(core, InputType.FIRST_PERSON));
      addDelegate(screenshotDelegate = new ScreenshotDelegate(core));
      addDelegate(visualizationDataDelegate = new VisualizationDataDelegate(core, this));
      addDelegate(displayDelegate);
      addDelegate(videoDelegate = new VideoCaptureDelegate(core));
   }
   
   
	// --------------------------------------------------------------------------
	// --- adapters -------------------------------------------------------------
	// --------------------------------------------------------------------------
   @Override
   public void setupAdapters() {
      addAdapter(new DisplayModeAdapter(core, this));
   }
   
   
	// --------------------------------------------------------------------------
	// --- DisplayDelegate ------------------------------------------------------
	// --------------------------------------------------------------------------
   private ASimDelegate displayDelegate = new ASimDelegate(core) {

      public void initSimulation() {
         
         // The field
         try {
            Node field = factory.createField(core.getPhysicsSpace(), core.getDisplay(), props.getFieldModelType());
            core.getStatics().attachChild(field);
         } catch (SimModelException exc) {
            log.error("Error while creating the RoboCupField", exc);
            core.stop();
         } catch (SimModelResourceException exc) {
            log.error("Error while loading model data!", exc);
         }
      };
   };

   
	// --------------------------------------------------------------------------
	// --- IVisDataInterpreterProvider, IMatchControlProvider -------------------
	// --------------------------------------------------------------------------
   @Override
   public IVisualizationDataInterpreter getVisualizationDataInterpreter() {
      return visualizationDataDelegate;
   }
   
   
   @Override
   public IBallModelControl getBallControl()
   {
   	return null;	// No ball available
   }
   
   
   @Override
   public List<? extends IBasicBotModelControl> getBotControls()
   {
   	return new ArrayList<IBasicBotModelControl>();	// No bots here
   }
}
