package edu.dhbw.mannheim.tigers.sim.engine.plugins.cam;

import java.util.List;

import org.apache.commons.configuration.SubnodeConfiguration;

import edu.dhbw.mannheim.tigers.sim.engine.model.TimedModelLocation;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.PluginID;


/**
 * Defines the default cam-output builder with two cameras.
 * Cam 0 is left (-Zsim , -Xssl), Cam 1 is right (+Zsim, +Xssl)
 * 
 * @see ACamOutputPlugin
 * @author Gero
 */
public class DefaultTwoCamOutput extends ACamOutputPlugin {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	public static final PluginID	ID						= new PluginID("DEFAULT_TWO_CAM");
	
	private static final int		DEFAULT_OVERLAP_X	= 500;
	
	private float						offset;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see DefaultTwoCamOutput
	 * @param config
	 */
	public DefaultTwoCamOutput(SubnodeConfiguration config) {
		super(config);
		
		this.offset = config.getFloat("overlapx", DEFAULT_OVERLAP_X) / 2;
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	protected void createCameras(List<ACam> cameras) {
		final float length = getRobocupFieldData().getFieldLength() + getRobocupFieldData().getFieldBorderWidthL() * 2;
		final float width = getRobocupFieldData().getFieldWidth() + getRobocupFieldData().getFieldBorderWidthW() * 2;
		
		// Cam 0, left
		cameras.add(new TwoCam(0, TwoCam.Side.LEFT, length, width, offset));
		
		// Cam 1, right
		cameras.add(new TwoCam(1, TwoCam.Side.RIGHT, length, width, offset));
	}
	

	@Override
	public void checkConfigConsistency(SubnodeConfiguration config) {
		if (config.getProperty("properties.overlapx") == null) {
			config.setProperty("properties.overlapx", DEFAULT_OVERLAP_X);
		}
	}
	
	
	/**
	 * @return the offset
	 */
	protected float getOffset() {
		return this.offset;
	}
	

	@Override
	public PluginID getID() {
		return ID;
	}
	
	
	protected static class TwoCam extends ACam {
		
		protected enum Side {
			LEFT,
			RIGHT;
		}
		
		private final Side	side;
		private final float	halfLength;
		private final float	halfWidth;
		private final float	offset;
		
		
		public TwoCam(int id, Side side, float length, float width, float offset) {
			super(id);
			
			this.side = side;
			this.halfLength = length / 2;
			this.halfWidth = width / 2;
			this.offset = offset;
		}
		

		@Override
		protected boolean isObjVisible(TimedModelLocation obj) {
			// Check for field-bounds
			if (isOnField(obj, halfLength, halfWidth)) {
				
				if (side == Side.LEFT) {
					if (obj.getZmm() < 0 + offset) {
						return true;
					}
				} else {
					if (obj.getZmm() > 0 - offset) {
						return true;
					}
				}
			}
			
			return false;
		}
		

		protected boolean isOnField(TimedModelLocation obj, float halfLength, float halfWidth) {
			return (obj.getXmm() < halfWidth && obj.getXmm() > -halfWidth)
					&& (obj.getZmm() < halfLength && obj.getZmm() > -halfLength);
		}
	}
}
