package edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import com.jme.scene.Node;
import com.jme.system.DisplaySystem;

import edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization.SimVisMessages.SimVisData;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization.SimVisMessages.SimVisData.Poi;
import edu.dhbw.mannheim.tigers.sim.engine.plugins.visualization.SimVisMessages.SimVisData.PoiList;


/**
 * Abstract Drawer class that provides some methods for managing {@link Poi}, but let the
 * subclass draw everything.
 * The idea is, that incoming information is simply passed by calling {@link #displayInfo(Map)}.
 * The mapping between {@link Poi} and its representing {@link AVisualObject} takes place in this
 * class, while actual creation and drawing is delegated to any subclasses.
 * 
 * @author Gero
 */
public abstract class APOIDrawer implements IDrawer {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private final HashMap<Integer, List<AVisualObject>>	predictedPOIs	= new HashMap<Integer, List<AVisualObject>>();
	
	protected final Node												localRoot;
	protected final int												renderQueueMode;
	protected final DisplaySystem									display;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see APOIDrawer
	 * @param root
	 * @param renderQueueMode
	 * @param display
	 */
	public APOIDrawer(Node root, int renderQueueMode, DisplaySystem display) {
		
		this.renderQueueMode = renderQueueMode;
		this.display = display;
		
		this.localRoot = new Node();
		this.localRoot.setRenderQueueMode(renderQueueMode);
		root.attachChild(localRoot);
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public abstract void displayInfo(SimVisData data);
	

	/**
	 * Draws info with the given {@link Poi}s, and stores the resulting {@link AVisualObject}s in the
	 * drawers store
	 * 
	 * @param pois
	 */
	protected void doDisplayInfo(List<PoiList> infos) {
		
		for (PoiList info : infos) {
			List<AVisualObject> visuals = predictedPOIs.get(identify(info));
			
			if (visuals == null) {
				visuals = new ArrayList<AVisualObject>(); // Nothing there, we have to draw every thing
				// from the scratch
			}
			drawInfo(info.getPoisList(), visuals);
			
			predictedPOIs.put(identify(info), visuals);
		}
	}
	

	/**
	 * May be overridden by subclasses!
	 * 
	 * @param info
	 * @return The unique id of this {@link PoiList}
	 */
	protected Integer identify(PoiList info) {
		return info.getId();
	}
	

	/**
	 * Draws a list of {@link AVisualObject}s with the given {@link Poi}s associated with each other
	 * using {@link #createSinglePoi(Poi)} and {@link #drawSinglePOI(AVisualObject, Poi)}. Tries to
	 * reuse already created objects!
	 * 
	 * @param poi
	 * @return
	 */
	protected void drawInfo(final List<Poi> pois, final List<AVisualObject> store) {
		
		final List<AVisualObject> createdVisuals = new ArrayList<AVisualObject>();
		
		final Iterator<AVisualObject> it = store.iterator();
		for (Poi poi : pois) {
			
			AVisualObject visual = null;
			if (it.hasNext()) {
				visual = it.next(); // There is another visual we can reuse for our purposes
			} else {
				visual = createSinglePoi(poi); // We have to create a new one
				
				createdVisuals.add(visual);
			}
			
			visual.drawSingle(poi);
			visual.updateRenderState();
		}
		
		// If there had already been _more visuals then we needed to draw our pois, we have to
		// undraw the now invalid old, odd visuals
		while (it.hasNext()) {
			AVisualObject oddVisual = it.next();
			oddVisual.undraw();
			it.remove();
		}
		
		// Other way round: There had been _less visuals then we needed for our pois, so we had to
		// create some, and have to register them now!
		if (!createdVisuals.isEmpty()) {
			store.addAll(createdVisuals);
		}
	}
	

	/**
	 * Actually _creates the {@link Poi}!
	 * 
	 * @param poi
	 * @return
	 */
	protected abstract AVisualObject createSinglePoi(Poi poi);
	

	@Override
	public void undisplayInfo() {
		if (!predictedPOIs.isEmpty()) {
			for (Entry<Integer, List<AVisualObject>> entry : predictedPOIs.entrySet()) {
				undrawInfo(entry.getValue());
			}
			predictedPOIs.clear();
		}
	}
	

	/**
	 * Simply removes the given list of {@link AVisualObject}s from their parents (but it is <b>not</b> removed from
	 * 'visuals'!)
	 */
	protected void undrawInfo(final List<AVisualObject> visuals) {
		for (AVisualObject visual : visuals) {
			visual.undraw();
		}
	}
}
