package edu.dhbw.mannheim.tigers.sim.gui.swing.settings;


import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import edu.dhbw.mannheim.tigers.sim.controller.IInputController;
import edu.dhbw.mannheim.tigers.sim.engine.model.EModelColorMode;
import edu.dhbw.mannheim.tigers.sim.engine.model.ModelDataProvider;
import edu.dhbw.mannheim.tigers.sim.engine.model.bot.EBotModelType;
import edu.dhbw.mannheim.tigers.sim.settings.ISettingsProvider;
import edu.dhbw.mannheim.tigers.sim.settings.data.IDataFactory;
import edu.dhbw.mannheim.tigers.sim.settings.data.IMatchModeProperties;
import edu.dhbw.mannheim.tigers.sim.settings.data.ITeamProperties;


/**
 * Settingspanel responsible for the settings for a RoboCup-match
 * 
 * @author Malte, Gero
 */
public class MatchSettingsPanel extends ASettingsPanel {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final long		serialVersionUID	= 7272584077192415452L;
	
	private IMatchModeProperties	properties;
	
	// Match Properties
	private JPanel						matchPropertiesPanel;
	
	private JComboBox					textFieldModel;
	private JComboBox					modelColorMode;
	
	private final TeamPanel			teamPanelBlue;
	private final TeamPanel			teamPanelYellow;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @see MatchSettingsPanel
	 * @param inputController
	 */
	public MatchSettingsPanel(IInputController inputController) {
		super("Match", inputController);
		
		teamPanelYellow = new TeamPanel(inputController, ETeam.YELLOW);
		teamPanelBlue = new TeamPanel(inputController, ETeam.BLUE);
	}
	

	@Override
	protected void loadSettings() {
		properties = ISettingsProvider.INSTANCE.getSimProperties().getMatchModeProperties();
		
		teamPanelYellow.loadSettings();
		teamPanelBlue.loadSettings();
	}
	

	@Override
	protected void createGui() {
		// Panels
		matchPropertiesPanel = createDefaultSettingsPanel("Match Properties");
		
		matchPropertiesPanel.add(new JLabel("Field model:"));
		textFieldModel = new JComboBox();
		for (String type : ModelDataProvider.getInstance().getAvailableFieldTypes()) {
			textFieldModel.addItem(type);
		}
		matchPropertiesPanel.add(textFieldModel);
		

		matchPropertiesPanel.add(new JLabel("Model color mode:"));
		modelColorMode = new JComboBox();
		for (EModelColorMode mode : EModelColorMode.values()) {
			modelColorMode.addItem(mode.toString());
		}
		matchPropertiesPanel.add(modelColorMode);
		

		add(matchPropertiesPanel, "grow");
		

		// Team Panels
		teamPanelYellow.createGui();
		add(teamPanelYellow.teamPanel, "grow");
		
		teamPanelBlue.createGui();
		add(teamPanelBlue.teamPanel, "grow");
		

		addButtonsPanel();
	}
	

	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void defaults() {
		fillFields(IDataFactory.eINSTANCE.createMatchModeProperties());
		
		teamPanelYellow.defaults();
		teamPanelBlue.defaults();
	}
	

	private void fillFields(IMatchModeProperties matchProperties) {
		textFieldModel.setSelectedItem(String.valueOf(matchProperties.getFieldModelType()));
		modelColorMode.setSelectedItem(String.valueOf(matchProperties.getModelColorMode()));
	}
	

	@Override
	public void revert() {
		fillFields(properties);
		
		teamPanelYellow.revert();
		teamPanelBlue.revert();
	}
	

	@Override
	public void apply() {
		properties.setFieldModelType(String.valueOf(textFieldModel.getSelectedItem()));
		properties.setModelColorMode(String.valueOf(modelColorMode.getSelectedItem()));
		
		teamPanelYellow.apply();
		teamPanelBlue.apply();
	}
	

	@Override
	public boolean checkInput() {
		// Nothing to check here!
		return true && teamPanelYellow.checkInput() && teamPanelBlue.checkInput();
	}
	
	private enum ETeam {
		BLUE,
		YELLOW;
	}
	
	private class TeamPanel extends ASettingsPanel {
		private static final long	serialVersionUID	= 3462170382065082563L;
		
		private final ETeam			team;
		
		private ITeamProperties		props;
		
		private JPanel					teamPanel;
		
		private JComboBox				textNumberOfBots;
		private JTextField			textPort;
		private JTextField			textIDs;
		private JComboBox				textBotModel;
		
		private JTextField			mcastInterface;
		private JTextField			mcastDiscoverGroup;
		private JTextField			mcastServerPort;
		private JTextField			mcastListenPort;
		
		
		public TeamPanel(IInputController inputController, ETeam team) {
			super("Team " + team, inputController);
			
			this.team = team;
		}
		

		@Override
		protected void loadSettings() {
			IMatchModeProperties matchProps = ISettingsProvider.INSTANCE.getSimProperties().getMatchModeProperties();
			
			switch (team) {
				case BLUE:
					props = matchProps.getTeamBlue();
					break;
				
				case YELLOW:
					props = matchProps.getTeamYellow();
					break;
			}
		}
		

		@Override
		protected void createGui() {
			teamPanel = createDefaultSettingsPanel(title);
			
			teamPanel.add(new JLabel("Number of bots:"));
			teamPanel.add(textNumberOfBots = new JComboBox());
			for (int number = 0; number <= 5; number++) {
				textNumberOfBots.addItem(number);
			}
			
			teamPanel.add(new JLabel("Bots IDs:"));
			teamPanel.add(textIDs = new JTextField());
			
			teamPanel.add(new JLabel("First port for team:"));
			teamPanel.add(textPort = new JTextField());
			
			teamPanel.add(new JLabel("Bot model for team:"));
			textBotModel = new JComboBox();
			for (EBotModelType e : EBotModelType.values()) {
				textBotModel.addItem(e.toString());
			}
			teamPanel.add(textBotModel);

			
			teamPanel.add(new JLabel("Multicast interface:"));
			teamPanel.add(mcastInterface = new JTextField());
			
			teamPanel.add(new JLabel("Multicast discover group:"));
			teamPanel.add(mcastDiscoverGroup = new JTextField());
			
			teamPanel.add(new JLabel("Multicast server port:"));
			teamPanel.add(mcastServerPort = new JTextField());
			
			teamPanel.add(new JLabel("Multicast listen port:"));
			teamPanel.add(mcastListenPort = new JTextField());
		}
		

		@Override
		public void defaults() {
			ITeamProperties teamProps = IDataFactory.eINSTANCE.createTeamProperties();
			
			switch (team) {
				case BLUE:
					teamProps.setBotIDs("6,7,8,9,10");
					teamProps.setFirstBotInputPort(10100);
					teamProps.setMcastDiscoverGroup("225.42.42.42");
					teamProps.setMcastListenPort(10141);
					teamProps.setMcastServerPort(10140);
					break;
				
				case YELLOW:
					teamProps.setBotIDs("1,2,3,4,5");
					teamProps.setFirstBotInputPort(10000);
					teamProps.setMcastDiscoverGroup("225.42.42.42");
					teamProps.setMcastListenPort(10041);
					teamProps.setMcastServerPort(10040);
					break;
			}
			
			fillFields(teamProps);
		}
		

		private void fillFields(ITeamProperties teamProperties) {
			textPort.setText(String.valueOf(teamProperties.getFirstBotInputPort()));
			textNumberOfBots.setSelectedItem(teamProperties.getNumberOfBots());
			textBotModel.setSelectedItem(String.valueOf(teamProperties.getBotModelType()));
			textIDs.setText(String.valueOf(teamProperties.getBotIDs()));
			
			mcastInterface.setText(teamProperties.getMcastInterface());
			mcastDiscoverGroup.setText(teamProperties.getMcastDiscoverGroup());
			mcastServerPort.setText(String.valueOf(teamProperties.getMcastServerPort()));
			mcastListenPort.setText(String.valueOf(teamProperties.getMcastListenPort()));
		}
		

		@Override
		public void revert() {
			fillFields(props);
		}
		

		@Override
		public boolean checkInput() {
			return true;
		}
		

		@Override
		public void apply() {
			props.setBotModelType(String.valueOf(textBotModel.getSelectedItem()));
			props.setFirstBotInputPort(Integer.parseInt((textPort.getText())));
			props.setNumberOfBots(Integer.parseInt(String.valueOf(textNumberOfBots.getSelectedItem())));
			props.setBotIDs(textIDs.getText());
			
			props.setMcastInterface(mcastInterface.getText());
			props.setMcastDiscoverGroup(mcastDiscoverGroup.getText());
			props.setMcastListenPort(Integer.valueOf(mcastListenPort.getText()));
			props.setMcastServerPort(Integer.valueOf(mcastServerPort.getText()));
		}
	}
}
