package edu.dhbw.mannheim.tigers.sim.util.activeobserver;


import java.util.LinkedList;


/**
 * {@link AActiveObserver}
 * 
 * @param <S> Observable type (Subject)
 * @param <O> Observer type
 * @param <E> Event type
 */
public class ActiveObservable<S extends IActiveObservable<S, O, E>, O extends IActiveObserver<S, O, E>, E> implements
		IActiveObservable<S, O, E> {
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private LinkedList<O>	registeredObservers	= null;
	private final Object		listSync					= new Object();
	
	private E					lastEvent;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Adds the given observer
	 * 
	 * @param observers: The {@link AActiveObserver} that should listen to this observable.
	 */
	public ActiveObservable(E initEvent, O... observers) {
		registeredObservers = new LinkedList<O>();
		
		this.lastEvent = initEvent;
		
		addObservers(observers);
	}
	

	/**
	  * @param initEvent
	  */
	public ActiveObservable(E initEvent) {
		registeredObservers = new LinkedList<O>();
		
		this.lastEvent = initEvent;
	}
	

	// --------------------------------------------------------------------------
	// --- notify ---------------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * Notifies all registered observers.
	 * 
	 * @param event: The {@link Event} that changed the model
	 */
	@Override
	public void notifyObservers(E event) {
		if (event == null) {
			return;
		}
		
		synchronized (listSync) {
			lastEvent = event;
			
			doNotify(event);
		}
	}
	

	@Override
	public void notifyObservers() {
		
		synchronized (listSync) {
			doNotify(lastEvent);
		}
	}
	

	@SuppressWarnings("unchecked")
	private void doNotify(E event) {
		for (O observer : registeredObservers) {
			observer.update((S) this, event);
		}
	}
	

	// --------------------------------------------------------------------------
	// --- observer management --------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @return Whether all observers has been added without problems
	 */
	@Override
	public boolean addObservers(O... observers) {
		
		if (observers == null) {
			return false;
		}
		
		synchronized (listSync) {
			boolean result = true;
			
			for (O obs : observers) {
				if (obs != null) {
					registeredObservers.add(obs);
					obs.onAdd();
				} else {
					result = false;
				}
			}
			
			notifyObservers(lastEvent);
			
			return result;
		}
	}
	

	@Override
	public boolean removeObserver(O observer) {
		synchronized (listSync) {
			boolean containedObs = registeredObservers.remove(observer);
			
			if (containedObs) {
				observer.onRemove();
			}
			
			return containedObs;
		}
	}
}
